/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <string.h>
#include <locale.h>
#include <signal.h>
#include <unistd.h>
#include "numexp-kernel.h"

#define EXIT_TIMEOUT 10000
#define FACTORY_OAFIID		"OAFIID:Numexp_KernelFactory"
#define KERNEL_COMPONENT_OAFIID	"OAFIID:Numexp_Kernel"


static guint timeout_id = 0;
static gint refcount = 0;

static gchar *default_session_file = NULL;


gboolean timeout_func(gpointer data)
{
    g_message("Server has been inactive for some time. Quitting...");
    if (bonobo_main_level())
        bonobo_main_quit();
    else
        exit(0);
    return FALSE;
}

void decref(void)
{
    g_message("Object destroyed, refcount=%i.", refcount);
    if (--refcount <= 0 && !timeout_id) {
	g_message("No CORBA objects left alive.");
	timeout_id = g_timeout_add(EXIT_TIMEOUT, timeout_func, NULL);
    }
}


static BonoboObject * factory_callback(BonoboGenericFactory *factory,
				       const char           *component_id,
				       gpointer              closure)
{
    BonoboObject *object;
    if (strcmp(component_id, KERNEL_COMPONENT_OAFIID) == 0)
	object = BONOBO_OBJECT(numexp_kernel_new());
    else
	object = NULL;
    if (object) {
	++refcount;
	if (timeout_id) {
	    g_source_destroy(g_main_context_find_source_by_id(NULL, timeout_id));
	    timeout_id = 0;
	    g_message("Server shutdown canceled due to new request");
	}
	g_signal_connect(object, "destroy", G_CALLBACK(decref), NULL);
    }
    return object;
}



static void enter_shutdown(void)
{
    static gboolean in_shutdown = FALSE;

    if (in_shutdown) return;
    in_shutdown = TRUE;

    /* TODO: when albie makes this more accessible */
    /* numexp_save_session(default_session_file); */
}


static void
signal_handler(int signo)
{
    static gint in_fatal = 0;

    /* avoid loops */
    if (in_fatal > 0)
	return;

    ++in_fatal;

    switch (signo)
    {
	/* Fast cleanup only */
    case SIGSEGV:
    case SIGBUS:
    case SIGILL:
	enter_shutdown();
	g_message(_("Received signal %d, dumping core. Please report a NumExp bug."), 
		  signo);
	abort();
	break;

    case SIGFPE:
    case SIGPIPE:
	/* Go ahead and try the full cleanup on these,
	 * though it could well not work out very well.
	 */
	enter_shutdown();

	/* let the fatal signals interrupt us */
	--in_fatal;

	g_message(_("Received signal %d, shutting down abnormally. Please file a NumExp bug report."), 
		  signo);


	bonobo_main_quit();

	break;

    case SIGTERM:
    case SIGHUP:
	enter_shutdown();

	/* let the fatal signals interrupt us */
	--in_fatal;

	g_message(_("Received signal %d, shutting down cleanly"), signo);

        if (bonobo_main_level())
            bonobo_main_quit();
        else
            exit(0);
	break;

    default:
	break;
    }
}



int main(int argc, char *argv[])
{
    BonoboGenericFactory *factory;
    BONOBO_FACTORY_INIT("Numexp Kernel server", VERSION, &argc, argv);

    if (getenv("NUMEXP_DEBUG"))
	g_log_set_always_fatal(G_LOG_LEVEL_WARNING|G_LOG_LEVEL_CRITICAL);

    /* enable i18n */
    setlocale(LC_ALL, "");
    setlocale(LC_NUMERIC, "C");


    if (argv[1] && strcmp(argv[1], "--private") == 0) {
        CORBA_Environment ev;
        int ior_fd;
        CORBA_char *ior;
        if (argv[2] == NULL)
            ior_fd = 1;
        else
            ior_fd = atoi(argv[2]);
        g_message("starting private bonobo server, %i", ior_fd);
	factory = g_object_new(BONOBO_TYPE_GENERIC_FACTORY, NULL);
        bonobo_generic_factory_construct_noreg(
            factory, FACTORY_OAFIID,
            g_cclosure_new(G_CALLBACK(factory_callback), NULL, NULL));

        CORBA_exception_init(&ev);
        ior = CORBA_ORB_object_to_string(bonobo_orb(), BONOBO_OBJREF(factory), &ev);
        CORBA_exception_free(&ev);
        if (ior) {
            write(ior_fd, ior, strlen(ior));
            write(ior_fd, "\n", 1);
            CORBA_free(ior);
        } else
            g_warning("Unable to write IOR");
        if (ior_fd != 1)
            close(ior_fd);
        
    } else {
        factory = bonobo_generic_factory_new(FACTORY_OAFIID, factory_callback, NULL);
    }

    if (factory) {
	signal(SIGSEGV, signal_handler);
	signal(SIGBUS,  signal_handler);
	signal(SIGILL,  signal_handler);
	signal(SIGFPE,  signal_handler);
	signal(SIGPIPE, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP,  signal_handler);

	nxp_kernel_init();
	default_session_file = g_build_filename(g_get_home_dir(),
						".numexp-session.xml", NULL);
	if (g_file_test(default_session_file, G_FILE_TEST_EXISTS)) {
	    /* TODO */
	    /* nxp_load_session(default_session_file); */
	    unlink(default_session_file);
	}

	bonobo_main();
	bonobo_object_unref(BONOBO_OBJECT(factory));
	enter_shutdown();
	return bonobo_debug_shutdown();
    } else
	return 1;
}


