/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include "nxp-ticket-source.h"

typedef struct {
    guint16  ref_count;
    guint32  key;
    gpointer value;
} Ticket;


struct _NxpTicketSource {
    GHashTable *hash;		/* key -> Ticket */
    GHashTable *reverse_hash;	/* value -> Ticket */
};



/**
 * nxp_ticket_source_new:
 * 
 * Create a new #NxpTicketSource.
 * 
 * Return value: the ticket source
 **/
NxpTicketSource* nxp_ticket_source_new(void)
{
    NxpTicketSource *self = g_new(NxpTicketSource, 1);
    self->hash         = g_hash_table_new(g_direct_hash, g_direct_equal);
    self->reverse_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
    return self;
}


/**
 * nxp_ticket_source_create_ticket:
 * @self: the ticket source
 * @data: pointer to associate with ticket
 * 
 * Create a new ticket from the ticket source @self.
 * 
 * Return value: the new ticket
 **/
guint32 nxp_ticket_source_create_ticket(NxpTicketSource *self, gpointer data)
{
    Ticket *ticket;
    guint32 key;

    g_return_val_if_fail(self, 0);

    if ((ticket = g_hash_table_lookup(self->reverse_hash, data)) != NULL) {
	++ticket->ref_count;
	return ticket->key;
    }
    while (1) {
	key = g_random_int();
        if (key == 0) continue;
	if (g_hash_table_lookup(self->hash, GUINT_TO_POINTER(key)) == NULL) {
	    ticket = g_new(Ticket, 1);
	    ticket->ref_count = 1;
	    ticket->key       = key;
	    ticket->value     = data;
	    g_hash_table_insert(self->hash, GUINT_TO_POINTER(ticket->key), ticket);
	    g_hash_table_insert(self->reverse_hash, ticket->value, ticket);
	    return key;
	}
    }
    return 0;
}


/**
 * nxp_ticket_source_free_ticket:
 * @self: a ticket source
 * @ticket: a ticket obtained from nxp_ticket_source_create_ticket() from this source
 * 
 * Free a ticket from a ticket source. The data associated with the ticket is untouched.
 **/
void nxp_ticket_source_free_ticket(NxpTicketSource *self, guint32 ticket)
{
    Ticket *ticket_holder;

    g_return_if_fail(self);
    ticket_holder = g_hash_table_lookup(self->hash, GUINT_TO_POINTER(ticket));
    if (!ticket_holder) {
	g_warning("%s: Ticket %i is invalid in source %p",
		  __FUNCTION__, ticket, self);
	return;
    }
    if (!--ticket_holder->ref_count) {
	g_hash_table_remove(self->hash, GUINT_TO_POINTER(ticket));
	g_hash_table_remove(self->reverse_hash, ticket_holder->value);
	g_free(ticket_holder);
    }
}


/**
 * nxp_ticket_source_lookup_ticket:
 * @self: a ticket source
 * @ticket: a ticket obtained from nxp_ticket_source_create_ticket() from this source
 * 
 * Looks up a ticket in the ticket source.
 *
 * Return value: the data associated with the ticket, or %NULL if the
 * ticket is not valid.
 **/
gpointer nxp_ticket_source_lookup_ticket(NxpTicketSource *self, guint32 ticket)
{
    Ticket *ticket_holder;
    g_return_val_if_fail(self, NULL);
    ticket_holder = g_hash_table_lookup(self->hash, GUINT_TO_POINTER(ticket));
    if (!ticket_holder)
	return NULL;
    else
	return ticket_holder->value;
}

/**
 * nxp_ticket_source_lookup_data:
 * @self: a ticket source
 * @data: a data pointer for which exists a ticket
 * 
 * Does a reverse lookup data->ticket in the ticket source.
 *
 * Return value: the ticket associated with @data, or 0 the ticket was
 * not found.
 **/
guint32 nxp_ticket_source_lookup_data(NxpTicketSource *self, gpointer data)
{
    Ticket *ticket_holder;
    g_return_val_if_fail(self, 0);
    ticket_holder = g_hash_table_lookup(self->reverse_hash, data);
    if (!ticket_holder)
	return 0;
    else
	return ticket_holder->key;
}


/**
 * nxp_ticket_source_destroy:
 * @self: a ticket source
 * 
 * Frees a ticket source.
 **/
void nxp_ticket_source_destroy(NxpTicketSource *self)
{
    g_return_if_fail(self);
    g_hash_table_foreach(self->hash, (GHFunc) g_free, NULL);
    g_hash_table_destroy(self->hash);
    g_hash_table_destroy(self->reverse_hash);
    g_free(self);
}


