// -*- Mode: c; c-file-style: "stroustrup" -*-
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <string.h>

#include "nxp_kernel.h"
#include "types/int.h"
#include "types/real.h"
#include "types/nxp_complex.h"
#include "types/container.h"
#include "functions.h"
#include "parser.h"
#include "i18n.h"

//~~~ ln
NxpElement* stdlib_eval_r_ln(NxpElementR* el, ETC) {
    return (nxp_element_r_compare_int(el, 0) > 0)?
	nxp_element_real_new(log(nxp_element_r_get_value_as_double(el))):
	nxp_element_complex_new_from_dbl(clog(nxp_element_r_get_value_as_double(el)));
}

#ifdef HAVE_COMPLEX_H
NxpElement* stdlib_eval_complex_ln(NxpElementComplex* el, ETC) {
    return nxp_element_complex_new_from_dbl(clog(nxp_element_complex_get_value(el)));
}
#endif

NxpElement* stdlib_derive_ln (NxpIdentifier *var, NxpElement* arg, ETC) {
    return elem_OVER(nxp_element_derive(arg, var, context, NULL), NXP_ELEMENT_REF(arg));
}


NxpElement* bifunc_load(NxpElementString *s, ETC) {
    GError *erro = NULL;
    NxpElement* el;
    el = nxp_files_load(s->string, context, &erro);
    if (erro) g_propagate_error(error, erro);
    return el;    
}

static void bifunc_load_session_modules(xmlNode *node, NxpEvalContext *context) {
    xmlNode *iterator;
    if (node->children)
	for (iterator = node->children; iterator; iterator = iterator->next) {
	    if (g_str_equal(iterator->name, "module") && iterator->children &&
		xmlNodeIsText(iterator->children)) {
		nxp_module_load(xmlNodeGetContent(iterator->children), context);
	    } else {
		/* Ignore error */
	    }
	}
}

static void bifunc_load_session_namespaces(xmlNode *node, NxpEvalContext *context) {
    xmlNode *iterator;
    guchar *namespace = NULL;
    /* GError *err = NULL; */

    namespace = xmlGetProp(node, (guchar *)"name");
    if (namespace) {
	NxpElement *el, *ell;
	NxpNameSpace *ns, *parent_ns;
	NxpIdentifier *id;
	const char *name;

	/* Create the Namespace */
	if (!g_str_equal(namespace, "::")) {
	    namespace[xmlStrlen(namespace)-strlen(NXP_IDENTIFIER_PATH_SEP)]='\0';
	    id = nxp_identifier_from_string((char *)namespace, context, NULL);

	    if (id) {
		nxp_identifier_resolve_from_dir(id, context->root_name_space,
						&parent_ns, &name, context);
		if (parent_ns && name) ns = nxp_name_space_new(parent_ns, name);
	    }
	}

	/* Treat children */
	if (node->children)
	    for (iterator = node->children; iterator; iterator = iterator->next) {
		if (g_str_equal(iterator->name, "function") && iterator->children &&
		    xmlNodeIsText(iterator->children)) {

		    /* load function */
		    guchar *prototype  = xmlGetProp(iterator, (guchar *)"prototype");
		    guchar *expression = xmlNodeGetContent(iterator->children);
		    if (prototype && expression) {
			char *exp;
			exp = g_strdup_printf("define[%s%s,%s]",
					      namespace,
					      prototype,
					      expression);
			el = exp_parse_string(exp, context, NULL); /* FIXME/GERROR */
			if (el) {
			    ell = nxp_element_eval(el, context, NULL); /* FIXME/GERROR */
			    NXP_ELEMENT_UNREF(el);
			    NXP_ELEMENT_XUNREF(ell);
			}
		    }

		} else if (g_str_equal(iterator->name, "variable") && iterator->children &&
			   xmlNodeIsText(iterator->children)) {
		    
		    /* load variable */
		    guchar *varname = xmlGetProp(iterator, (guchar *)"name");
		    guchar *value = xmlNodeGetContent(iterator->children);
		    if (varname && value) {
			char *exp;
			exp = g_strdup_printf("%s%s=%s",namespace,varname,value);
			el = exp_parse_string(exp, context, NULL); /* FIXME/GERROR */
			if (el) {
			    ell = nxp_element_eval(el, context, NULL); /* FIXME/GERROR */
			    NXP_ELEMENT_UNREF(el);
			    NXP_ELEMENT_XUNREF(ell);
			}
		    }
		} else {
		    /* ignore error */
		}
	    }
    }
}

NxpElement *bifunc_load_session(NxpElementString *s, ETC)
{
    xmlDoc *doc;
    xmlNode *node, *iterator;
    doc = xmlParseFile(s->string);
    if (!doc) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_XML_PARSER_ERROR,
		    _("Error opening XML file '%s'"), s->string);
	return NULL;
    }

    node = xmlDocGetRootElement(doc);
    if (g_str_equal(node->name, "numexp-session")) {
	if (node->children)
	    for (iterator = node->children; iterator; iterator = iterator->next) {
		if (g_str_equal(iterator->name, "modules")) {
		    bifunc_load_session_modules(iterator, context);
		} else if (g_str_equal(iterator->name, "namespace")) {
		    bifunc_load_session_namespaces(iterator, context);
		} else {
		    /* Ignore error...go ahead */
		}
	    }
    } else {
	xmlFreeDoc(doc);
	g_set_error(error, nxp_error_domain, NXP_ERROR_XML_PARSER_ERROR,
		    "Error parsing XML file. Found '%s' and expecting 'numexp-session'",
		    node->name);
	return NULL;
    }
    xmlFreeDoc(doc);
    return nxp_element_void_new();
}

void session_dump(gpointer _key, gpointer _value, gpointer _user_data)
{
    GString *repr;
    xmlNode *variable, *function;
    char          *key = (char*)_key;
    NxpDefinition *def = (NxpDefinition*)_value;
    xmlNode *namespace = (xmlNode*)_user_data;

    switch (def->type) {
    case NXP_DEF_VARIABLE:
    {
	repr = nxp_element_repr(((NxpDefVariable*)def)->value);
	variable = xmlNewTextChild(namespace, NULL, (guchar *)"variable", (guchar *)repr->str);
	xmlNewProp(variable, (guchar *)"name", (guchar *)def->name);
	g_string_free(repr, TRUE);
    }
    break;
    case NXP_DEF_BI_FUNC:
	break;
    case NXP_DEF_USER_FUNC:
    {
	GSList *ptr;
	GString *prototype = g_string_new(key);
	int offset = ((char*)strstr(prototype->str,"/")-(char*)(prototype->str));
	prototype=g_string_truncate(prototype, offset);
	g_string_append_c(prototype, '[');
	ptr = ((NxpDefUserFunc*)def)->argument_list;
	for(;ptr;ptr=g_slist_next(ptr)) {
	    g_string_append(prototype, (char*)(ptr->data));
	    if (g_slist_next(ptr)) g_string_append_c(prototype, ',');
	}
	g_string_append_c(prototype, ']');
	repr = nxp_element_repr(((NxpDefUserFunc*)def)->expression);
	function = xmlNewTextChild(namespace, NULL, (guchar *)"function", (guchar *)repr->str);
	xmlNewProp(function, (guchar *)"prototype", (guchar *)prototype->str);
	g_string_free(repr, TRUE);
	g_string_free(prototype, TRUE);
    }
    break;
    }
}

void session_add_namespace(NxpNameSpace *ns, xmlNode *node)
{
    xmlNode *namespace;
    GString *path;

    namespace = xmlNewChild(node, NULL, (guchar *)"namespace", NULL);
    path = nxp_name_space_get_full_path_name(ns);
    xmlNewProp(namespace, (guchar *)"name", (guchar *)path->str);
    g_string_free(path, TRUE);

    g_hash_table_foreach(ns->definitions, session_dump, namespace);

    if (ns->child) session_add_namespace(ns->child, node);
    if (ns->next)  session_add_namespace(ns->next, node);
}

NxpElement *bifunc_save_session(NxpElementString *s, ETC)
{
    NxpNameSpace *ns;
    /* NxpNameSpace *child; 
       xmlChar   *rv = NULL; */
    xmlDoc    *doc;
    /* 
     * <session>
     *  <modules>
     *   <module>builtin</module>
     *  </modules>
     *  <namespace name="::">
     *   <variable name="ans">foo</variable>
     *   <function protoype="id[x]">x</function>  ----- define[id[x],x]
     *  </namespace>
     * </session>
     *
     */

    xmlNode   *session, *modules;
    /* xmlNode *module; */

    doc        = xmlNewDoc((guchar *)"1.0");

    session    = xmlNewChild((xmlNode *) doc, NULL, (guchar *)"numexp-session", NULL);
    modules    = xmlNewChild(session, NULL, (guchar *)"modules", NULL);
    /* Treat modules */
    {
	NxpModuleEntry **modules_list;
	int i;
	modules_list = nxp_modules_listing();
	for (i = 0; modules_list[i]; i++) 
	    if (modules_list[i]->loaded)
		xmlNewTextChild(modules, NULL, (guchar *)"module", (guchar *)modules_list[i]->name);
	nxp_modules_list_free(modules_list);
    }
    /* Treat namespaces */
    ns = context->root_name_space;
    session_add_namespace(ns, session);

    xmlSaveFile(s->string, doc);
    xmlFreeDoc(doc);
    return nxp_element_void_new();
}


/* undef */
NxpElement* bifunc_eval_undef_function(NxpElementVariable *var, ETC)
{
    char *varname;
    NxpNameSpace *ns, *actual_ns;
    NxpDefinition *def;
    const char *name;

    g_assert(NXP_ELEMENT_GET_TYPE(var) == NXP_ELEMENT_TYPE_VARIABLE);
    varname = nxp_identifier_to_string(var->ident);
    if (varname[0] == '_' || strstr(varname, "::_")) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't undefine a numexp private variable"));
	return NULL;
    }
    g_free(varname);

    nxp_identifier_resolve(var->ident, context, &ns, &name);
    if (ns && name) {
	def = nxp_name_space_lookup_definition_extended(ns, name, &actual_ns, FALSE);

	if (def) {
	    if (def->type == NXP_DEF_VARIABLE) {
//		NXP_ELEMENT_UNREF(NXP_DEF_VARIABLE(def)->value);
		nxp_definition_unref(def);
		nxp_name_space_remove_definition(ns, def);
	    }
	}
    }
    return nxp_element_void_new();    
}

/* aka mkdir */
NxpElement* bifunc_create_world(NxpElementVariable* el, ETC)
{
    NxpNameSpace *ns, *parent_ns;
    const char *name;

    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(el)->ident, context, &parent_ns, &name);
    if (!parent_ns) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("mkdir: Invalid identifier: it should not end with path separator."));
	return NULL;
    }
    if (!name) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("mkdir: Invalid identifier: it should not end with path separator."));
	return NULL;
    }

    ns = nxp_name_space_new(parent_ns, name);

    return nxp_element_void_new();
}


NxpElement* bifunc_eval_get_from_matrix(NxpElementMatrix *matrix,
					NxpElementInt *x, NxpElementInt *y, ETC)
{
    NxpElement *elem;
    gint row, col;

    row = nxp_element_int_get_si(NXP_ELEMENT(x), error);
    col = nxp_element_int_get_si(NXP_ELEMENT(y), error);

    if (col == 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("getFromMatrix: invalid second index value (column)"));
	return NULL;
    }
    if (row == 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("getFromMatrix: invalid first index value (row)"));
	return NULL;
    }
    /* Check boundaries -- row */
    if (row > matrix->height) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("getFromMatrix: row number (= %i) exceeds matrix height (= %i)"),
		    row, matrix->height);
	return NULL;
    }
    /* Check boundaries -- column */
    if (col > matrix->width) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("getFromMatrix: column number (= %i) exceeds matrix width (= %i)"),
		    col, matrix->width);
	return NULL;
    }
    elem = _element_matrix_get(matrix, row - 1, col - 1);
    return elem;
}


/* replaceOnMatrix[matrix, row, col, replacement] */
NxpElement* bifunc_eval_replace_on_matrix(NxpElementMatrix* matrix,
					  NxpElementInt* row_el,
					  NxpElementInt* col_el,
					  NxpElement* value, ETC)
{				
    gint row, col;
    NxpElement *result;

    row = nxp_element_int_get_si(NXP_ELEMENT(row_el), error);
    if (row <= 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("replaceOnMatrix: row number is not greater than zero."));
	return NULL;
    }

    col = nxp_element_int_get_si(NXP_ELEMENT(col_el), error);
    if (col <= 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("replaceOnMatrix: column number is not greater than zero."));
	return NULL;
    }

    result = nxp_element_matrix_real_put(matrix, value, row - 1, col - 1);

    return result;
}


NxpElement* bifunc_eval_replace_on_vecvar(NxpElementVariable* var,
					  NxpElementInt* pos,
					  NxpElement* value, ETC)
{
    NxpElement *vector;
    NxpElement *new_vector = NULL;

    vector = nxp_variable_get(var->ident, context);
    if (vector)
	if (NXP_ELEMENT_IS_VECTOR(vector)) {
	    new_vector = bifunc_eval_replace_on_vector(NXP_ELEMENT_VECTOR(vector),
						       pos, value, NULL, context, error);
	    nxp_user_variable_define(var->ident, new_vector, context, error);
	} else g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, 
			   _("variable should contain a vector"));
    else {
	char *name = nxp_identifier_to_string(var->ident);
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, _("undefined variable '%s'"), name);
	g_free(name);
    }

    return new_vector;
}


NxpElement* bifunc_eval_replace_on_matvar(NxpElementVariable* var,
					  NxpElementInt *row_el,
					  NxpElementInt *col_el,
					  NxpElement* value, ETC)
{
    NxpElement *matrix;
    NxpElement *new_matrix = NULL;

    matrix = nxp_variable_get(var->ident, context);
    if (matrix)
	if (NXP_ELEMENT_IS_MATRIX(matrix)) {
	    new_matrix = bifunc_eval_replace_on_matrix(NXP_ELEMENT_MATRIX(matrix),
						       row_el, col_el, value, NULL, context, error);
	    nxp_user_variable_define(var->ident, new_matrix, context, error);
	} else g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, 
			   _("variable should contain a matrix"));
    else {
	char *name = nxp_identifier_to_string(var->ident);
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, _("undefined variable '%s'"), name);
	g_free(name);
    }
    
    return new_matrix;
}

NxpElement* bifunc_eval_get_from_matvar(NxpElementVariable* var,
					NxpElementInt *row_el,
					NxpElementInt *col_el, ETC)
{
    NxpElement *matrix;
    NxpElement *elem = NULL;

    matrix = nxp_variable_get(var->ident, context);
    if (matrix)
	if (NXP_ELEMENT_IS_MATRIX(matrix)) {
	    elem = bifunc_eval_get_from_matrix(NXP_ELEMENT_MATRIX(matrix),
					       row_el, col_el, NULL,
					       context, error);
	}
	else g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, 
			   _("variable should contain a matrix"));
    else {
	char *name = nxp_identifier_to_string(var->ident);
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, _("undefined variable '%s'"), name);
	g_free(name);
    }

    return elem;
}

NxpElement* bifunc_eval_replace_on_vector(NxpElementVector* vector,
					  NxpElementInt* pos,
					  NxpElement* value, ETC)
{
    gint index;
    NxpElement *result;

    index = nxp_element_int_get_si(NXP_ELEMENT(pos), error);
    if (index <= 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("replaceOnVector: index number is not greater than zero."));
	return NULL;
    }

    result = nxp_element_vector_real_put(vector, value, index - 1);
    return result;
}



static nxp_int_t factorial_i(nxp_int_t j)
{
    nxp_int_t f = 1;
    for (; j > 1; j--) f *= j;
    return f;
}

static nxp_real_t factorial_r(nxp_real_t j)
{
    nxp_real_t f = 1;
    for (; j > 1; j--) f *= j;
    return f;
}

NxpElement* bifunc_eval_factorial(NxpElementInt* a, ETC)
{
    nxp_int_t j = nxp_element_int_get_si(NXP_ELEMENT(a), error);

    if (j < 0)
    {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("Invalid argument to the factorial function (a non-negative integer must be provided)!"));
  	return NULL;
    }
    if (j > 25)
	return nxp_element_real_new(factorial_r(j));
    else
	return nxp_element_int_new(factorial_i(j));
}

NxpElement* bifunc_eval_round(NxpElementReal* element, ETC)
{
    return nxp_element_int_new((nxp_int_t)(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(element)) + 0.5));
}

NxpElement* bifunc_eval_round2(NxpElementReal* element,
			       NxpElementInt* n, ETC)
{
    int decimal = ABS(nxp_element_int_get_si(NXP_ELEMENT(n), error));
    double t;
    int mult;
    mult = pow(10, decimal);
    t = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(element));
    /* TODO: Check if 'long long int' can be nxp_int or something like that. */
    return nxp_element_real_new(((double)((long long int)(t*mult + 0.5))) / mult);
}

NxpElement* bifunc_eval_print(GSList* args, ETC)
{
    NxpElement *el;
    GSList *l;
    GString *str, *final;

    final = g_string_new("");
    l = args;
    while (l)
    {
	el = nxp_element_eval(NXP_ELEMENT(l->data), context, NULL);
	if (!el) return NULL;

	// this 'if' is to remove quotes from string elements
	if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_STRING)
	    str = g_string_new(NXP_ELEMENT_STRING(el)->string);
	else
	    elem2string(&str, el);

	g_string_append(final, str->str);
	NXP_ELEMENT_UNREF(el);
	g_string_free(str, TRUE);
	l = g_slist_next(l);
    }
    nxp_output_message(context, "%s", final->str);
    return nxp_element_void_new();
}

NxpElement* bifunc_simp_print(GSList* args, ETC)
{
    NxpElement *el;
    GSList *l;
    GString *str, *final;

    final = g_string_new("");
    l = args;
    while (l)
    {
	el = nxp_element_simplify(NXP_ELEMENT(l->data), context, NULL);
	if (!el) return NULL;

	// this 'if' is to remove quotes from string elements
	if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_STRING)
	    str = g_string_new(NXP_ELEMENT_STRING(el)->string);
	else
	    elem2string(&str, el);

	g_string_append(final, str->str);
	NXP_ELEMENT_UNREF(el);
	g_string_free(str, TRUE);
	l = g_slist_next(l);
    }
    nxp_output_message(context, "%s", final->str);
    return nxp_element_void_new();
}


NxpElement* bifunc_eval_quote(NxpElement* el, ETC)
{
    return NXP_ELEMENT_REF(el);
}




NxpElement* bifunc_eval_eval(NxpElement* el, ETC)
{
    NxpElement *el1, *el2;
    GError *err = NULL;

    el1 = nxp_element_eval(el, context, &err);
    if (err || !el1) {
	g_propagate_error(error, err);
	return NULL;
    }

    el2 = nxp_element_eval(el1, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    NXP_ELEMENT_UNREF(el1);
    return el2;
}


/* Simplify[Expression] */
NxpElement* bifunc_eval_simplify(NxpElement* el, ETC)
{
    GError *myerror = NULL;
    NxpElement *result;

    result = nxp_element_simplify(el, context, &myerror);
    if (myerror) g_propagate_error(error, myerror);
    return result;
}



/* ToInt[Expression] */
NxpElement* bifunc_eval_toint(GSList* args, ETC)
{
    NxpElement_t* return_value;
    NxpElement_t *element;
    if (!(element = (NxpElement_t*)args->data)) return NULL;

    if (!(element = nxp_element_eval(element, context, NULL))) return NULL;

    if (NXP_ELEMENT_GET_TYPE(element) == NXP_ELEMENT_TYPE_REAL) {
        return_value = nxp_element_real_new((long long int)NXP_ELEMENT_REAL(element)->value);
        NXP_ELEMENT_UNREF(element);
    }
    else {
        NXP_ELEMENT_UNREF(element);
        return_value = NULL;
    }
    return return_value;
}


/* Well... D[If[Condicao,x,y]] = If[Condicao,D[x],D[y]] */
NxpElement* bifunc_derivate_if(NxpIdentifier *var, GSList* args, ETC)
{
    gint n = g_slist_length(args);
    switch (n)
    {
    case 2:
        return elem_IF_THEN(NXP_ELEMENT_REF((NxpElement*)args->data), 
			    nxp_element_derive((NxpElement*)g_slist_next(args)->data, 
					       var, context, NULL));
    case 3:
        return
	    elem_IF_THEN_ELSE(NXP_ELEMENT_REF((NxpElement*)
					      args ->data), 
			      nxp_element_derive((NxpElement*)
						 g_slist_next(args)-> data, var,
						 context, NULL), 
			      nxp_element_derive((NxpElement*)
						 g_slist_next(g_slist_next(args))
						 -> data, var, context, NULL));
    default:
        return NULL;
    }
}

NxpElement* bifunc_eval_d(NxpElement* tree, NxpElementVariable* var, ETC) {
    NxpElement* arg;
    NxpElement* el;
    GError *myerr = NULL;

    arg =  nxp_element_derive(tree, var->ident, context, &myerr);
    if (myerr) {
	g_propagate_error(error, myerr);
	return NULL;
    }
    g_return_val_if_fail(arg, NULL);
    el = nxp_element_eval(arg, context,&myerr);
    if (myerr) g_propagate_error(error, myerr);
    NXP_ELEMENT_UNREF(arg);
    return el; 
}

NxpElement* bifunc_eval_n_d(NxpElement* tree, NxpElementVariable* var, NxpElementInt* n, ETC) {
    NxpElement* arg = NULL;
    NxpElement* el;
    GError *myerr = NULL;
    int times = nxp_element_int_get_si(NXP_ELEMENT(n), error);
    gint tofree = 0;

    if (times <= 0) {
	/* TODO: give some more usefull error */
	return NULL;
    } else {
	while(times--) {
	    arg =  nxp_element_derive(tree, var->ident, context, &myerr);
	    if (myerr) {
		g_propagate_error(error, myerr);
		return NULL;
	    }
	    /* This means we will free all times, except the very
	     * first one */
	    if (tofree) NXP_ELEMENT_UNREF(tree);
	    tofree = 1;

	    tree = arg;

	}
    }
    g_return_val_if_fail(arg, NULL);
    el = nxp_element_eval(arg, context, &myerr);
    if (myerr) g_propagate_error(error, myerr);
    NXP_ELEMENT_UNREF(arg);
    return el; 
}


NxpElement* bifunc_simplify_d(NxpElement* tree, NxpElementVariable* var, ETC) {
    NxpElement* arg, *arg2;
    GError *myerr = NULL;

    arg =  nxp_element_derive(tree,var->ident, context,&myerr);
    if (myerr) {
	g_propagate_error(error, myerr);
	return NULL;
    }
    g_return_val_if_fail(arg, NULL);
    arg2 = nxp_element_simplify(arg, context, NULL);
    NXP_ELEMENT_UNREF(arg);
    return arg2;
}


NxpElement* bifunc_simplify_n_d(NxpElement* tree, NxpElementVariable* var, NxpElementInt* n, ETC) {
    NxpElement* arg;
    GError *myerr = NULL;
    int times = nxp_element_int_get_si(NXP_ELEMENT(n), error);

    if (times <= 0) {
	/* TODO: give some more usefull error */
	return NULL;
    }
    NXP_ELEMENT_REF(tree);
    while(times--) {
	arg = nxp_element_derive(tree, var->ident, context, &myerr);
	NXP_ELEMENT_UNREF(tree);
	if (myerr) {
	    g_propagate_error(error, myerr);
	    return NULL;
	}
	tree = nxp_element_simplify(arg, context, NULL); 
	NXP_ELEMENT_UNREF(arg);
    }
    g_return_val_if_fail(tree, NULL);
    return tree;
}

NxpElement* bifunc_eval_is_real(NxpElement* el, ETC) {
    NxpElement *x = nxp_element_eval(el, context, NULL);
    if (x) {
	NxpElement *r = nxp_element_int_new(NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_REAL);
	nxp_element_unref(x);
	return r;
    }
    return nxp_element_int_new(0);
}
NxpElement* bifunc_eval_is_int(NxpElement* el, ETC) {
    NxpElement *x = nxp_element_eval(el, context, NULL);
    if (x) {
	NxpElement *r = nxp_element_int_new(NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_INT);
	nxp_element_unref(x);
	return r;
    }
    return nxp_element_int_new(0);
}
NxpElement* bifunc_eval_is_complex(NxpElement* el, ETC) {
    NxpElement *x = nxp_element_eval(el, context, NULL);
    if (x) {
	NxpElement *r = nxp_element_int_new(NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_COMPLEX);
	nxp_element_unref(x);
	return r;
    }
    return nxp_element_int_new(0);
}
NxpElement* bifunc_eval_is_matrix(NxpElement* el, ETC) {
    NxpElement *x = nxp_element_eval(el, context, NULL);
    if (x) {
	NxpElement *r = nxp_element_int_new(NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_MATRIX);
	nxp_element_unref(x);
	return r;
    }
    return nxp_element_int_new(0);
}

NxpElement* bifunc_eval_is_vector(NxpElement* el, ETC) {
    NxpElement *x = nxp_element_eval(el, context, NULL);
    if (x) {
	NxpElement *r = nxp_element_int_new(NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_VECTOR);
	nxp_element_unref(x);
	return r;
    }
    return nxp_element_int_new(0);
}

