/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "element_type.h"
#include "op_bindings.h"
#include "typecasts.h"
#include <string.h>


/* A GHashTable where values are NxpElementTypeInfo structures and
   indices are element types (ElementType). */
static GHashTable      *element_types_table		= NULL;
int             nxp_element_last_type            = 0;

static void type_info_ghfunc(gpointer key, gpointer value, gpointer user_data);

/* -------------- type info -------------------------------------- */

#define MAKE_TYPE_INFO_KEY(type) GINT_TO_POINTER(type)


/**
 * nxp_element_types_init:
 * 
 * Initialize element types subsystem
 **/
void nxp_element_type_init(void)
{
    g_return_if_fail(element_types_table == NULL);

    element_types_table = g_hash_table_new(NULL, NULL);
}

int nxp_element_next_type(void) {
    return ++nxp_element_last_type;
}

/**
 * nxp_element_types_init:
 * 
 * Finalize element types subsystem
 **/
void nxp_element_type_fini(void)
{
    g_return_if_fail(element_types_table);
    g_hash_table_foreach(element_types_table, type_info_ghfunc, NULL);
    g_hash_table_destroy(element_types_table);
    element_types_table = NULL;
}

typedef struct _ChildrenData {
    NxpElementType type;
    GSList *lst;
} ChildrenData;

static void _get_children(gpointer key, gpointer value, gpointer user_data)
{
    NxpElementTypeInfo *typeinfo = (NxpElementTypeInfo *)value;
    ChildrenData *data = (ChildrenData *)user_data;

    if (typeinfo && typeinfo->parent && typeinfo->parent->type == data->type)
	data->lst = g_slist_append(data->lst, GINT_TO_POINTER(typeinfo->type));
}

GSList* nxp_element_type_get_children_types(NxpElementType type)
{
    ChildrenData data = {type, NULL};

    g_hash_table_foreach(element_types_table, _get_children, &data);
    return data.lst;
}

static void nxp_element_type_info_destroy(NxpElementTypeInfo *typeinfo)
{
    g_hash_table_destroy(typeinfo->user_data);
    g_free(typeinfo->name);
    g_free(typeinfo);
}


static NxpElementTypeInfo* nxp_element_type_info_new(guint size)
{
    NxpElementTypeInfo* type_info;
    type_info = (NxpElementTypeInfo *)g_malloc0(size);

    /* TODO : If a child destroys totaly its user_data, its parent can't do it
     * because it is already done. So we call here g_hash_table_new instead of
     * g_hash_table_new_full and when user_data is destroyed the datas pointed
     * to are not erased and we don't have the previous difficulty. There is a
     * memory lack instead... */
    type_info->user_data = g_hash_table_new(g_str_hash, g_str_equal);
    return type_info;
}

/**
 * Function needed by nxp_element_type_register_info to copy a hash table
 **/
static void copy_table(gpointer key, gpointer value, gpointer data)
{
    NxpElementTypeInfo *typeinfo = data;

    g_hash_table_insert(typeinfo->user_data, key, value);
}

/**
 * nxp_element_type_register_info:
 * @type: the type to register
 * @parent: the type of the parent, if 0 is given, the new element comes from
 * nothing.
 * @size: This is the size of the TypeInfo structure of our new element.
 *
 * Creates and fills the TypeInfo structure of the element of type @type from
 * the element of type @parent.
 * Return value: a pointer to the new NxpElementTypeInfo structure.
 **/
NxpElementTypeInfo* nxp_element_type_register_info(NxpElementType type,
						   guint size,
						   NxpElementType parent,
						   guint parent_size)
{
    NxpElementTypeInfo *typeinfo;
    NxpElementTypeInfo *parent_typeinfo = NULL;

    g_return_val_if_fail((parent && size >= parent_size)
	    || (parent == 0 && parent_size == sizeof(NxpElementTypeInfo)), NULL);

    if (!element_types_table) {
	g_message("Uninitialized kernel!!!");
	g_on_error_query("numexp");
    }
    g_return_val_if_fail(element_types_table, NULL);

    if (parent) {
	parent_typeinfo = g_hash_table_lookup(element_types_table,
					      MAKE_TYPE_INFO_KEY(parent));
	if (!parent_typeinfo) {
	    g_message("Parent does not exist, I can not create child!!!");
	    g_on_error_query("numexp");
	    g_return_val_if_fail(parent_typeinfo, NULL);
	}
    }

    typeinfo = g_hash_table_lookup(element_types_table, 
				   MAKE_TYPE_INFO_KEY(type));
    if (!typeinfo) {
	typeinfo = nxp_element_type_info_new(size);
	typeinfo->parent = parent_typeinfo;
	typeinfo->type = type;
	if (parent) {
	    nxp_typecasts_heritate_caster(parent, type);
	    g_hash_table_foreach(parent_typeinfo->user_data,
				 copy_table, typeinfo);
	    memcpy((gpointer)typeinfo + 3*sizeof(gpointer) + sizeof(NxpElementType),
		   (gpointer)parent_typeinfo + 3*sizeof(gpointer) + sizeof(NxpElementType),
		   parent_size - 3*sizeof(gpointer) - sizeof(NxpElementType));
	}
	g_hash_table_insert(element_types_table,
			    MAKE_TYPE_INFO_KEY(type),
			    typeinfo);
	nxp_op_binding_propagate(typeinfo);
    }
    return typeinfo;
}

/**
 * nxp_element_type_get_info:
 * @type: the type to lookup
 * 
 * Get the NxpElementTypeInfo structure associated with type @type. This
 * structure can be accessed directly to register all sorts of functions.
 * Functions can be NULL, which means no function is available for that
 * task.
 * Return value: a pointer to the NxpElementTypeInfo structure.
 **/
NxpElementTypeInfo* nxp_element_type_get_info(NxpElementType type)
{
    NxpElementTypeInfo *typeinfo;
    if (!element_types_table) {
	g_message("Uninitialized kernel!!!");
	g_on_error_query("numexp");
    }
    g_return_val_if_fail(element_types_table, NULL);
    typeinfo = g_hash_table_lookup(element_types_table, 
				   MAKE_TYPE_INFO_KEY(type));
    if (!typeinfo) {
	g_message("Uninitialized object!!!");
//	g_on_error_query("numexp");
    }
    return typeinfo;
}


static void type_info_ghfunc(gpointer key, gpointer value, gpointer user_data)
{
    nxp_element_type_info_destroy((NxpElementTypeInfo *)value);
}

gpointer nxp_element_type_get_data (NxpElementTypeInfo  *type_info,
				    const char* property_name)
{
    return g_hash_table_lookup((GHashTable*)(type_info->user_data), property_name);
}

void nxp_element_type_set_data(NxpElementTypeInfo  *type_info,
				const char* property_name, gpointer data)
{
    GSList *children, *lst;
    NxpElementTypeInfo *child_typeinfo;

    g_hash_table_insert((GHashTable*)(type_info->user_data), g_strdup(property_name), data);

    /* We set the data for each child of this type_info */
    children = nxp_element_type_get_children_types(type_info->type);
    for (lst = children; lst; lst = lst->next) {
	child_typeinfo = nxp_element_type_get_info(GPOINTER_TO_INT(lst->data));
	nxp_element_type_set_data(child_typeinfo, property_name, data);
    }
    g_slist_free(children);
}

