/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <glob.h>
#include <stdlib.h>
#include <regex.h>

#include <glib.h>

#include "nxp_kernel.h"
#include "i18n.h"

static GSList *module_path = NULL;

GPtrArray *nxp_loaded_modules = NULL;
NxpModule *nxp_current_module = NULL;


static void get_module_description(NxpModuleEntry *entry)
{
    /* ... */
}


static void init_module(NxpModule *module, NxpEvalContext *context)
{
    gpointer sym;
    NxpModuleInitFunc init_func;

    if (!g_module_symbol(module->module, "nxp_module_init", &sym)) {
	g_warning("Module is not for numexp");
	return;
    }
    nxp_current_module = module;

    init_func = (NxpModuleInitFunc) sym;
    g_return_if_fail(init_func);

    init_func(context, NULL);
    nxp_current_module = NULL;
}


static void
nxp_module_entry_free(NxpModuleEntry *entry)
{
    g_free(entry->name);
    g_free(entry->location);
    g_free(entry->description);
    g_free(entry->longer_desc);
    g_free(entry);
}


NxpModule* nxp_module_load(const unsigned char *name, NxpEvalContext *context)
{
    GSList *dir;
    NxpModule *nxpmodule = NULL;
    GModule *module = NULL;
    gchar *file = NULL;
    gchar *module_dir = NULL;

    g_return_val_if_fail(name, NULL);
    if (!module_path) nxp_module_rebuild_search_path();

    if (name[0] == '/') {
	file = g_strdup((char *)name);
	module = g_module_open(file, 0);
        module_dir = g_path_get_dirname(file);
    } else {
	for (file = NULL, dir = module_path; dir; dir = dir->next)
	{
	    struct stat buf;
            char *modname;
	    // free 'file' mem each time it loops
	    if (file) g_free(file);
            modname = g_strconcat((char *)name, ".la", NULL);
            file = g_build_filename((gchar*)dir->data, modname, NULL);
            g_free(modname);
	    if (stat(file, &buf)) {
                g_free(file);
                file = g_module_build_path((gchar*)dir->data, (gchar *)name);
            }
	    if (stat(file, &buf) == 0) {
		module = g_module_open(file, 0);
		if (module) {
                    module_dir = g_strdup((gchar *) dir->data);
                    break;
                }
		g_warning("Failed to load module %s from %s (%s)",
			  name, (gchar*)dir->data, g_module_error());
	    }
	}
    }
    if (module) {
	nxpmodule = g_new0(NxpModule, 1);
	g_ptr_array_add(nxp_loaded_modules, nxpmodule);
	nxpmodule->module   = module;
	nxpmodule->filename = file;
	nxpmodule->name     = g_strdup((gchar *)name);
	// get_module_xml_file(nxpmodule);
	g_module_make_resident(module);
	g_message("Module %s has been loaded from %s.", name, module_dir);
        g_free(module_dir);
	init_module(nxpmodule, context);
    } else {
	g_message("Failed to load module %s (%s).", name, g_module_error());
	g_free(file);
    }
    return nxpmodule;
}




void nxp_module_rebuild_search_path(void)
{
    char *env_path;
    gchar **dirs, **dir;
    if (module_path) {
	g_slist_foreach(module_path, (GFunc)g_free, NULL);
	module_path = NULL;
    }
    module_path = g_slist_prepend(module_path, g_strdup(NUMEXP_MODULES_DIR));

    env_path = getenv("NUMEXP_MODULE_PATH");
    if (env_path)
    {
	dirs = g_strsplit(env_path, ":", 0);
	for (dir = dirs; *dir; dir++)
	    module_path = g_slist_prepend(module_path, g_strdup(*dir));
	g_strfreev(dirs);
    }
}


static NxpElement* import_eval_func(NxpElementString *str, ETC)
{
    const char *module;

    module = str->string;
    nxp_module_load((guchar *)module, context);
    return nxp_element_void_new();
}


static NxpElement* listModules_eval_func(NxpEvalContext *context, GError **error)
{
    NxpModuleEntry **entries, *entry;
    GPtrArray *outer_vec_contents;
    int i;
    NxpElement       *vec;
    NxpElementVector *vec1;

    entries = nxp_modules_listing();
    g_return_val_if_fail(entries, NULL);
    outer_vec_contents = g_ptr_array_new();
    for (i = 0; (entry = entries[i]); ++i)
    {
	vec1 = NXP_ELEMENT_VECTOR(nxp_element_vector_new(5, NULL));
	_element_vector_index(vec1, 0) = nxp_element_string_new(entry->name);
	_element_vector_index(vec1, 1) = nxp_element_string_new(entry->location);
	_element_vector_index(vec1, 2) = entry->description?
	    nxp_element_string_new(entry->description) : nxp_element_void_new();
	_element_vector_index(vec1, 3) = nxp_element_int_new(entry->loaded? 1 : 0);
	_element_vector_index(vec1, 4) = entry->longer_desc?
	    nxp_element_string_new(entry->longer_desc) : nxp_element_void_new();
	g_ptr_array_add(outer_vec_contents, vec1);
    }
    nxp_modules_list_free(entries);
    vec = nxp_element_vector_new(outer_vec_contents->len,
				 (NxpElement **) outer_vec_contents->pdata);
    g_ptr_array_add(outer_vec_contents, FALSE);
    return vec;
}

/**
 * nxp_module_init:
 * @context: 
 * 
 * This function initializes the module loading subsystem, not any specific module!
 **/
void nxp_module_init(NxpEvalContext *context, GError **error)
{
    /* FIXME: so that the help works, these two functions must be
     * registered into the builtin "namespace" */
    nxp_bifunc_define_global("import", &nxp_prototype__STRING,
			     import_eval_func, 
			     NULL, import_eval_func, context, NULL);
    nxp_bifunc_define_global("listModules", &nxp_prototype__VOID,
			     listModules_eval_func, NULL, NULL, context, NULL);

    nxp_bifunc_define_global("gcd", &nxp_prototype__INT_INT,
			     nxp_element_int_gcd, NULL, nxp_element_int_gcd, context, NULL);
    nxp_bifunc_define_global("inverse", &nxp_prototype__INT_INT,
			     nxp_element_int_inverse, NULL, nxp_element_int_inverse, context, NULL);
    nxp_bifunc_define_global("normalize", &nxp_prototype__INT_INT,
			     nxp_element_int_normalize, NULL, nxp_element_int_normalize, context, NULL);
    if (nxp_loaded_modules == NULL)
        nxp_loaded_modules = g_ptr_array_new();
}

char* nxp_help(const char* topic, NxpEvalContext *context)
{
    /* First, check if it is a namespace or not */
    int topic_size = strlen(topic);
    char *contents;
    GString *path = g_string_new(topic);
    if (strncmp(&topic[topic_size-2],
                NXP_IDENTIFIER_PATH_SEP,
                strlen(NXP_IDENTIFIER_PATH_SEP)) == 0) {
        /* Should be a namespace */
        char *t = path->str;
        while(*t) {
            if (*t == ':') *t = '/';
            t++;
        }
        path = g_string_append(path, "_.xml");
        path = g_string_prepend(path, "/docs");
        path = g_string_prepend(path, NXP_DOCUMENTATION_DIR);

        if (!g_file_get_contents(path->str, &contents, NULL, NULL)) {
            contents = g_strdup_printf(_("Documentation not found for '%s'"),
                                       topic);        
        }
    } else {
        /* Should not be a namespace */
        char *t = path->str;
        while(*t) {
            if (*t == ':') *t = '/';
            t++;
        }
        path = g_string_append(path, ".xml");
        path = g_string_prepend(path, "/docs");
        path = g_string_prepend(path, NXP_DOCUMENTATION_DIR);

        if (!g_file_get_contents(path->str, &contents, NULL, NULL)) {
            contents = g_strdup_printf(_("Documentation not found for '%s'"),
                                       topic);        
        }
    }
    g_string_free(path, TRUE);
    return contents;
}

GList* nxp_help_list(NxpEvalContext *context)
{
    return NULL;
}


static NxpModuleEntry* _nxp_modules_listing_helper(const char *path)
{
    NxpModuleEntry  *entry;
    static gboolean  rx_initialized = FALSE;
    static regex_t   rx;
    regmatch_t       matches[2];
    gchar           *name;
    int              i;

    /* get module name from path name */
    if (!rx_initialized) {
	regcomp(&rx,".*/([^/]+)\\.(la|"G_MODULE_SUFFIX")$", REG_EXTENDED);
	rx_initialized = TRUE;
    }
    if (regexec(&rx, path, 2, matches, 0) == 0) {
	int off, len;
	off  = matches[1].rm_so;
	len  = matches[1].rm_eo - matches[1].rm_so;
	name = g_strndup(path + off, len);
    } else {
	g_warning("Module file '%s' does not match regexp", path);
	return NULL;
    }
    entry = g_new0(NxpModuleEntry, 1);
    entry->location = g_strdup(path);
    entry->name     = name;
    /* check if the module has been loaded already */
    entry->loaded   = FALSE;
    for (i = 0; i < nxp_loaded_modules->len; ++i) {
	NxpModule *module = g_ptr_array_index(nxp_loaded_modules, i);
	if (strcmp(entry->name, module->name) == 0) {
	    entry->loaded = TRUE;
	    break;
	}
    }
    get_module_description(entry);
    return entry;
}


NxpModuleEntry** nxp_modules_listing(void)
{
    GPtrArray      *array;
    NxpModuleEntry *entry, **retval;
    GSList         *path_entry;
    glob_t          globvec;
    gboolean        globvec_used = FALSE;

    memset(&globvec, 0, sizeof(globvec));
    array = g_ptr_array_new();
    if (!module_path)
        nxp_module_rebuild_search_path();
    for (path_entry = module_path; path_entry; path_entry = path_entry->next)
    {
	char const *dir;
	char       *pattern;
	int         globerr;
	int         gl_i;
        int         i;
	char const *globentry;

	dir = (char const *) path_entry->data;

	pattern = g_strdup_printf("%s/*.la", dir);
	globerr = glob(pattern, 0, NULL, &globvec);
	g_free(pattern);
        if (globerr == GLOB_NOMATCH)
            continue;
	if (G_UNLIKELY(globerr)) {
            g_warning("glob error %i", globerr);
            continue;
        }

	pattern = g_strdup_printf("%s/*."G_MODULE_SUFFIX, dir);
	globerr = glob(pattern, GLOB_APPEND, NULL, &globvec);
	g_free(pattern);

	globvec_used = TRUE;
	for (gl_i = globvec.gl_offs; gl_i < globvec.gl_pathc; gl_i++) {
	    globentry = globvec.gl_pathv[gl_i];
	    entry = _nxp_modules_listing_helper(globentry);
            /* look for a module with the same name already found */
            for (i = 0; i < array->len; ++i) {
                NxpModuleEntry *old_entry;
                old_entry = (NxpModuleEntry *) array->pdata[i];
                if (!strcmp(old_entry->name, entry->name)) {
                    /* we found duplicate module; throw this new one
                     * away, keep the existing one */
                    nxp_module_entry_free(entry);
                    entry = NULL;
                    break;
                }
            }
            if (entry)
                g_ptr_array_add(array, entry);
	}
    }
    if (globvec_used) globfree(&globvec);
    g_ptr_array_add(array, NULL);
    retval = (NxpModuleEntry **) array->pdata;
    g_ptr_array_free(array, FALSE);
    return retval;
}

void nxp_modules_list_free(NxpModuleEntry **entries)
{
    NxpModuleEntry *entry;
    int i;
    for (i = 0; (entry = entries[i]); ++i) {
	g_free(entry->name);
	g_free(entry->location);
	if (entry->description)
	    g_free(entry->description);
	g_free(entry);
    }
    g_free(entries);
}

/* void nxp_open_help_db(DB **db) { */
/*     GSList *dir; */
/*     nxp_module_rebuild_search_path(); */
/*     for (dir = module_path; dir; dir = dir->next) */
/*     { */
/* 	GString *str; */
/* 	FILE *fd; */
	
/* 	str = g_string_new((gchar*)dir->data); */
/* 	str = g_string_append(str, "/language.db"); */
/* 	fd = fopen(str->str, "r"); */
/* 	if (fd) { */
/* 	    fclose(fd); */
/* 	    open_db_file(str->str, db); */
/* 	    g_string_free(str, TRUE); */
/* 	    break; */
/* 	} else { */
/* 	    g_string_free(str, TRUE); */
/* 	} */
/*     } */
/* } */
