/* -*- Mode: C; c-file-style: "bsd"; -*- */

/* 
   Operator bindings
   -----------------
*/

#include "op_bindings.h"
#include "derive.h"
#include "i18n.h"
#include "nxp_debug.h"

/* Important!: The order of these entries must be the same as in the
   NxpOp enumeration. */
static NxpOpInfo opinfo[NXP_NUM_OPS] = 
{
    { 
	NXP_OP_PLUS,
	"+",
	NXP_OP_BINARY,
	50
    },
    { 
	NXP_OP_MINUS,
	"-",
	NXP_OP_BINARY,
	50
    },
    { 
	NXP_OP_MINUS_LEFT,
	"-",
	NXP_OP_UNARY_LEFT,
	45
    },
    { 
	NXP_OP_STAR,
	"*",
	NXP_OP_BINARY,
	40
    },
    {
	NXP_OP_SLASH,
	"/",
	NXP_OP_BINARY,
	40
    },
    { 
	NXP_OP_HAT,
	"^",
	NXP_OP_BINARY,
	30
    },
    {
	NXP_OP_DOT3,
	"...",
	NXP_OP_BINARY,
	10
    },
    { 
	NXP_OP_PLUS2_LEFT,
	"++",
	NXP_OP_UNARY_LEFT,
	40
    },
    { 
	NXP_OP_MINUS2_LEFT,
	"--",
	NXP_OP_UNARY_LEFT,
	40
    },
    { 
	NXP_OP_PLUS2_RIGHT,
	"++",
	NXP_OP_UNARY_RIGHT,
	40
    },
    { 
	NXP_OP_MINUS2_RIGHT,
	"--",
	NXP_OP_UNARY_RIGHT,
	40
    },
    {
	NXP_OP_GT,
	">",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_GT_EQ,
	">=",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_LT,
	"<",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_LT_EQ,
	"<=",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_EQ2,
	"==",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_EXCL_EQ,
	"!=",
	NXP_OP_BINARY,
	60
    },
    {
	NXP_OP_SLASH_BSLASH,
	"/\\",
	NXP_OP_BINARY,
	70
    },
    {
	NXP_OP_BSLASH_SLASH,
	"\\/",
	NXP_OP_BINARY,
	80
    },
    {
	NXP_OP_TILDE_LEFT,
	"~",
	NXP_OP_UNARY_LEFT,
	65
    },
    {
	NXP_OP_EXCL_RIGHT,
	"!",
	NXP_OP_UNARY_RIGHT,
	10
    },
    {
	NXP_OP_EQ,
	"=",
	NXP_OP_BINARY,
	100
    },
    {
	NXP_OP_COLON_EQ,
	":=",
	NXP_OP_BINARY,
	100
    },
    {
	NXP_OP_PERCENT,
	"%",
	NXP_OP_BINARY,
	40
    }
};


/* --- */


/* Maybe in a distant future numexp will allow defining new operators,
   but for now we have a fixed number of operators, though with an
   (almost) unlimited number of types. */
static GHashTable *bindings_table[NXP_NUM_OPS];
/* For each operator we should define a derive function with prototype
   NxpElement* func(NxpOp op, NxpIdentifier *id, NxpElement *left, NxpElement* right, Errorstuff)
 */

static gboolean bindings_initialized = FALSE;

/* The key used in the hash tables is an integer containing the
   combination (bitwise OR) of the two types */
#define MAKE_BINDING_KEY(type_left, type_right) \
	GUINT_TO_POINTER( (((guint32)(type_left)) << 16) | ((guint32)type_right) )

/* error stuff */
static GQuark error_domain = -1;


/*
 * nxp_op_bindings_init: - initialize operator bindings subsystem
 */

void nxp_op_bindings_init(void)
{
    int op;
    g_return_if_fail(!bindings_initialized);
    /* Check the integrity of the 'opinfo' table */
    for (op = 1; op <= NXP_NUM_OPS; op++)
	if (opinfo[op - 1].op != op)
	    g_error("Integrity check at the 'opinfo' table failed!\n"
		    "If you are a developer, check the file kernel/op_bindings.c"
		    " for the 'opinfo' vector. The broken entry is at position"
		    " %i (counting from 1)", op);
    for (op = 1; op <= NXP_NUM_OPS; op++)
	bindings_table[op - 1] = g_hash_table_new_full(NULL, NULL, NULL, g_free);
    error_domain = g_quark_from_static_string("NXP_OP_BINDINGS_ERROR");

    nxp_op_derivation_init();

    bindings_initialized = TRUE;
}

/*
void static _display(gpointer key, gpointer value, gpointer user_data)
{
    guint left = GPOINTER_TO_UINT(key) >> 16;
    guint right = GPOINTER_TO_UINT(key) & 0xffff;
    gchar *namel, *namer;

    if (left == NXP_ELEMENT_TYPE_ANY)
	namel = "any";
    else
	namel = nxp_element_type_get_info(left)->name;

    if (right == NXP_ELEMENT_TYPE_ANY)
	namer = "any";
    else
	namer = nxp_element_type_get_info(right)->name;

    g_print("\tleft : %s right : %s\n", namel, namer);
}
*/

/**
 * nxp_op_bindings_fini: - finalize (shutdown) operator bindings subsystem
 **/
void nxp_op_bindings_fini(void)
{
    int op;
    g_return_if_fail(bindings_initialized);

/*    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	for (op = 1; op <= NXP_NUM_OPS; op++) {
	    g_print("op %s\n", nxp_op_get_info(op)->token);
	    g_hash_table_foreach(bindings_table[op - 1], _display, NULL);
	    g_print("\n");
	}
    }*/

    for (op = 1; op <= NXP_NUM_OPS; op++)
	 g_hash_table_destroy(bindings_table[op - 1]);
    bindings_initialized = FALSE;    
}


/**
 * nxp_bind_operator_function: - bind functions to handle a type/operator/type case
 * @op: 
 * @type_left: 
 * @type_right: 
 * @eval: function used to evaluate the expression
 * @simp: function used to simplify the expression or %NULL
 * @der: function used to derive the expression or %NULL
 * @flags: if NXP_OP_BINDING_SYMMETRIC is set, another binding is created with @type_left and @type_right reversed and the above functions will be called with their parameter elements also reversed
 * @error: may be set to ALREADY_BOUND if the same binding already exists
 * 
 * Adds functions to handle expressions involving operator @op with an
 * element of type @type_left at the left of the operator and an
 * element of type @type_right at the right of the operator. In case
 * of unary operators, one of the element types is ignored considering
 * the side of the operator relative to the operand.
 **/
void nxp_bind_operator_function(NxpOp op,
				NxpElementType type_left,
				NxpElementType type_right,
				NxpOpEvalFunc eval,
				NxpOpSimpFunc simp, 
				NxpOpBindFlags flags,
				GError **error)
{
    NxpOpBinding *binding;
    GSList *children_left, *children_right, *lst1, *lst2;
    NxpElementTypeInfo *typeinfo;
    gchar *str1, *str2;
    gchar *any = "any";

    g_return_if_fail(bindings_initialized);
    g_return_if_fail(op >= 1 && op <= NXP_NUM_OPS);

    /* Check if there is already a binding in the same place */
    if (g_hash_table_lookup(bindings_table[op - 1],
			    MAKE_BINDING_KEY(type_left, type_right))) {
	typeinfo = nxp_element_type_get_info(type_left);
	if (typeinfo)
	    str1 = typeinfo->name;
	else
	    str1 = any;
	typeinfo = nxp_element_type_get_info(type_right);
	if (typeinfo)
	    str2 = typeinfo->name;
	else
	    str2 = any;

	g_print("Binding of %s for %s-%s is replaced by a new one.\n",
			    nxp_op_get_info(op)->token,
			    str1, str2);
	g_hash_table_remove(bindings_table[op - 1],
			    MAKE_BINDING_KEY(type_left, type_right));
    }

    /* If operator is symmetric, check also reversed binding */
    if (flags & NXP_OP_BIND_SYMMETRIC)
	if (g_hash_table_lookup(bindings_table[op - 1],
				MAKE_BINDING_KEY(type_right, type_left)))
	    g_hash_table_remove(bindings_table[op - 1],
			    MAKE_BINDING_KEY(type_right, type_left));

    binding = nxp_op_binding_new();
    binding->eval_func = eval;
    binding->simp_func = simp;
    g_hash_table_insert(bindings_table[op - 1],
			MAKE_BINDING_KEY(type_left, type_right),
			binding);

    if (flags & NXP_OP_BIND_SYMMETRIC) {
	binding = nxp_op_binding_new();
	binding->eval_func = eval;
	binding->simp_func = simp;
	binding->flags = NXP_OP_BINDING_REVERSE;
	g_hash_table_insert(bindings_table[op - 1],
			    MAKE_BINDING_KEY(type_right, type_left),
			    binding);
    }

/*    typeinfo = nxp_element_type_get_info(type_left);
    if (typeinfo->parent) {
	type_parent_left = typeinfo->parent->type;
	nxp_bind_operator_function(op, type_parent_left,
				   type_right, eval, simp, flags, error);
    }
    else
	type_parent_left = -1;
    typeinfo = nxp_element_type_get_info(type_right);
    if (typeinfo->parent) {
	type_parent_right = typeinfo->parent->type;
	nxp_bind_operator_function(op, type_left,
				   type_parent_right, eval, simp, flags, error);
    }
    else
	type_parent_right = -1;
    if (type_parent_left >= 0 && type_parent_right >= 0)
	nxp_bind_operator_function(op, type_parent_left,
				   type_parent_right, eval, simp, flags, error);
*/

    children_left = nxp_element_type_get_children_types(type_left);
    children_right = nxp_element_type_get_children_types(type_right);

    for (lst1 = children_left; lst1; lst1 = lst1->next)
	nxp_bind_operator_function(op, GPOINTER_TO_INT(lst1->data),
				   type_right,
				   eval, simp, flags, error);
    for (lst2 = children_right; lst2; lst2 = lst2->next)
	nxp_bind_operator_function(op, type_left,
				   GPOINTER_TO_INT(lst2->data),
				   eval, simp, flags, error);
    for (lst1 = children_left; lst1; lst1 = lst1->next) {
	for (lst2 = children_right; lst2; lst2 = lst2->next)
	    nxp_bind_operator_function(op, GPOINTER_TO_INT(lst1->data),
				       GPOINTER_TO_INT(lst2->data),
				       eval, simp, flags, error);
    }
    g_slist_free(children_left);
    g_slist_free(children_right);
}

typedef struct _OpData {
    NxpElementType parent;
    NxpElementType child;
    NxpOp op;
    GHashTable *new_bindings;
} OpData;

static void _bind_operator_function(GHashTable *new_bindings,
				    NxpElementType type_left,
				    NxpElementType type_right,
				    NxpOpEvalFunc eval,
				    NxpOpSimpFunc simp, 
				    NxpOpBindFlags flags)
{
    NxpOpBinding *binding;

    binding = nxp_op_binding_new();
    binding->eval_func = eval;
    binding->simp_func = simp;
    binding->flags = flags;
    g_hash_table_insert(new_bindings,
			MAKE_BINDING_KEY(type_left, type_right),
			binding);
}

static void _propagate(gpointer key, gpointer value, gpointer user_data)
{
    OpData *data = (OpData *)user_data;
    guint left = GPOINTER_TO_UINT(key) >> 16;
    guint right = GPOINTER_TO_UINT(key) & 0xffff;
    NxpOpBinding *binding1 = (NxpOpBinding *)value;

    g_return_if_fail(left != data->child && right != data->child);
//    if (left == data->parent || right == data->parent) {
//	g_print("\tleft : %s right : %s\n", nxp_element_type_get_info(left)->name, nxp_element_type_get_info(right)->name);
//    }
    if (left == data->parent) {
	_bind_operator_function(data->new_bindings, data->child, right,
				    binding1->eval_func, binding1->simp_func, binding1->flags);
    }
    if (right == data->parent) {
	_bind_operator_function(data->new_bindings, left, data->child,
				    binding1->eval_func, binding1->simp_func, binding1->flags);
    }
    if (left == data->parent && right == data->parent) {
	_bind_operator_function(data->new_bindings, data->child, data->child,
				    binding1->eval_func, binding1->simp_func, binding1->flags);
    }
}

gboolean _copy(gpointer key, gpointer value, gpointer user_data)
{
    NxpOp op = ((OpData *)user_data)->op;

    if (g_hash_table_lookup(bindings_table[op - 1],
			    key)) {
	g_warning("Binding of %s for key %p is replaced by a new one.\n",
			    nxp_op_get_info(((OpData *)user_data)->op)->token,
			    key);
	g_hash_table_remove(bindings_table[op - 1],
			    key);
    }
    g_hash_table_insert(bindings_table[op - 1],
			key, value);
    return TRUE;
}

void nxp_op_binding_propagate(NxpElementTypeInfo *typeinfo)
{
    NxpOp op;
    OpData user_data;
    
    if (typeinfo->parent) {
	user_data.parent = typeinfo->parent->type;
	user_data.child = typeinfo->type;
	user_data.new_bindings = g_hash_table_new(NULL, NULL);
	for (op = 1; op <= NXP_NUM_OPS; op ++) {
	    user_data.op = op;
//	    g_print ("op : %s\n", nxp_op_get_info(op)->token);
	    g_hash_table_foreach(bindings_table[op - 1], _propagate, &user_data);
	    g_hash_table_foreach_remove(user_data.new_bindings, _copy, &user_data);
	}
	g_hash_table_destroy(user_data.new_bindings);
    }
}

/**
 * nxp_get_operator_binding:
 * @op: operator
 * @type_left: type at left of operator (may be ignored for unary operators)
 * @type_right: type at right of operator (may be ignored for unary operators)
 * @error: 
 * 
 * Get the binding matching the case @type_left/@op/@type_right.
 * 
 * Return value: The binding structure (an existing reference) or NULL if not found.
 **/
NxpOpBinding* nxp_get_operator_binding(NxpOp op, NxpElementType type_left, 
				       NxpElementType type_right, GError **error)
{
    NxpOpBinding *binding;
    g_return_val_if_fail(bindings_initialized, NULL);
    g_return_val_if_fail(op >= 1 && op <= NXP_NUM_OPS, NULL);
    binding = g_hash_table_lookup(bindings_table[op - 1],
				  MAKE_BINDING_KEY(type_left, type_right));
    return binding;
}

/*NxpOpBinding* nxp_get_higher_operator_binding(NxpOp op, NxpElement *left,
					      NxpElement *right, GError **error)
{
    NxpElementTypeInfo *info_left, *info_right;
    NxpOpBinding *binding = NULL;

    g_return_val_if_fail(bindings_initialized, NULL);
    g_return_val_if_fail(op >= 1 && op <= NXP_NUM_OPS, NULL);

    for (info_left = left->type_info; info_left; info_left = info_left->parent) {
	for (info_right = right->type_info; info_right; info_right = info_right->parent) {
	    binding = g_hash_table_lookup(bindings_table[op - 1],
			MAKE_BINDING_KEY(GPOINTER_TO_INT(info_left->type),
					 GPOINTER_TO_INT(info_right->type)));
	    if (binding)
		return binding;
	}
    }
    return binding;
}
*/
/* --- */


/**
 * nxp_op_binding_new:
 * @void: 
 * 
 * 
 * 
 * Return value: 
 **/
NxpOpBinding* nxp_op_binding_new(void)
{
    NxpOpBinding *binding;
    binding = g_new0(NxpOpBinding, 1);
    binding->ref_count = 1;
    return binding;
}

/**
 * nxp_op_binding_ref:
 * @self: 
 * 
 * 
 **/
void nxp_op_binding_ref(NxpOpBinding *self)
{
    g_return_if_fail(self);
    ++self->ref_count;
}

/**
 * nxp_op_binding_unref:
 * @self: 
 * 
 * 
 **/
void nxp_op_binding_unref(NxpOpBinding *self)
{
    g_return_if_fail(self);
    if (--self->ref_count == 0)
	g_free(self);
}


/**
 * nxp_op_get_info:
 * @op: an operator enum value
 * 
 * Get a pointer to the #NxpOpInfo structure regarding operator @op.
 * 
 * Return value: pointer to a #NxpOpInfo
 **/
const NxpOpInfo* nxp_op_get_info(NxpOp op)
{
    g_assert (op > 0 && op <= NXP_NUM_OPS);
    return opinfo + op - 1;
}
