/* -*- c -*- */

#include "function.h"
#include "identifier.h"
#include "namespace.h"
#include "int.h"
#include "tree.h"
#include "eval_context.h"
#include "bifuncs.h"
#include "message.h"
#include "errors.h"
#include "element_output.h"
#include "prototypes.h"
#include "nxp_debug.h"
#include "op_bindings.h"
#include "i18n.h"
#include "void.h"

static NxpElementTypeInfo *typeinfo = NULL;

/**
 * element_function_new:
 * @ident: identifier of the function
 * @args: the linked list of arguments; note: will take ownership of this list
 * 
 * Create a new NxpElementFunction element (function call)
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_function_new(NxpIdentifier *ident, 
				     NxpElement *args, GHashTable *named_args)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementFunction));
    NXP_ELEMENT_VARIABLE(elem)->ident = nxp_identifier_ref(ident);
    NXP_ELEMENT_FUNCTION(elem)->simple_args = NXP_ELEMENT_REF(args);
    NXP_ELEMENT_FUNCTION(elem)->named_args = named_args;
    return elem;
}

NxpElement* nxp_element_function_new_simple(char *name, NxpElement *args)
{
    NxpIdentifier *ident;
    NxpElement *el;

    ident = nxp_identifier_new_simple(name);
    el = nxp_element_function_new(ident, args, NULL);
    nxp_identifier_unref(ident);
    return el;
}

gboolean nxp_element_function_is(NxpElement *el, char *name, NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_FUNCTION(el), FALSE);
    char *str;

    str = nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(el)->ident);
    if (g_str_equal(str, name)) {
	g_free(str);
	return TRUE;
    }
    else {
	g_free(str);
	return FALSE;
    }
}

static NxpElement* eval_func(NxpElement *elem,
			     NxpEvalContext *context, GError **error)
{
    GHashTable *hash;
    NxpIdentifier *ident;
    NxpElementFunction *func;
    NxpElement *values, *evaled_values = NULL;
    NxpElement *result = NULL;
    int arity;
    NxpDefinition *def;
    NxpNameSpace *ns;
    const char *name;
    char *new_name;

    g_return_val_if_fail(elem, NULL);

    func   = NXP_ELEMENT_FUNCTION(elem);
    ident  = NXP_ELEMENT_VARIABLE(func)->ident;
    values = func->simple_args;
    hash = func->named_args;

    arity = nxp_element_list_length(values);

    /* First, look for a built-in function name */
    nxp_identifier_resolve(ident, context, &ns, &name);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    if (def) {
	if (def->type == NXP_DEF_BI_FUNC)
	{
	    NxpProtoInstance *proto;
	    GError *err = NULL;
	    NxpElement *new_values;

	    proto = nxp_find_eval_prototype(values, &new_values,
				    NXP_DEF_BI_FUNC(def)->prototypes,
				    context, &err);

	    if (err) {
		g_propagate_error(error, err);
		result = NULL;
	    }
	    else if (proto && proto->eval_func) {
		if (proto->flags & NXP_PROTO_INSTANCE_DETAILED) {
		    NxpElement* (*func) (NxpDefinition *def, GSList *arg_list,
					 NxpEvalContext *context, GError **error);
		    func = proto->eval_func;
		    result = func(def, NXP_ELEMENT_LIST(new_values)->list,
				  context, error);
		} else {
		    g_assert(proto->eval_func && proto->protoclass->eval_marshal);
		    result = proto->protoclass->eval_marshal(
					proto->eval_func,
					NXP_ELEMENT_LIST(new_values)->list,
					hash,
					proto->user_data,
					context,
					&err);
		    if (err)
			g_propagate_error(error, err);
		}
		NXP_ELEMENT_UNREF(new_values);
	    } else  {
		char *ident_str = nxp_identifier_to_string(ident);
		g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			    _("No suitable prototype found in function '%s'"), ident_str);
		g_free(ident_str);
		result = NULL;
	    }
	} else if (def->type == NXP_DEF_VARIABLE) {
	    char *ident_str = nxp_identifier_to_string(ident);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
			_("Cannot call variable '%s' as a function"), ident_str);
	    g_free(ident_str);
	    result = NULL;
	}
    }
    else {
	/* Is it a user defined function ? We test name/arity */
	new_name = g_strdup_printf("%s/%d", name, arity);
	def = nxp_name_space_lookup_definition_r(ns, new_name, TRUE);
	g_free(new_name);
	if (!def) {
	    /* We test name/-1 */
	    new_name = g_strconcat(name, "/-1", NULL);
	    def = nxp_name_space_lookup_definition_r(ns, new_name, TRUE);
	    g_free(new_name);
	}
	if (def) {
	    if (def->type == NXP_DEF_USER_FUNC) {
		NxpElement* exp;
		GSList* variables, *v;
		GError *myerr = NULL;
		exp = NXP_ELEMENT_REF(NXP_DEF_USER_FUNC(def)->expression);
		variables = NXP_DEF_USER_FUNC(def)->argument_list;

		evaled_values = nxp_element_eval(values, context, &myerr);
		if (myerr) {
		    g_propagate_error(error, myerr);
		    return NULL;
		}
		v = NXP_ELEMENT_LIST(evaled_values)->list;
		/* define the function arguments in a temporary namespace */
		nxp_eval_context_push_name_space(context);
		while (variables)
		{
		    const char *var;
		    NxpElement *el;
		    GError *myerr = NULL;
		    var = variables->data;
		    el = NXP_ELEMENT(v->data);
		    g_assert(var);
		    nxp_user_variable_define_simple(var, el, context, &myerr);
		    if (myerr) {
			NXP_ELEMENT_XUNREF(el);
			g_propagate_error(error, myerr);
			return NULL;
		    }

		    variables = g_slist_next(variables);
		    v = g_slist_next(v);
		}
		result = nxp_element_eval(exp, context, &myerr);
		if (myerr) {
		    NXP_ELEMENT_XUNREF(result); result = NULL;
		    g_propagate_error(error, myerr);
		}
		nxp_eval_context_pop_name_space(context);
		NXP_ELEMENT_UNREF(evaled_values);
		NXP_ELEMENT_UNREF(exp);
	    } else if (def->type == NXP_DEF_VARIABLE) {
		char *ident_str = nxp_identifier_to_string(ident);
		g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
			    _("Cannot call variable '%s' as a function"), ident_str);
		g_free(ident_str);
		result = NULL;
	    }
	}
	else if (!def) {
	    char *ident_str = nxp_identifier_to_string(ident);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			_("Undefined Function: %s, or function has a different arity"), ident_str);
	    g_free(ident_str);
	    return NULL;
	}
    }
    nxp_definition_unref(def);
    return result;
}



static NxpElement* nxp_function_derive(NxpElement *elem, NxpIdentifier *var,
			       NxpEvalContext *context, GError **error)
{
    GHashTable *hash;
    NxpElementFunction *func;
    NxpElement *result = NULL;
    int arity;
    NxpDefinition *def;
    NxpIdentifier *ident;
    NxpElement *values;
    NxpNameSpace *ns;
    const char *name;
    NxpIdentifier *fullname;

    g_return_val_if_fail(NXP_ELEMENT_IS_FUNCTION(elem), NULL);
    func = NXP_ELEMENT_FUNCTION(elem);
    ident = NXP_ELEMENT_VARIABLE(func)->ident;
    values = func->simple_args;
    hash   = func->named_args;

    g_return_val_if_fail(ident, NULL);
    g_return_val_if_fail(var, NULL);

    arity = nxp_element_list_length(values);

    nxp_identifier_resolve(ident, context, &ns, &name);
    g_return_val_if_fail(ns, NULL);
    g_return_val_if_fail(name, NULL);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);

    if (!def) {
	fullname = nxp_identifier_create_function_identifier(ident, arity);
	def = nxp_definition_get(fullname, context);
	nxp_identifier_unref(fullname);
    }

    if (def->type == NXP_DEF_BI_FUNC)
    {
	NxpProtoInstance* proto;
	GError *err = NULL;
	NxpElement *new_values;
	proto = nxp_find_simp_prototype(values, &new_values,
			    NXP_DEF_BI_FUNC(def)->prototypes, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    result = NULL;
	}
	else if (!proto) {
	    char *ident_str = nxp_identifier_to_string(ident);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			_("No suitable prototype found in function '%s'"), ident_str);
	    g_free(ident_str);
	    result = NULL;
	    
	}
	else if (proto && proto->derive_func && proto->protoclass && 
	    proto->protoclass->derive_marshal)
	    result = proto->protoclass->derive_marshal(proto->derive_func, 
						       var, 
						       NXP_ELEMENT_LIST(new_values)->list,
						       hash,
						       proto->user_data,
						       context,
						       error);
	else {
	    char *ident_str = nxp_identifier_to_string(ident);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			_("No implementation of '%s' function derivative"), ident_str);
	    g_free(ident_str);
	    result = NULL;
	}
	NXP_ELEMENT_UNREF(new_values);
    }
    else if (def->type == NXP_DEF_USER_FUNC) {
	NxpElement *exp;
	GError *myerr = NULL;

	exp = nxp_element_simplify(elem, context, &myerr);
	if (myerr) {
	    g_propagate_error(error, myerr);
	    nxp_definition_unref(def);
	    return NULL;
	}
	result = nxp_element_derive(exp, var, context, error);
	NXP_ELEMENT_UNREF(exp);
    } else {
	g_warning("Some strange error -- probably a bug!");
	result = NULL;
    }
    nxp_definition_unref(def);
    return result;
}


static NxpElement* eval_store(NxpOp op, NxpElement *a, NxpElement *b,
			      NxpEvalContext *context, GError **error)
{
    NxpIdentifier *fullname;
    const char *name;
    NxpNameSpace *ns;
    int size = nxp_element_list_length(NXP_ELEMENT_FUNCTION(a)->simple_args);
    NxpDefinition *def;

    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(a)->ident, context, &ns, &name);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    if (def && def->type == NXP_DEF_BI_FUNC) {
	if (g_str_equal(name, "getFromMatrix")) {
	    NxpElement *args, *f, *retval;
	    GSList *l;

	    args = nxp_element_list_new();
	    for (l = NXP_ELEMENT_LIST(NXP_ELEMENT_FUNCTION(a)->simple_args)->list; l; l = l->next)
		nxp_element_list_append(args, l->data);

	    nxp_element_list_append(args, b);
	    f = nxp_element_function_new_simple("replaceOnMatrix", args);
	    retval = nxp_element_eval(f, context, error);
	    NXP_ELEMENT_UNREF(args);
	    NXP_ELEMENT_UNREF(f);
	    return retval;
	}
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp built-in function"));
	    nxp_definition_unref(def);
	    return NULL;
	}
    }

    fullname = nxp_identifier_create_function_identifier(
					NXP_ELEMENT_VARIABLE(a)->ident, size);
    nxp_user_function_define(fullname,
			     NXP_ELEMENT_FUNCTION(a)->simple_args,
			     b, context, error);
    nxp_identifier_unref(fullname);
    return nxp_element_void_new();
}

static NxpElement* simp_assign(NxpOp op, NxpElement *a, NxpElement *b,
			      NxpEvalContext *context, GError **error)
{
    NxpIdentifier *fullname;
    NxpElement *value;
    const char *name;
    NxpNameSpace *ns;
    int size = nxp_element_list_length(NXP_ELEMENT_FUNCTION(a)->simple_args);
    NxpDefinition *def;
    GError *err = NULL;

    value = nxp_element_simplify(b, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(a)->ident, context, &ns, &name);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    if (def && def->type == NXP_DEF_BI_FUNC) {
	if (g_str_equal(name, "getFromVector")) {
	    NxpElement *args, *f, *retval;
	    GSList *l;

	    args = nxp_element_list_new();
	    for (l = NXP_ELEMENT_LIST(NXP_ELEMENT_FUNCTION(a)->simple_args)->list; l; l = l->next)
		nxp_element_list_append(args, l->data);

	    nxp_element_list_append(args, value);
	    NXP_ELEMENT_UNREF(value);
	    f = nxp_element_function_new_simple("replaceOnVector", args);
	    retval = nxp_element_simplify(f, context, error);
	    NXP_ELEMENT_UNREF(f);
	    return retval;
	}
	else if (g_str_equal(name, "getFromMatrix")) {
	    NxpElement *args, *f, *retval;
	    GSList *l;

	    args = nxp_element_list_new();
	    for (l = NXP_ELEMENT_LIST(NXP_ELEMENT_FUNCTION(a)->simple_args)->list; l; l = l->next)
		nxp_element_list_append(args, l->data);

	    nxp_element_list_append(args, value);
	    NXP_ELEMENT_UNREF(value);
	    f = nxp_element_function_new_simple("replaceOnMatrix", args);
	    retval = nxp_element_eval(f, context, error);
	    NXP_ELEMENT_UNREF(f);
	    return retval;
	}
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp built-in function"));
	    nxp_definition_unref(def);
	    return NULL;
	}
    }

    fullname = nxp_identifier_create_function_identifier(
					NXP_ELEMENT_VARIABLE(a)->ident, size);
    nxp_user_function_define(fullname,
			     NXP_ELEMENT_FUNCTION(a)->simple_args,
			     value, context, error);
    NXP_ELEMENT_UNREF(value);
    nxp_identifier_unref(fullname);
    return nxp_element_void_new();
}

static NxpElement* eval_assign(NxpOp op, NxpElement *a, NxpElement *b,
			      NxpEvalContext *context, GError **error)
{
    NxpIdentifier *fullname;
    NxpElement *value;
    const char *name;
    NxpNameSpace *ns;
    int size = nxp_element_list_length(NXP_ELEMENT_FUNCTION(a)->simple_args);
    NxpDefinition *def;
    GError *err = NULL;

    value = nxp_element_eval(b, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(a)->ident, context, &ns, &name);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    if (def && def->type == NXP_DEF_BI_FUNC) {
	if (g_str_equal(name, "getFromMatrix")) {
	    NxpElement *args, *f, *retval;
	    GSList *l;

	    args = nxp_element_list_new();
	    for (l = NXP_ELEMENT_LIST(NXP_ELEMENT_FUNCTION(a)->simple_args)->list; l; l = l->next)
		nxp_element_list_append(args, l->data);

	    nxp_element_list_append(args, value);
	    f = nxp_element_function_new_simple("replaceOnMatrix", args);
	    NXP_ELEMENT_UNREF(value);
	    NXP_ELEMENT_UNREF(args);
	    retval = nxp_element_eval(f, context, error);
	    NXP_ELEMENT_UNREF(f);
	    return retval;
	}
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp built-in function"));
	    nxp_definition_unref(def);
	    return NULL;
	}
    }

    fullname = nxp_identifier_create_function_identifier(
					NXP_ELEMENT_VARIABLE(a)->ident, size);
    nxp_user_function_define(fullname,
			     NXP_ELEMENT_FUNCTION(a)->simple_args,
			     value, context, error);
    NXP_ELEMENT_UNREF(value);
    nxp_identifier_unref(fullname);
    return nxp_element_void_new();
}

static NxpElement* simplify_func(NxpElement *elem,
				 NxpEvalContext *context, GError **error)
{
    GHashTable *hash;
    NxpIdentifier *ident;
    NxpElementFunction *func;
    NxpElement *values, *evaled_values = NULL;
    NxpElement *result = NULL;
    int arity;
    NxpDefinition *def;
    NxpNameSpace *ns;
    const char *name;
    char *new_name;
    GError *err = NULL;

    g_return_val_if_fail(elem, NULL);

    func   = NXP_ELEMENT_FUNCTION(elem);
    ident  = NXP_ELEMENT_VARIABLE(func)->ident;
    values = func->simple_args;
    hash = func->named_args;

    arity = nxp_element_list_length(values);

    /* First, look for a built-in function name */
    nxp_identifier_resolve(ident, context, &ns, &name);
    def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    if (def) {
	if (def->type == NXP_DEF_BI_FUNC)
	{
	    NxpProtoInstance *proto;
	    NxpElement *new_values;
	    GError *err = NULL;

	    proto = nxp_find_simp_prototype(values, &new_values,
				       NXP_DEF_BI_FUNC(def)->prototypes, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		result = NULL;
	    }
	    /* we need to check if there is a simplify_func because it can be
	       a prototype only for simplification, by example... */
	    else if (proto && proto->simplify_func) {
		if (proto->flags & NXP_PROTO_INSTANCE_DETAILED) {
		    NxpElement* (*func) (NxpDefinition *def, GSList *arg_list,
					 NxpEvalContext *context, GError **error);
		    func = proto->simplify_func;
		    result = func(def, NXP_ELEMENT_LIST(new_values)->list,
				  context, error);
		} else {
		    g_assert(proto->simplify_func && proto->protoclass->simplify_marshal);
		    result = proto->protoclass->simplify_marshal(
					proto->simplify_func,
					NXP_ELEMENT_LIST(new_values)->list,
					hash,
					proto->user_data,
					context,
					error);
		}
		NXP_ELEMENT_UNREF(new_values);
	    } else  {
		result = nxp_element_function_new(ident, values, hash ? nxp_named_args_ref(hash):NULL);
	    }
	} else if (def->type == NXP_DEF_VARIABLE) {
	    char *ident_str = nxp_identifier_to_string(ident);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
			_("Cannot call variable '%s' as a function"), ident_str);
	    g_free(ident_str);
	    result = NULL;
	}
    }
    else {
	/* Is it a user defined function ? We test name/arity */
	new_name = g_strdup_printf("%s/%d", name, arity);
	def = nxp_name_space_lookup_definition_r(ns, new_name, TRUE);
	g_free(new_name);
	if (!def) {
	    /* We test name/-1 */
	    new_name = g_strconcat(name, "/-1", NULL);
	    def = nxp_name_space_lookup_definition_r(ns, new_name, TRUE);
	    g_free(new_name);
	}
	if (def) {
	    if (def->type == NXP_DEF_USER_FUNC) {
		NxpElement* exp;
		GSList* variables, *v;
		GError *myerr = NULL;
		exp = NXP_ELEMENT_REF(NXP_DEF_USER_FUNC(def)->expression);
		variables = NXP_DEF_USER_FUNC(def)->argument_list;

		evaled_values = nxp_element_simplify(values, context, &myerr);
		if (myerr) {
		    g_propagate_error(error, myerr);
		    return NULL;
		}
		/* define the function arguments in a temporary namespace */
		nxp_eval_context_push_name_space(context);
		v = NXP_ELEMENT_LIST(evaled_values)->list;

		v = NXP_ELEMENT_LIST(evaled_values)->list;
		variables = NXP_DEF_USER_FUNC(def)->argument_list;

		while (variables)
		{
		    const char *var;
		    NxpElement *vv, *el, *tmp;
		    GError *myerr = NULL;
		    var = variables->data;
		    vv = nxp_element_variable_new_simple(var);
		    el = NXP_ELEMENT(v->data);
		    tmp = nxp_element_substitute(exp, vv, el, context, &myerr);
		    if (myerr) {
			NXP_ELEMENT_XUNREF(el);
			NXP_ELEMENT_XUNREF(tmp);
			nxp_eval_context_pop_name_space(context);
			g_propagate_error(error, myerr);
			return NULL;
		    }
		    NXP_ELEMENT_UNREF(exp);
		    exp = tmp;

		    variables = g_slist_next(variables);
		    v = g_slist_next(v);
		}
		result = nxp_element_simplify(exp, context, &myerr);
		if (myerr) {
		    NXP_ELEMENT_UNREF(exp);
		    g_propagate_error(error, myerr);
		    nxp_eval_context_pop_name_space(context);
		    return NULL;
		}

		nxp_eval_context_pop_name_space(context);
		NXP_ELEMENT_UNREF(evaled_values);
		NXP_ELEMENT_UNREF(exp);
	    } else if (def->type == NXP_DEF_VARIABLE) {
		char *ident_str = nxp_identifier_to_string(ident);
		g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
			    _("Cannot call variable '%s' as a function"), ident_str);
		g_free(ident_str);
		result = NULL;
	    }
	}
	else if (!def) {
	    evaled_values = nxp_element_simplify(values, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		return NULL;
	    }
	    result = nxp_element_function_new(ident, evaled_values, hash ? nxp_named_args_ref(hash):NULL);
	    NXP_ELEMENT_UNREF(evaled_values);
	    return result;
	}
    }
    nxp_definition_unref(def);
    return result;
}

static GString *str_or_repr_func(NxpElement *elem, int *prec, int str_or_repr)
{
    GString *str, *str1;
    NxpElementFunction *func = NXP_ELEMENT_FUNCTION(elem);
    
    str = nxp_identifier_to_gstring(NXP_ELEMENT_VARIABLE(func)->ident);
    if (str_or_repr == 1)
	str1 = nxp_element_str(NXP_ELEMENT(func->simple_args));
    else
	str1 = nxp_element_repr(NXP_ELEMENT(func->simple_args));
    str = g_string_append(str, str1->str);
    g_string_free(str1, TRUE);
    if (prec) *prec = 0;
    return str;
}


static GString *str_func(NxpElement *elem, int *prec)
{
    return str_or_repr_func(elem, prec, 1);
}


static GString *repr_func(NxpElement *elem, int *prec)
{
    return str_or_repr_func(elem, prec, 2);
}

static NxpElement* simp_eq2(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error) {
    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(a)->ident,
			     NXP_ELEMENT_VARIABLE(b)->ident, context)
	    && nxp_element_eq(NXP_ELEMENT(NXP_ELEMENT_FUNCTION(a)->simple_args),
			      NXP_ELEMENT(NXP_ELEMENT_FUNCTION(b)->simple_args),
			      context, error))
	return nxp_element_int_new(1);
    else
	return nxp_element_int_new(0);
}

static void destroy_func(NxpElement *elem)
{
    NxpElementFunction *func = NXP_ELEMENT_FUNCTION(elem);

    NXP_ELEMENT_UNREF(func->simple_args);
    if (func->named_args)
	g_hash_table_destroy(func->named_args);
    NXP_ELEMENT(elem)->type_info->parent->destroy_func(elem);
}

static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement* el;
    gboolean result;

    el = nxp_element_simplify(elem, context, error);
    if (el != elem && !NXP_ELEMENT_IS_FUNCTION(el)) {
	g_assert(*error == NULL);
	result = nxp_element_is_true(el, context, error);
	NXP_ELEMENT_UNREF(el);
	return result;
    }
    else {
	NXP_ELEMENT_UNREF(el);
	return TRUE;
    }
}

static gboolean dependson_func(NxpElement *self, NxpElement *variable,
			       NxpEvalContext *context, GError **error)
{
    gboolean retval;

    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(self)->ident,
			     NXP_ELEMENT_VARIABLE(variable)->ident, context))
	retval = TRUE;
    else {
	retval = nxp_element_depends_on(NXP_ELEMENT_FUNCTION(self)->simple_args, variable, context, error);
	/* If the function is defined by the user, maybe the expression itself
	 * depends of variable... The following code should be improved... */
	if (retval == FALSE) {
	    NxpElement *el = nxp_element_simplify(self, context, error);
	    if (!NXP_ELEMENT_IS_FUNCTION(el)) {
		retval = nxp_element_depends_on(el, variable, context, error);
	    }
	    else
		retval = FALSE;
	    NXP_ELEMENT_UNREF(el);
	}
    }
    return retval;
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var,
				   NxpElement *el, NxpEvalContext *context,
				   GError **error)
{
    NxpElement *args, *retval;
    GHashTable *hash = NXP_ELEMENT_FUNCTION(self)->named_args;

    args = nxp_element_substitute(NXP_ELEMENT_FUNCTION(self)->simple_args, var, el, context, error);
    retval = nxp_element_function_new(NXP_ELEMENT_VARIABLE(self)->ident, args,
				      hash ? nxp_named_args_ref(hash):NULL);
    NXP_ELEMENT_UNREF(args);
    return retval;
}

typedef struct {
    NxpElementVisitorFunc func;
    gpointer user_data;
    NxpEvalContext *context;
} iterator_func_data_t;


static void
_iterator_func_helper(gpointer key,
                      gpointer value,
                      gpointer user_data)
{
    iterator_func_data_t *data = user_data;
    NxpElement *arg = NXP_ELEMENT(value);

    data->func(arg, data->user_data);
    if (arg->type_info->iterator_func)
        arg->type_info->iterator_func(arg, data->func, data->user_data, data->context);
}

static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc cfunc,
              gpointer user_data,
              NxpEvalContext *context)
{
    NxpElementFunction *func = NXP_ELEMENT_FUNCTION(elem);
    NxpElement *args = NXP_ELEMENT(func->simple_args);
  
    if (args->type_info->iterator_func)
	args->type_info->iterator_func(args, cfunc, user_data, context);
//    nxp_element_foreach(func->simple_args, cfunc, user_data, context);

    if (func->named_args) {
        iterator_func_data_t data = {cfunc, user_data, context};
        g_hash_table_foreach(func->named_args,
                             _iterator_func_helper, &data);
    }
}

int nxp_element_function_get_type(void)
{
    static int nxp_element_type_function = 0;
    if (nxp_element_type_function == 0) {
	nxp_element_type_function = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_function, /* The new type */
					    sizeof(NxpElementTypeInfo), /* The size of our typeinfo */
					    NXP_ELEMENT_TYPE_VARIABLE, /* the type of the parent */
					    sizeof(NxpElementTypeInfo)); /* the size of the parent's typeinfo */
    }
    return nxp_element_type_function;
}

#define BIND(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_FUNCTION, \
							 NXP_ELEMENT_TYPE_FUNCTION, \
    			                                 eval, symp, 0, NULL);

void nxp_element_function_type_init(void)
{
    typeinfo = nxp_element_type_get_info(nxp_element_function_get_type());
    typeinfo->name = g_strdup("function");
    typeinfo->destroy_func = destroy_func;
    typeinfo->eval_func = eval_func;
    typeinfo->simplify_func = simplify_func;
    typeinfo->derive_func = nxp_function_derive;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = truth_func;
    typeinfo->iterator_func = iterator_func;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;

    BIND(NULL, simp_eq2, NXP_OP_EQ2);
    nxp_bind_operator_function(NXP_OP_COLON_EQ, NXP_ELEMENT_TYPE_FUNCTION,
			       NXP_ELEMENT_TYPE_ANY,
			       eval_store, eval_store,
			       0, NULL);
    nxp_bind_operator_function(NXP_OP_EQ, NXP_ELEMENT_TYPE_FUNCTION,
			       NXP_ELEMENT_TYPE_ANY,
			       eval_assign, simp_assign,
			       0, NULL);
}

static void ghref_elems(gpointer key, gpointer value, gpointer user_data) {
    char *char_key = (char*)key;
    NxpElement *el = NXP_ELEMENT(value);
    GHashTable *hash = (GHashTable*)user_data;
    g_hash_table_insert(hash, g_strdup(char_key), NXP_ELEMENT_REF(el));
}

GHashTable *nxp_named_args_ref(GHashTable *named_args) {
    GHashTable *hash = g_hash_table_new_full(g_str_hash, g_str_equal,
					     g_free, nxp_element_data_unref);

    g_hash_table_foreach(named_args, ghref_elems, hash);

    return hash;
}
