#include <stdio.h>
#include "matrix.h"

#include "int.h"
#include "real.h"
#include "tree.h"
#include "vector.h"

#include "message.h"
#include "types_misc.h"
#include "errors.h"
#include "element_output.h"
#include "void.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

#define unimplemented()\
 eval_warning(_("Function %s unimplemented in NumExp. Mail the developers!"), __FUNCTION__);\
 return NULL;



/**
 * nxp_element_matrix_new: -create matrix element
 * @width: the number of columns in the matrix
 * @height: the number of rows in the matrix
 * @data: an array of pointers to each element of the matrix; this block of memory should have a size of at least (and no more than) @width*@height*sizeof(&NxpElement*); the element pointers are stored as a sequence of rows, top->bottom left->right (like pixels in video memory); if %NULL, a new array of pointers (initialized to %NULL) will be created to hold @width*@height elements; Note: if @data is non null, the matrix element will take ownership (keep) of this array.
 * 
 * Create a new matrix element.
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_matrix_new(guint width, guint height, NxpElement **data)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementMatrix)); /* TODO */
    NXP_ELEMENT_MATRIX(elem)->width = width;
    NXP_ELEMENT_MATRIX(elem)->height = height;
    if (data)
	NXP_ELEMENT_MATRIX(elem)->data = data;
    else
	NXP_ELEMENT_MATRIX(elem)->data = (width > 0 && height > 0)?
	    g_new0(NxpElement *, width*height) : NULL;
    return elem;
}



static void nxp_destroy_matrix(NxpElement *elem)
{
    NxpElementMatrix *mat = NXP_ELEMENT_MATRIX(elem);
    int i, len;
    len = mat->width * mat->height;
    for (i = 0; i < len; ++i)
	NXP_ELEMENT_XUNREF(mat->data[i]);
    g_free(mat->data);
}


/**
 * _element_matrix_set:
 * @matrix: the matrix element
 * @row: row number (starts at 0)
 * @col: column number (starts at 0)
 * @element: value
 * 
 * Set a specific element of the matrix. The matrix should not already
 * have an element at that position.
 **/
void _element_matrix_set(NxpElementMatrix *matrix, guint row, guint col, NxpElement *element)
{
    /* make sure we have a matrix */
    g_return_if_fail(matrix);
    /* make sure we have an element */
    g_return_if_fail(matrix);
    /* make sure we are not exceeding dimmensions */
    g_return_if_fail(row < matrix->height && col < matrix->width);
    /* make sure the matrix doesn't already have an element there */
    g_return_if_fail(!_element_matrix_index(matrix, row, col));
    _element_matrix_index(matrix, row, col) = NXP_ELEMENT_REF(element);
}

/**
 * _element_matrix_set_row:
 * @matrix: the matrix element
 * @row: row number (starts at 0)
 * @element: the vector to assign to the row
 * 
 * Set a specific row of the matrix. The matrix should not already
 * have an element at that row.
 **/
void _element_matrix_set_row(NxpElementMatrix *matrix, guint row, NxpElement *element)
{
    int i;
    NxpElementVector *v = NXP_ELEMENT_VECTOR(element);

    /* make sure we have a matrix */
    g_return_if_fail(matrix);
    /* make sure we have an element */
    g_return_if_fail(NXP_ELEMENT_IS_MATRIX(NXP_ELEMENT(matrix)));
    /* make sure we are not exceeding dimmensions */
    g_return_if_fail(row < matrix->height && v->size <= matrix->width);

    for (i = 0; i < v->size; i++)
	_element_matrix_index(matrix, row, i) = _element_vector_get(v, i);
    for ( ; i < matrix->width; i++)
	_element_matrix_index(matrix, row, i) = nxp_element_int_new(0);
}

/**
 * _element_matrix_set_col:
 * @matrix: the matrix element
 * @row: row number (starts at 0)
 * @element: the vector to assign to the row
 * 
 * Set a specific column of the matrix. The matrix should not already
 * have an element at that column.
 **/
void _element_matrix_set_col(NxpElementMatrix *matrix, guint col, NxpElement *element)
{
    int i;
    NxpElementVector *v = NXP_ELEMENT_VECTOR(element);

    /* make sure we have a matrix */
    g_return_if_fail(matrix);
    /* make sure we have an element */
    g_return_if_fail(NXP_ELEMENT_IS_MATRIX(NXP_ELEMENT(matrix)));
    /* make sure we are not exceeding dimmensions */
    g_return_if_fail(col < matrix->width && v->size <= matrix->height);

    for (i = 0; i < v->size; i++)
	_element_matrix_index(matrix, i, col) = _element_vector_get(v, i);
    for ( ; i < matrix->width; i++)
	_element_matrix_index(matrix, i, col) = nxp_element_int_new(0);
}

/**
 * _element_matrix_get:
 * @matrix: the matrix element
 * @row: row number (starts at 0)
 * @col: column number (starts at 0)
 * 
 * Get a specific element of the matrix. 
 *
 * Return value: a new reference to the element at the specified position of the matrix if it has one, or the void element if there is no element in that position.
 **/
NxpElement* _element_matrix_get(NxpElementMatrix *matrix, guint row, guint col)
{
    NxpElement *el;
    /* make sure we have a matrix */
    g_return_val_if_fail(matrix, NULL);
    /* make sure we are not exceeding dimmensions */
    g_return_val_if_fail(row < matrix->height && col < matrix->width, NULL);
    el = _element_matrix_index(matrix, row, col);
    return el? NXP_ELEMENT_REF(el) : nxp_element_void_new();
}



/**
 * matrix_real_put:
 * @matrix: 
 * @element: 
 * @row: 
 * @col: 
 * 
 * 
 * 
 * Return value: 
 **/
NxpElement* nxp_element_matrix_real_put(NxpElementMatrix *matrix, NxpElement *element,
					guint row, guint col)
{
    NxpElementMatrix *m;
    NxpElement *el;
    int i, j, width, height;

    g_return_val_if_fail(NXP_ELEMENT_MATRIX(matrix), NULL);
    g_return_val_if_fail(NXP_ELEMENT(element), NULL);

    width  = MAX(matrix->width, col + 1);
    height = MAX(matrix->height, row + 1);

    m = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(width, height, NULL));
    for (i = 0; i < height; i++)
	for (j = 0; j < width; j++) {
	    if (i == row && j == col)
		el = NXP_ELEMENT_REF(element);
	    else {
		if (i < matrix->width && j < matrix->height) {
		    el = _element_matrix_get(matrix, i, j);
		} else {
		    el = NULL;
		}
	    }
	    _element_matrix_index(m, i, j) = el;
	}
    return NXP_ELEMENT(m);
}





/**
 * matrix_times_scalar:
 * @mat: 
 * @scalar: 
 * 
 * 
 * 
 * Return value: 
 **/
NxpElement *matrix_times_scalar(NxpElementMatrix *mat, NxpElement *scalar,
				NxpEvalContext *context, GError **error)
{
    NxpElementMatrix *result;
    NxpElement *el, *tree;
    gint i, j;

    g_return_val_if_fail(NXP_ELEMENT_MATRIX(mat), NULL);
    g_return_val_if_fail(scalar, NULL);
/*      g_return_val_if_fail(nxp_element_is_scalar(scalar), NULL); */
    
    result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(mat->width, mat->height, NULL));
    for (i = 0; i < mat->height; i++)
	for (j = 0; j < mat->width; j++)
	{
	    tree = elem_TIMES(_element_matrix_get(mat, i, j),
			      NXP_ELEMENT_REF(scalar));
	    el = nxp_element_simplify(tree, context, NULL);
	    NXP_ELEMENT_UNREF(tree);
	    _element_matrix_index(result, i, j) = el;
	}
    return NXP_ELEMENT(result);
}



/**
 * vector_times_matrix:
 * @mat: 
 * @vec: 
 * 
 * Multiply matrix @mat (on the left) by vector @vec (on the
 * right). The vector is assumed to be horizontal or vertical,
 * whichever allows multiplication by algebra rules -- in this case,
 * it has to be vertical.
 * 
 * Return value: a matrix with the result
 **/
NxpElement *matrix_times_vector(NxpElementMatrix *mat, NxpElementVector *vec,
				NxpEvalContext *context, GError **error)
{
    NxpElementMatrix *result;
    NxpElement *tree, *el;
    gint i, j;

    g_return_val_if_fail(NXP_ELEMENT_MATRIX(mat), NULL);
    g_return_val_if_fail(NXP_ELEMENT_VECTOR(vec), NULL);

    if (vec->size != mat->width) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("Dimensions don't agree: matrix width (%i) != vector size (%i)"), mat->width, vec->size);
	return NULL;
    }

    /* allocate matrix for result width = 1; height = mat->height*/
    result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(1, mat->height, NULL));
    for (i = 0; i < mat->height; i++) {
	tree = NULL;
	for (j = 0; j < mat->width; j++) {
	    el = elem_TIMES(_element_matrix_get(mat, i, j),
			    _element_vector_get(vec, j));
	    if (!tree)
		tree = el;
	    else 
		tree = elem_PLUS(tree, el);
	}
	el = nxp_element_simplify(tree, context, NULL);
	NXP_ELEMENT_UNREF(tree);
	_element_matrix_set(result, i, 0, el);
	NXP_ELEMENT_UNREF(el);
    }
    return NXP_ELEMENT(result);
}



/**
 * matrix_times_matrix:
 * @mat1: 
 * @mat2: 
 * 
 * 
 * 
 * Return value: 
 **/
static NxpElement *mul(NxpOp op,NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error)
{
    NxpElementMatrix *mat1 = NXP_ELEMENT_MATRIX(a);
    NxpElementMatrix *mat2 = NXP_ELEMENT_MATRIX(b);

    NxpElementMatrix *result;
    NxpElement *tree, *el;

    gint i, j, k;

    g_return_val_if_fail(mat1, NULL);
    g_return_val_if_fail(mat2, NULL);

    if (mat1->width != mat2->height) {
	// TODO: - GERROR: 
	// matrix1 times matrix2: dimmensions don't agree:
	// matrix1 width (= %i) != matrix2 height (= %i).", 
	return NULL;
    }

    /* allocate matrix for result: width = mat2->width; height = mat1->height */
    result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(mat2->width,
						       mat1->height, NULL));
    for (i = 0; i < mat1->height; i++) {
	for (j = 0; j < mat2->width; j++) {
	    tree = NULL;
	    for (k = 0; k < mat1->width; k++) {
		el = elem_TIMES(_element_matrix_get(mat1, i, k),
				_element_matrix_get(mat2, k, j));
		if (!tree)
		    tree = el;
		else 
		    tree = elem_PLUS(tree, el);
	    }
	    el = nxp_element_eval(tree, context, NULL);
	    NXP_ELEMENT_UNREF(tree);
	    _element_matrix_set(result, i, j, el);
	    NXP_ELEMENT_UNREF(el);
	}
    }
    return NXP_ELEMENT(result);
}




/**
 * nxp_element_matrix_determinant:
 * @mat: 
 * 
 * 
 * 
 * Return value: 
 **/
NxpElement* nxp_element_matrix_determinant(NxpElementMatrix* mat, NxpEvalContext *context, GError **error)
{
    NxpElementMatrix *mat1;
    gint n, x, y;
    NxpElement *sum, *ret;

    g_return_val_if_fail(mat, NULL);

    if (!NXP_ELEMENT_MATRIX_IS_SQUARE(mat)) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("The argument should be a square matrix"));
	return NULL;
    }

    if (mat->width == 1)
    {
	NxpElement* zbr;
	zbr = _element_matrix_get(mat, 0, 0);
	return zbr;
    }

    if (mat->height == 2)
    {
	NxpElement* zbr, *ret;
	zbr = elem_MINUS(
	    elem_TIMES(
		_element_matrix_get(mat, 0, 0), 
		_element_matrix_get(mat, 1, 1)), 
	    elem_TIMES(
		_element_matrix_get(mat, 0, 1), 
		_element_matrix_get(mat, 1, 0)));
	ret = nxp_element_simplify(zbr, context, NULL);
	NXP_ELEMENT_UNREF(zbr);
	return ret;
    }

    if (mat->height == 3)
    {
	NxpElement* zbr, *ret;
	zbr = elem_MINUS(
	    elem_PLUS(
		elem_TIMES(
		    _element_matrix_get(mat, 0, 0), 
		    elem_TIMES(_element_matrix_get(mat, 1, 1), 
			       _element_matrix_get(mat, 2, 2))), 
		elem_PLUS(
		    elem_TIMES(
			_element_matrix_get(mat, 0, 1), 
			elem_TIMES(_element_matrix_get(mat, 2, 0), 
				   _element_matrix_get(mat, 1, 2))), 
		    elem_TIMES(
			_element_matrix_get(mat, 2, 1), 
			elem_TIMES(_element_matrix_get(mat, 0, 2), 
				   _element_matrix_get(mat, 2, 1))))), 
	    elem_PLUS(
		elem_TIMES(
		    _element_matrix_get(mat, 0, 2), 
		    elem_TIMES(_element_matrix_get(mat, 1, 1), 
			       _element_matrix_get(mat, 2, 0))), 
		elem_PLUS(
		    elem_TIMES(
			_element_matrix_get(mat, 0, 1), 
			elem_TIMES(_element_matrix_get(mat, 1, 0), 
				   _element_matrix_get(mat, 2, 2))), 
		    elem_TIMES(
			_element_matrix_get(mat, 2, 1), 
			elem_TIMES(_element_matrix_get(mat, 1, 2), 
				   _element_matrix_get(mat, 0, 0))))));
	ret = nxp_element_simplify(zbr, context, NULL);
	NXP_ELEMENT_UNREF(zbr);
	return ret;
    }

    /* bigger than 3x3 -- do it recursively */
    sum = nxp_element_real_new(0);

    /* Laplace expansion (column sweep) */
    for (n = 0; n < mat->height; n++)
    {
	NxpElement *det;
	NxpElement *pivot, *el;

	pivot = _element_matrix_get(mat, n, 0);
	mat1 = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(mat->width - 1, mat->height - 1, NULL));

	/* Build new matrix (mat1) by eliminating this column (1) and this row (n) */
	for (y = 0; y < mat->height; y++)
	{
	    if (y == n) continue;
	    for (x = 1; x < mat->width; x++) {
		el = _element_matrix_get(mat, y, x);
		_element_matrix_set(mat1, (y < n)? y: y - 1, x - 1, el);
		NXP_ELEMENT_UNREF(el);
	    }
	}
	det = nxp_element_matrix_determinant(mat1, context, NULL);
	NXP_ELEMENT_UNREF(mat1);	
	if (!det) {
	    /* garbage collection */
	    NXP_ELEMENT_UNREF(pivot);
	    NXP_ELEMENT_UNREF(sum);
	    return NULL;
	}
	if (n & 1)
	    sum = elem_MINUS(sum, elem_TIMES(pivot, det));
	else
	    sum = elem_PLUS(sum, elem_TIMES(pivot, det));
    }

#if 0
    /* Simplify some of the matrices (the ones with even size) */
    if (mat->width & 1)
	ret = NXP_ELEMENT_REF(sum);
    else
	ret = nxp_element_simplify(sum);
#else
    /* Simplify always  */
    ret = nxp_element_simplify(sum, context, NULL);
#endif
    NXP_ELEMENT_UNREF(sum);
    return ret;
}



static GString* str_func(NxpElement *elem, int *precedence) {
    GString *str = g_string_new("{");

    int i, j;
    NxpElement *el;
    NxpElementMatrix* m;
    GString* old;
    
    if (precedence) *precedence = 0;

    m = NXP_ELEMENT_MATRIX(elem);
    for (i = 0; i < m->height; i++) {
      for (j = 0; j < m->width; j++) {
	el = _element_matrix_get(m, i, j);

	old = nxp_element_str(el);


	NXP_ELEMENT_UNREF(el);
	str = g_string_append(str, old->str);
	str = g_string_append(str, ", ");
	g_string_free(old, TRUE);
      }
      str = g_string_truncate(str, (str)->len - 2);
      str = g_string_append(str, "; ");
    }
    str = g_string_truncate(str, (str)->len - 2);
    
    str = g_string_append_c(str, '}');

    return str;
}

#define BIND(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_MATRIX, \
							 NXP_ELEMENT_TYPE_MATRIX, \
    			                                 eval, symp,  0, NULL);
#define BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_MATRIX, \
   			                                 eval, symp,  0, NULL);


static NxpElement* sum(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error) 
{
    NxpElementMatrix *m_a, *m_b, *result;
    NxpElement *tree, *el;

    m_a = NXP_ELEMENT_MATRIX(a);
    m_b = NXP_ELEMENT_MATRIX(b);

    if (m_a->width!=m_b->width||m_a->height!=m_b->height) {
	// TODO -- GERROR SIZE DIFFERS
    } else {
	int i,j;
	result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(m_a->width,
							   m_b->height,
							   NULL));
	for(i = 0; i < m_a->width; i++)
	    for (j = 0; j < m_b->height; j++) 
	    {
		tree = elem_PLUS(_element_matrix_get(m_a, j, i),
				 _element_matrix_get(m_b, j, i));
		el = nxp_element_eval(tree, context, NULL);
		NXP_ELEMENT_UNREF(tree);
		_element_matrix_index(result, j, i) = el;
	    }
	return NXP_ELEMENT(result);
    }
    return NULL;
}


static NxpElement* eq(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) 
{
    NxpElementMatrix *m_a, *m_b;
    NxpElement *tree, *el;

    m_a = NXP_ELEMENT_MATRIX(a);
    m_b = NXP_ELEMENT_MATRIX(b);

    if (m_a->width!=m_b->width||m_a->height!=m_b->height) {
	return nxp_element_int_new(0);
    } else {
	int i,j;

	for(i = 0; i < m_a->width; i++)
	    for (j = 0; j < m_b->height; j++) {
		tree = elem_EQ(_element_matrix_get(m_a, j, i),
			       _element_matrix_get(m_b, j, i));
		el = nxp_element_eval(tree, context, NULL);
		NXP_ELEMENT_UNREF(tree);
		if (!NXP_ELEMENT_INT(el)->value) {
		    return el;
		}
	    }
    }
    return nxp_element_int_new(1);
}

static NxpElement* ne(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    NxpElement *el;
    el = eq(op,a,b,context,error);
    if (NXP_ELEMENT_INT(el)->value) {
	NXP_ELEMENT_UNREF(el);
	return nxp_element_int_new(0);
    }
    NXP_ELEMENT_UNREF(el);
    return nxp_element_int_new(1);
}
    

static NxpElement* uminus(NxpOp op, NxpElement *a, NxpElement *b, 
			  NxpEvalContext *context, GError **error) 
{
    NxpElementMatrix *m_b, *result;
    NxpElement *tree, *el;
    int i,j;
    
    m_b = NXP_ELEMENT_MATRIX(b);

    result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(m_b->width,
						       m_b->height,
						       NULL));
    for(i = 0; i < m_b->width; i++)
	for (j = 0; j < m_b->height; j++) 
	{
	    tree = elem_UMINUS(_element_matrix_get(m_b, j, i));
	    el = nxp_element_eval(tree, context, NULL);
	    NXP_ELEMENT_UNREF(tree);
	    _element_matrix_index(result, j, i) = el;
	}
    return NXP_ELEMENT(result);
}


static NxpElement* sub(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error) 
{
    NxpElementMatrix *m_a, *m_b, *result;
    NxpElement *tree, *el;
    
    m_a = NXP_ELEMENT_MATRIX(a);
    m_b = NXP_ELEMENT_MATRIX(b);
    
    if (m_a->width!=m_b->width||m_a->height!=m_b->height) {
	// TODO -- GERROR SIZE DIFFERS
    } else {
	int i,j;
	result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(m_a->width,
							   m_b->height,
							   NULL));
	for(i = 0; i < m_a->width; i++)
	    for (j = 0; j < m_b->height; j++) 
	    {
		tree = elem_MINUS(_element_matrix_get(m_a, j, i),
				  _element_matrix_get(m_b, j, i));
		el = nxp_element_eval(tree, context, NULL);
		NXP_ELEMENT_UNREF(tree);
		_element_matrix_index(result, j, i) = el;
	    }
	return NXP_ELEMENT(result);
    }
    return NULL;
}

NxpElement* nxp_eval_matrix(NxpElement *elem, NxpEvalContext *context, GError **error) {
	NxpElementMatrix* matrix = NXP_ELEMENT_MATRIX(elem);
	int width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
	int height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
	NxpElement* new = nxp_element_matrix_new(width, height, NULL);
	int i,j;

	for (i=0; i<height; i++) 
	    for (j=0; j<width; j++) {
		_element_matrix_index(NXP_ELEMENT_MATRIX(new),i,j) = 
		    nxp_element_eval(NXP_ELEMENT_MATRIX_ELEM_YX(matrix, i,j), context, NULL); /* FIXME: use error */
	    }
	return new;
	
}

NxpElement* nxp_simplify_matrix(NxpElement *elem, NxpEvalContext *context, GError **error) {
    NxpElementMatrix* matrix = NXP_ELEMENT_MATRIX(elem);
    int width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    int height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    NxpElement* new = nxp_element_matrix_new(width, height, NULL);
    int i,j;
    for (i=0; i<height; i++) 
	for (j=0; j<width; j++) {
	    NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new),i,j) = 
		nxp_element_simplify(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, i,j), context, NULL);
	    /* FIXME: use error */
	}
    return new;
}

static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc func,
              gpointer user_data,
              NxpEvalContext *context)
{
    NxpElementMatrix *mat = NXP_ELEMENT_MATRIX(elem);
    int i;
  
    for (i = 0; i < mat->width*mat->height; ++i)
    {
        NxpElement *item = NXP_ELEMENT(mat->data[i]);
        func(item, user_data);
        if (item->type_info->iterator_func)
            item->type_info->iterator_func(item, func, user_data, context);
    }
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var, NxpElement *el, NxpEvalContext *context, GError **error)
{
    guint i, w, h, size;
    NxpElementMatrix *retval, *a;
    GError *err = NULL;

    a = NXP_ELEMENT_MATRIX(self);
    h = a->height;
    w = a->width;
    size = h * w;
    retval = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(w, h, NULL));
    for (i = 0; i < size; i++) {
	retval->data[i] = nxp_element_substitute(a->data[i], var, el, context, &err);
	if (err) {
	    NXP_ELEMENT_UNREF(NXP_ELEMENT(retval));
	    g_propagate_error(error, err);
	    return NULL;
	}
    }
    return NXP_ELEMENT(retval);
}

static gboolean
dependson_func(NxpElement *self, NxpElement *variable, NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    NxpElementMatrix *mat = NXP_ELEMENT_MATRIX(self);
    gboolean retval = FALSE;
    int i, size = mat->width * mat->height;

    for (i = 0; !retval && i < size; ++i) {
	retval = nxp_element_depends_on(mat->data[i], variable, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    break;
	}
    }
    return retval;
}
	
int nxp_element_matrix_get_type(void)
{
    static int nxp_element_type_matrix = 0;

    if (nxp_element_type_matrix == 0) {
	nxp_element_type_matrix= nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_matrix,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_matrix;
}

void nxp_element_matrix_type_init() 
{
    typeinfo = nxp_element_type_get_info(nxp_element_matrix_get_type());
    typeinfo->name = g_strdup("matrix");
    typeinfo->eval_func = nxp_eval_matrix;
    typeinfo->destroy_func = nxp_destroy_matrix;
    typeinfo->simplify_func = nxp_simplify_matrix;
    typeinfo->derive_func = NULL;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->iterator_func = iterator_func;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;

    BIND( sum, NULL, NXP_OP_PLUS);
    BIND( sub, NULL, NXP_OP_MINUS);
    BIND_LEFT(uminus, NULL, NXP_OP_MINUS_LEFT);
    BIND( mul, NULL, NXP_OP_STAR);

    //    NXP_OP_SLASH --> Matrix division?
    //    NXP_OP_HAT   --> Matrix exponentiation?

    /// not implementable ///////
    //
    //    NXP_OP_DOT3
    //    NXP_OP_PLUS2_LEFT
    //    NXP_OP_MINUS2_LEFT
    //    NXP_OP_PLUS2_RIGHT
    //    NXP_OP_MINUS2_RIGHT
    /////////////////////////////

    // >,<,>= and <= compares what size?
    //    BIND( gt, gt, NXP_OP_GT);
    //    BIND( le, le, NXP_OP_LT_EQ);
    //    BIND( lt, lt, NXP_OP_LT);
    //    BIND( ge, ge, NXP_OP_GT_EQ);

    // == and != compares ALL
    BIND( eq, NULL, NXP_OP_EQ2);
    BIND( ne, NULL, NXP_OP_EXCL_EQ);

    // There are no empty matrices
    //   BIND( and, and, NXP_OP_SLASH_BSLASH);
    //   BIND( or,  or,  NXP_OP_BSLASH_SLASH);
    //   BIND_LEFT( not, not, NXP_OP_TILDE_LEFT);

    ///// not implementable ///
    //
    //    NXP_OP_EXCL_RIGHT
    //    NXP_OP_EQ
    //    NXP_OP_COLON_EQ
    ///////////////////////////
}

