/* -*- Mode: C; c-file-style: "stroustrup"; coding: latin-1-unix -*- */
#include <stdio.h>
#include "mpoly.h"

#include "message.h"

#include "errors.h"
#include "variable.h"
#include "tree.h"
#include "frac.h"
#include "element.h"
#include "element_output.h"
#include "types_misc.h"
#include "void.h"
#include "i18n.h"
#include "bifuncs.h"
#include "prototypes.h"
#include "polynomials/nxpintpoly.h"
#include "polynomials/nxpintpoly_mod.h"
#include "named_args.h"

static NxpElementTypeInfo *mpoly_typeinfo = NULL;

inline NxpElement*
nxp_monomial_new(NxpElement *coeff, NxpElement *vector)
{
    NxpElement *elem;

    elem = nxp_element_vector_new(2, NULL);
    NXP_MONOMIAL_COEFF(elem) = NXP_ELEMENT_REF(coeff);
    NXP_MONOMIAL_VECTOR(elem) = NXP_ELEMENT_REF(vector);

    return elem;
}

static gint nxp_monomial_compare(NxpElement *a, NxpElement *b)
{
    return -nxp_element_vector_int_compare(NXP_MONOMIAL_VECTOR(a), NXP_MONOMIAL_VECTOR(b));
}

/**
 * nxp_monomial_perm:
 * @mon: the monomial we work with.
 * @from: index of the element we want to move.
 * @to: new index of the element.
 *
 * Move the monomial variable of index @from to the index @to.
 *
 * Return value: the new monomial.
 *
 * Example : With from=0 and to=2, c^8*a^2*b^3 becomes a^2*b^3*c^8 !!! :-)
 **/
NxpElement *nxp_monomial_perm(NxpElement *mon, int from, int to)
{
    NxpElement *retval;
    NxpElement *coeff;
    int i;
    NxpElementVector *v, *vector = NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mon));
    int size = vector->size;

    g_return_val_if_fail(from < size && to < size, NULL);

    coeff = NXP_MONOMIAL_COEFF(mon);
    v = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < size; i++) {
	if (i == to)
	    _element_vector_set(v, i, _element_vector_index(vector, from));
	else if (i == from)
	    _element_vector_set(v, i, _element_vector_index(vector, to));
	else
	    _element_vector_set(v, i, _element_vector_index(vector, i));
    }
    retval = nxp_monomial_new(coeff, NXP_ELEMENT(v));
    NXP_ELEMENT_UNREF(v);
    return retval;
}

static NxpElement* nxp_monomial_insert(NxpElement *mon, int index,
				       NxpElement *el)
{
    NxpElement *retval;
    NxpElement *coeff;
    NxpElement *vector;

    coeff = NXP_MONOMIAL_COEFF(mon);
    vector = nxp_element_vector_insert(NXP_MONOMIAL_VECTOR(mon), index, el);
    retval = nxp_monomial_new(coeff, vector);
    NXP_ELEMENT_UNREF(vector);
    return retval;
}

static NxpElement* nxp_monomial_reduce(NxpElement *mon, int i1, int i2,
				       NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElement *coeff;
    NxpElement *vector;
    NxpElement *number;
    int size;
    int i, j;

    coeff = NXP_MONOMIAL_COEFF(mon);
    size = NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mon))->size;
    vector = nxp_element_vector_new(size - 1, NULL);
    for (i = 0, j = 0; i < size; i++) {
	if (i == i1) {
	    number = nxp_element_tree_simplify_op(NXP_OP_PLUS,
		    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i1),
		    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i2),
		    context, error);
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), j++) = number;
	}
	else if (i != i2)
	    _element_vector_set(NXP_ELEMENT_VECTOR(vector),
			j++,
			_element_vector_index(NXP_ELEMENT_VECTOR(vector), i));
    }
    retval = nxp_monomial_new(coeff, vector);
    NXP_ELEMENT_UNREF(vector);
    return retval;
}

/************************** Multivariable polynomial *************************/

NxpElement *nxp_element_mpoly_new_simple(NxpElement *monomial, NxpElement *args)
{
    NxpElement *elem;

    elem = nxp_element_alloc(mpoly_typeinfo, sizeof(NxpElementMpoly));
    NXP_ELEMENT_MPOLY(elem)->mon = NXP_ELEMENT_LIST(nxp_element_list_new_simple(monomial));
    NXP_ELEMENT_MPOLY(elem)->args = NXP_ELEMENT_LIST(NXP_ELEMENT_REF(args));
    return (NxpElement *)elem;
}

NxpElement *nxp_element_mpoly_new(NxpElement *mon, NxpElement *args)
{
    NxpElement *elem;
    elem = nxp_element_alloc(mpoly_typeinfo, sizeof(NxpElementMpoly));
    NXP_ELEMENT_MPOLY(elem)->mon = NXP_ELEMENT_LIST(NXP_ELEMENT_REF(mon));
    NXP_ELEMENT_MPOLY(elem)->args = NXP_ELEMENT_LIST(NXP_ELEMENT_REF(args));
    return (NxpElement *)elem;
}

static void nxp_mpoly_destroy(NxpElement *elem)
{
    NxpElementMpoly *pol = NXP_ELEMENT_MPOLY(elem);

    NXP_ELEMENT_UNREF(pol->mon);
    NXP_ELEMENT_UNREF(pol->args);
}

static GString* pol_repr_func(NxpElement *elem, int *precedence)
{
    GSList *args, *mon;
    NxpElementMpoly *pol = NXP_ELEMENT_MPOLY(elem);
    NxpElementTypeInfo *typeinfo;
    NxpElement *nb;
    GString *string, *var, *var1, *coeff, *coeff1, *var_den;
    int prec;
    int i, pow, n_mon;
    int s, c, c1;
    gboolean star = FALSE, star_den, powers;

    string = g_string_new("");
    n_mon = 0;
    for (mon = pol->mon->list; mon; mon = mon->next) {
	NxpElementNumber *numb = NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(mon->data));
	s = nxp_element_number_compare_int(numb, 0);
	c = nxp_element_number_compare_int(numb, 1);
	c1 = nxp_element_number_compare_int(numb, -1);

	coeff = g_string_new("");

	if (s != 0) {
	    n_mon++;
	    if (string->str[0] && s > 0)
		coeff = g_string_append(coeff, " + ");
	    else if (string->str[0] && s < 0)
		coeff = g_string_append(coeff, " - ");
	    else if (string->str[0] == 0 && s < 0)
		coeff = g_string_append(coeff, "-");

	    powers = nxp_element_is_true(NXP_MONOMIAL_VECTOR(mon->data),
					 NULL, NULL);
	    if ((c && c1) || !powers) {
		if (s > 0) {
		    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(numb);
		    coeff1 = typeinfo->repr_func(NXP_ELEMENT(numb), &prec);
		    if (prec > 40) {
			coeff1 = g_string_prepend(coeff1, "(");
			coeff1 = g_string_append(coeff1, ")");
		    }
		}
		else {
		    nb = nxp_element_number_opposed(numb);
		    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(nb);
		    coeff1 = typeinfo->repr_func(nb, &prec);
		    if (prec > 40) {
			coeff1 = g_string_prepend(coeff1, "(");
			coeff1 = g_string_append(coeff1, ")");
		    }
		    NXP_ELEMENT_UNREF(nb);
		}
		coeff = g_string_append(coeff, coeff1->str);
		g_string_free(coeff1, TRUE);
	    }

	    string = g_string_append(string, coeff->str);

	    if (powers) {
		if (c && c1)
		    string = g_string_append(string, " * ");

		var = g_string_new("");
		var_den = g_string_new("");
		for (args = pol->args->list, i = 0; args; args = args->next, i++) {
		    star_den = FALSE;
		    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(args->data);
		    var1 = typeinfo->repr_func(args->data, &prec);
		    if (prec >= 40) {
			var1 = g_string_prepend(var1, "(");
			var1 = g_string_append(var1, ")");
		    }
		    pow = nxp_element_int_get_si(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mon->data)), i), NULL);
		    if (pow > 0) {
			if (var->str[0] != 0) {
			    var = g_string_append(var, " * ");
			    star = TRUE;
			}
			if (pow == 1)
			    g_string_append_printf(var, "%s", var1->str);
			else
			    g_string_append_printf(var, "%s ^ %d", var1->str, pow);
		    }
		    else if (pow < 0) {
			if (var_den->str[0] != 0) {
			    var_den = g_string_append(var_den, " * ");
			    star_den = TRUE;
			}
			if (pow == -1)
			    g_string_append_printf(var_den, "%s", var1->str);
			else
			    g_string_append_printf(var_den, "%s ^ %d", var1->str, -pow);
		    }
		    g_string_free(var1, TRUE);
		}
		if (star_den) {
		    var_den = g_string_prepend(var_den, "(");
		    var_den = g_string_append(var_den, ")");
		}
		if (var->str[0] != 0) {
		    if (var_den->str[0] != 0)
			g_string_append_printf(var, " / %s", var_den->str);
		}
		else {
		    if (var_den->str[0] != 0)
			g_string_append_printf(var, "1 / %s", var_den->str);
		}
		g_string_free(var_den, TRUE);
		string = g_string_append(string, var->str);
		g_string_free(var, TRUE);
	    }
	}
    }
    if (string->str[0] == 0)
	string = g_string_append(string, "0");

//    string = g_string_prepend(string, "##");
//    string = g_string_append(string, "##");

    if (precedence) {
	if (n_mon > 1)
	    * precedence = 50;
	else if (star)
	    * precedence = 40;
	else
	    * precedence = 30;
    }

    return string;
}

NxpElement *nxp_element_mpoly_insert(NxpElement *p, NxpElement *el,
					    int index)
{
    NxpElement *mon, *args, *m;
    NxpElementMpoly *pol;
    NxpElement *retval;
    NxpElement *zero = nxp_element_int_new(0);
    GSList *l;

    pol = NXP_ELEMENT_MPOLY(p);
    args = nxp_element_list_insert(NXP_ELEMENT(pol->args), el, index);

    mon = nxp_element_list_new();
    for (l = pol->mon->list; l; l = l->next) {
	m = nxp_monomial_insert(NXP_ELEMENT(l->data), index, zero);
	nxp_element_list_append(mon, m);
	NXP_ELEMENT_UNREF(m);
    }

    NXP_ELEMENT_UNREF(zero);
    retval = nxp_element_mpoly_new(mon, args);
    NXP_ELEMENT_UNREF(mon);
    NXP_ELEMENT_UNREF(args);
    return retval;
}

NxpElement *nxp_element_mpoly_opposed(NxpElement *p)
{
    NxpElement *mon, *m1;
    NxpElement *coeff;
    NxpElementMpoly *pol;
    NxpElement *retval;
    GSList *l;

    pol = NXP_ELEMENT_MPOLY(p);
    mon = nxp_element_list_new();
    for (l = pol->mon->list; l; l = l->next) {
	coeff = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(l->data)));
	m1 = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(l->data));
	nxp_element_list_append(mon, NXP_ELEMENT(m1));
	NXP_ELEMENT_UNREF(m1);
	NXP_ELEMENT_UNREF(coeff);
    }
    retval = nxp_element_mpoly_new(mon, NXP_ELEMENT(pol->args));
    NXP_ELEMENT_UNREF(mon);
    return retval;
}

NxpElement *nxp_element_mpoly_add_monomial(NxpElement *p, NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElement *mon, *args, *m;
    NxpElementMpoly *pol;
    NxpElement *retval, *coeff;
    GSList *l;
    gboolean seen = FALSE;

    pol = NXP_ELEMENT_MPOLY(p);
    args = NXP_ELEMENT(pol->args);

    mon = nxp_element_list_new();
    for (l = pol->mon->list; l; l = l->next) {
	if (nxp_element_eq(NXP_MONOMIAL_VECTOR(el), NXP_MONOMIAL_VECTOR(l->data), context, error)) {
	    coeff = nxp_element_tree_simplify_op(NXP_OP_PLUS, NXP_MONOMIAL_COEFF(el), NXP_MONOMIAL_COEFF(l->data), context, error);
	    m = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(el));
	    nxp_element_list_append(mon, m);
	    NXP_ELEMENT_UNREF(m);
	    NXP_ELEMENT_UNREF(coeff);
	    seen = TRUE;
	}
	else
	    nxp_element_list_append(mon, NXP_ELEMENT(l->data));
    }

    if (!seen)
	nxp_element_list_append(mon, el);

    retval = nxp_element_mpoly_new(mon, args);
    NXP_ELEMENT_UNREF(mon);
    return retval;
}

static NxpElement *nxp_element_mpoly_perm(NxpElement *p, int i1, int i2)
{
    GSList *l;
    NxpElement *args, *mon;
    NxpElementMpoly *pol;
    NxpElement *el1, *el2, *tmp;
    NxpElement *retval;
    int i;

    if (i1 != i2) {
	pol = NXP_ELEMENT_MPOLY(p);
	l = pol->args->list;
	el1 = g_slist_nth_data(l, i1);
	el2 = g_slist_nth_data(l, i2);

	args = nxp_element_list_new();
	for (i = 0; l; i++, l = l->next) {
	    if (i == i1)
		nxp_element_list_append(args, el2);
	    else if (i == i2)
		nxp_element_list_append(args, el1);
	    else
		nxp_element_list_append(args, l->data);
	}
	mon = nxp_element_list_new();
	for (l = pol->mon->list; l; l = l->next) {
	    tmp = nxp_monomial_perm(NXP_ELEMENT(l->data), i1, i2);
	    nxp_element_list_append(mon, tmp);
	    NXP_ELEMENT_UNREF(tmp);
	}

	retval = nxp_element_mpoly_new(mon, args);
	NXP_ELEMENT_UNREF(mon);
	NXP_ELEMENT_UNREF(args);

	return retval;
    }
    else
	return NXP_ELEMENT_REF(p);
}

static NxpElement *nxp_element_mpoly_reduce(NxpElement *p, int i1, int i2,
			NxpEvalContext *context, GError **error)
{
    GSList *l;
    NxpElement *args, *mon;
    NxpElement *retval;
    NxpElementMpoly *pol;
    int i;

    if (i1 != i2) {
	pol = NXP_ELEMENT_MPOLY(p);
	args = nxp_element_list_new();
	for (i = 0, l = pol->args->list; l; i++, l = l->next) {
	    if (i != i2)
		nxp_element_list_append(args, l->data);
	}

	mon = nxp_element_list_new();
	for (l = pol->mon->list; l; l = l->next) {
	    mon = nxp_element_list_new();
	    nxp_element_list_append(mon, nxp_monomial_reduce(NXP_ELEMENT(l->data), i1, i2, context, error));
	}
	retval = nxp_element_mpoly_new(mon, args);
	NXP_ELEMENT_UNREF(mon);
	NXP_ELEMENT_UNREF(args);
	return retval;
    }
    else
	return NXP_ELEMENT_REF(p);
}

static void _element_mpoly_reduce(NxpElement *p, NxpEvalContext *context, GError **error)
{
    GSList *l;
    NxpElement *mona, *monb;
    NxpElement *nb;

    l = NXP_ELEMENT_MPOLY(p)->mon->list;
    while(l && l->next) {
	mona = l->data;
	monb = l->next->data;
	if (nxp_element_vector_int_compare(NXP_MONOMIAL_VECTOR(mona),
				       NXP_MONOMIAL_VECTOR(monb)) == 0) {
	    nb = nxp_element_tree_simplify_op(NXP_OP_PLUS, NXP_MONOMIAL_COEFF(mona), NXP_MONOMIAL_COEFF(monb), context, error);
	    NXP_ELEMENT_UNREF(NXP_MONOMIAL_COEFF(mona));
	    NXP_MONOMIAL_COEFF(mona) = nb;
	    l->next = l->next->next;
	    NXP_ELEMENT_UNREF(monb);
	}
	else
	    l = l->next;
    }
}

static NxpElement *pol_eval(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement *args, *retval;
    NxpElement *coeff, *vector;
    NxpElement *tmp, *tmp1, *tmp2;
    GSList *l, *lst;
    GError *err = NULL;
    int i;

    args = nxp_element_eval(NXP_ELEMENT(NXP_ELEMENT_MPOLY(elem)->args), context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = nxp_element_int_new(0);
    for (l = NXP_ELEMENT_MPOLY(elem)->mon->list; l; l = l->next) {
	coeff = NXP_MONOMIAL_COEFF(l->data);
	vector = NXP_MONOMIAL_VECTOR(l->data);
	tmp1 = NXP_ELEMENT_REF(coeff);
	for (lst = NXP_ELEMENT_LIST(args)->list, i = 0; lst; lst = lst->next, i++) {
	    tmp = nxp_element_tree_eval_op(NXP_OP_HAT, NXP_ELEMENT(lst->data),
					   _element_vector_index(NXP_ELEMENT_VECTOR(vector), i),
					   context, error);
	    tmp2 = nxp_element_tree_eval_op(NXP_OP_STAR, tmp, tmp1,
					    context, error);
	    NXP_ELEMENT_UNREF(tmp);
	    NXP_ELEMENT_UNREF(tmp1);
	    tmp1 = tmp2;
	}
	tmp = nxp_element_tree_eval_op(NXP_OP_PLUS, retval, tmp1, context, error);
	NXP_ELEMENT_UNREF(retval);
	retval = tmp;
    }
    NXP_ELEMENT_UNREF(args);
    return retval;
}

static NxpElement *pol_simplify(NxpElement *elem, NxpEvalContext *context,
				GError **error)
{
    NxpElement *args, *retval;
    NxpElement *coeff, *vector;
    NxpElement *tmp, *tmp1, *tmp2;
    GSList *l, *lst;
    GError *err = NULL;
    int i;

    args = nxp_element_simplify(NXP_ELEMENT(NXP_ELEMENT_MPOLY(elem)->args),
				context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = nxp_element_int_new(0);
    for (l = NXP_ELEMENT_MPOLY(elem)->mon->list; l; l = l->next) {
	coeff = NXP_MONOMIAL_COEFF(l->data);
	vector = NXP_MONOMIAL_VECTOR(l->data);
	tmp1 = NXP_ELEMENT_REF(coeff);
	for (lst = NXP_ELEMENT_LIST(args)->list, i = 0; lst; lst = lst->next, i++) {
	    tmp = nxp_element_tree_simplify_op(NXP_OP_HAT,
					   NXP_ELEMENT(lst->data),
					   _element_vector_index(NXP_ELEMENT_VECTOR(vector), i),
					   context, error);
	    tmp2 = nxp_element_tree_simplify_op(NXP_OP_STAR, tmp, tmp1,
						context, error);
	    NXP_ELEMENT_UNREF(tmp);
	    NXP_ELEMENT_UNREF(tmp1);
	    tmp1 = tmp2;
	}
	tmp = nxp_element_tree_simplify_op(NXP_OP_PLUS, retval,
					   tmp1, context, error);
	NXP_ELEMENT_UNREF(retval);
	retval = tmp;
    }
    NXP_ELEMENT_UNREF(args);
    return retval;
}

static gboolean pol_dependson(NxpElement *elem, NxpElement *var, NxpEvalContext *context, GError **error)
{
    return nxp_element_depends_on(NXP_ELEMENT(NXP_ELEMENT_MPOLY(elem)->args),
				  var, context, error);
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var,
				   NxpElement *el, NxpEvalContext *context,
				   GError **error)
{
    NxpElement *retval, *args;

    args = nxp_element_substitute(NXP_ELEMENT(NXP_ELEMENT_MPOLY(self)->args), var, el, context, error);
    retval = nxp_element_mpoly_new(NXP_ELEMENT(NXP_ELEMENT_MPOLY(self)->mon), args);
    NXP_ELEMENT_UNREF(args);
    return retval;
}

static NxpElement *mon_derive(NxpElement *mon, NxpElement *args, 
			      NxpIdentifier *ident, NxpEvalContext *context,
			      GError **error)
{
    int i, j;
    NxpElement *tmp, *tmp1, *tmp2, *new;
    NxpElement *retval = NULL;
    GSList *largsd = NXP_ELEMENT_LIST(args)->list, *l;
    int size = nxp_element_list_length(args);
    NxpElement *one = nxp_element_int_new(1);
    NxpElementVector *vector = (NxpElementVector *)NXP_MONOMIAL_VECTOR(mon);
    NxpElement *coeff = NXP_MONOMIAL_COEFF(mon);
    GError *err = NULL;

    for (i = 0; i < size; i++) {
	tmp = nxp_element_derive(largsd->data, ident, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(one);
	    return NULL;
	}
	if (nxp_element_is_true(_element_vector_index(vector, i), context, error)) {
	    g_assert(*error == NULL);
	    tmp1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
						_element_vector_index(vector, i),
						tmp, context, error);
	    g_assert(*error == NULL);
	    NXP_ELEMENT_UNREF(tmp);
	    tmp2 = nxp_element_tree_simplify_op(NXP_OP_MINUS,
					       _element_vector_index(vector, i),
					       one, context, error);
	    g_assert(*error == NULL);
	    tmp = nxp_element_tree_simplify_op(NXP_OP_HAT,
					       largsd->data,
					       tmp2, context, error);
	    g_assert(*error == NULL);
	    NXP_ELEMENT_UNREF(tmp2);
	    tmp2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					  tmp1, tmp, context, error);
	    g_assert(*error == NULL);
	    NXP_ELEMENT_UNREF(tmp1);
	    NXP_ELEMENT_UNREF(tmp);
	    tmp = nxp_element_tree_simplify_op(NXP_OP_STAR,
					   coeff, tmp2, context, error);
	    g_assert(*error == NULL);

	    /* Here, tmp is coeff*a0*arg0^(a0-1)*arg0' */

	    new = NXP_ELEMENT_REF(one);
	    for (j = 0, l = NXP_ELEMENT_LIST(args)->list; j < size; j++, l = l->next) {
		if (j == i) {
		    tmp2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
						   new, tmp, context, error);
	    g_assert(*error == NULL);
		    NXP_ELEMENT_UNREF(tmp);
		    NXP_ELEMENT_UNREF(new);
		    new = tmp2;
		}
		else {
		    tmp1 = nxp_element_tree_simplify_op(NXP_OP_HAT,
						    l->data,
						    _element_vector_index(vector, j),
						    context, error);
	    g_assert(*error == NULL);
		    tmp2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
						   new, tmp1, context, error);
	    g_assert(*error == NULL);
		    NXP_ELEMENT_UNREF(tmp1);
		    NXP_ELEMENT_UNREF(new);
		    new = tmp2;
		}
	    }
	}
	else {
	    NXP_ELEMENT_UNREF(tmp);
	    new = nxp_element_int_new(0);
	}
	if (retval) {
	    tmp = nxp_element_tree_simplify_op(NXP_OP_PLUS,
		    retval,
		    new, context, error);
	    g_assert(*error == NULL);
	    NXP_ELEMENT_UNREF(retval);
	    NXP_ELEMENT_UNREF(new);
	    retval = tmp;
	}
	else
	    retval = new;

	largsd = largsd->next;
    }
    NXP_ELEMENT_UNREF(one);
    return retval;
}

static NxpElement *pol_derive(NxpElement *elem, NxpIdentifier *ident,
			      NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    GSList * mons = NXP_ELEMENT_MPOLY(elem)->mon->list;
    NxpElement *args = nxp_element_simplify(NXP_ELEMENT(NXP_ELEMENT_MPOLY(elem)->args),
					    context, error);
    NxpElement *retval = NULL, *tmp, *tmp1;

    while (mons) {
	tmp = mon_derive(mons->data, args, ident, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(args);
	    NXP_ELEMENT_XUNREF(retval);
	    return NULL;
	}
	if (retval) {
	    tmp1 = nxp_element_tree_simplify_op(NXP_OP_PLUS,
						retval, tmp,
						context, error);
	    NXP_ELEMENT_UNREF(retval);
	    NXP_ELEMENT_UNREF(tmp);
	    retval = tmp1;
	}
	else
	    retval = tmp;
	mons = mons->next;
    }
    return retval;
}

static inline void _mpoly_sort(NxpElementMpoly *p)
{
    p->mon->list = g_slist_sort(p->mon->list, (GCompareFunc)nxp_monomial_compare);
}

static GSList *_element_mpoly_same_args(NxpElement *a, NxpElement *b, NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *aa, *bb;
    GSList *argsa, *argsb, *retval = NULL;
    NxpElement *p, *el_l;
    int i = 0, j;

    aa = NXP_ELEMENT_MPOLY(NXP_ELEMENT_REF(a));
    bb = NXP_ELEMENT_MPOLY(NXP_ELEMENT_REF(b));
    argsa = aa->args->list;
    el_l = nxp_element_list_new();
    while (argsa) {
	j = nxp_element_list_contains(el_l, argsa->data, context, error);
	if (j < 0) {
	    nxp_element_list_append(el_l, argsa->data);
	    j = nxp_element_list_contains(NXP_ELEMENT(bb->args), argsa->data, context, error);
	    if (j < 0) {
		p = nxp_element_mpoly_insert(NXP_ELEMENT(bb), argsa->data, i);
		NXP_ELEMENT_UNREF(bb);
		bb = NXP_ELEMENT_MPOLY(p);
	    }
	    else {
		p = nxp_element_mpoly_perm(NXP_ELEMENT(bb), j, i);
		NXP_ELEMENT_UNREF(bb);
		bb = NXP_ELEMENT_MPOLY(p);
	    }
	    i++;
	    argsa = argsa->next;
	}
	else {
	    p = nxp_element_mpoly_reduce(NXP_ELEMENT(aa), j, i, context, error);
	    NXP_ELEMENT_UNREF(aa);
	    aa = NXP_ELEMENT_MPOLY(p);
	}
    }
    argsb = g_slist_nth(bb->args->list, i);

    while (argsb) {
	j = nxp_element_list_contains(el_l, argsb->data, context, error);
	if (j < 0) {
	    nxp_element_list_append(el_l, argsb->data);
	    p = nxp_element_mpoly_insert(NXP_ELEMENT(aa), argsb->data, i);
	    NXP_ELEMENT_UNREF(aa);
	    aa = NXP_ELEMENT_MPOLY(p);
	    i++;
	    argsb = argsb->next;
	}
	else {
	    p = nxp_element_mpoly_reduce(NXP_ELEMENT(bb), j, i, context, error);
	    NXP_ELEMENT_UNREF(bb);
	    bb = NXP_ELEMENT_MPOLY(p);
	}
    }
    NXP_ELEMENT_UNREF(el_l);
    retval = g_slist_append(retval, aa);
    retval = g_slist_append(retval, bb);
    _mpoly_sort(aa);
    _mpoly_sort(bb);
    return retval;
}

static NxpElement *pol_sub(NxpOp op, NxpElement *a, NxpElement *b, NxpEvalContext *context, GError **error)
{
    NxpElement *mona, *monb;
    NxpElement *coeff, *monb1;
    NxpElement *el, *mon, *nb;
    GSList *la, *lb, *l;
    int test;
    NxpElement *mons;
    NxpElementMpoly *aa, *bb;

    l = _element_mpoly_same_args(a, b, context, error);
    aa = l->data;
    bb = l->next->data;
    g_slist_free(l);
    l = NULL;

    la = aa->mon->list;
    lb = bb->mon->list;

    mons = nxp_element_list_new();
    do {
	mona = la->data;
	monb = lb->data;

	test = nxp_element_vector_int_compare(NXP_MONOMIAL_VECTOR(mona), NXP_MONOMIAL_VECTOR(monb));
	if (test < 0) {
	    nxp_element_list_append(mons, NXP_ELEMENT(mona));
	    la = la->next;
	}
	else if (test > 0) {
	    coeff = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(monb)));
	    monb1 = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(monb));
	    NXP_ELEMENT_UNREF(coeff);
	    nxp_element_list_append(mons, NXP_ELEMENT(monb1));
	    NXP_ELEMENT_UNREF(monb1);
	    lb = lb->next;
	}
	else {
	    nb = nxp_element_tree_simplify_op(op, NXP_MONOMIAL_COEFF(mona), NXP_MONOMIAL_COEFF(monb), context, error);
	    mon = nxp_monomial_new(nb, NXP_MONOMIAL_VECTOR(mona));
	    NXP_ELEMENT_UNREF(nb);
	    nxp_element_list_append(mons, mon);
	    NXP_ELEMENT_UNREF(mon);
	    la = la->next;
	    lb = lb->next;
	}
    }
    while (la && lb);

    while (la) {
	mona = la->data;
	nxp_element_list_append(mons, NXP_ELEMENT(mona));
	la = la->next;
    }

    while (lb) {
	monb = lb->data;
	coeff = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(monb)));
	monb1 = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(monb));
	NXP_ELEMENT_UNREF(coeff);
	nxp_element_list_append(mons, NXP_ELEMENT(monb1));
	NXP_ELEMENT_UNREF(monb1);
	lb = lb->next;
    }
    el = nxp_element_mpoly_new(mons, NXP_ELEMENT(aa->args));
    NXP_ELEMENT_UNREF(aa);
    NXP_ELEMENT_UNREF(bb);
    NXP_ELEMENT_UNREF(mons);
    return el;
}

static NxpElement *pol_add(NxpOp op, NxpElement *a, NxpElement *b, NxpEvalContext *context, GError **error)
{
    NxpElement *mona, *monb;
    NxpElement *el, *mon, *nb;
    GSList *la, *lb, *l;
    int test;
    NxpElement *mons;

    NxpElementMpoly *aa, *bb;

    l = _element_mpoly_same_args(a, b, context, error);
    aa = l->data;
    bb = l->next->data;
    g_slist_free(l);

    la = aa->mon->list;
    lb = bb->mon->list;

    mons = nxp_element_list_new();
    do {
	mona = la->data;
	monb = lb->data;

	test = nxp_element_vector_int_compare(NXP_MONOMIAL_VECTOR(mona), NXP_MONOMIAL_VECTOR(monb));
	if (test > 0) {
	    nxp_element_list_append(mons, mona);
	    la = la->next;
	}
	else if (test < 0) {
	    nxp_element_list_append(mons, monb);
	    lb = lb->next;
	}
	else {
	    nb = nxp_element_tree_simplify_op(op, NXP_MONOMIAL_COEFF(mona), NXP_MONOMIAL_COEFF(monb), context, error);
	    mon = nxp_monomial_new(nb, NXP_MONOMIAL_VECTOR(mona));
	    NXP_ELEMENT_UNREF(nb);
	    nxp_element_list_append(mons, NXP_ELEMENT(mon));
	    la = la->next;
	    lb = lb->next;
	}
    }
    while (la && lb);

    while (la) {
	mona = la->data;
	nxp_element_list_append(mons, mona);
	la = la->next;
    }

    while (lb) {
	monb = lb->data;
	nxp_element_list_append(mons, monb);
	lb = lb->next;
    }
    el = nxp_element_mpoly_new(mons, NXP_ELEMENT(aa->args));
    NXP_ELEMENT_UNREF(aa);
    NXP_ELEMENT_UNREF(bb);
    NXP_ELEMENT_UNREF(mons);
    return el;
}

static NxpElement *pol_minus(NxpOp op, NxpElement *a, NxpElement *b, NxpEvalContext *context, GError **error)
{
    return nxp_element_mpoly_opposed(b);
}

static NxpElement *pol_div(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *aa = NXP_ELEMENT_MPOLY(NXP_ELEMENT_REF(a));
    NxpElementMpoly *bb = NXP_ELEMENT_MPOLY(NXP_ELEMENT_REF(b));
    NxpElement *el, *el_mon, *mona, *monb, *mon, *nb, *vector;
    GSList *l;

    if (nxp_element_list_length(NXP_ELEMENT(bb->mon)) == 1
	    && nxp_element_list_length(NXP_ELEMENT(aa->mon)) == 1) {
	l = _element_mpoly_same_args(a, b, context, error);
	aa = l->data;
	bb = l->next->data;
	g_slist_free(l);

	el_mon = nxp_element_list_new();
	mona = aa->mon->list->data;
	monb = bb->mon->list->data;
	nb = nxp_element_tree_simplify_op(NXP_OP_SLASH, NXP_MONOMIAL_COEFF(mona), NXP_MONOMIAL_COEFF(monb), context, error);
	vector = nxp_element_tree_simplify_op(NXP_OP_MINUS, NXP_MONOMIAL_VECTOR(mona), NXP_MONOMIAL_VECTOR(monb), context, error);
	mon = nxp_monomial_new(nb, vector);
	nxp_element_list_prepend(el_mon, mon);
	NXP_ELEMENT_UNREF(mon);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(nb);

	el = nxp_element_mpoly_new(el_mon, NXP_ELEMENT(aa->args));
	NXP_ELEMENT_UNREF(el_mon);

	NXP_ELEMENT_UNREF(aa);
	NXP_ELEMENT_UNREF(bb);

	return el;
    }
    else
	return nxp_element_tree_new(op, a, b);
}

static NxpElement *nxp_element_mpoly_deg(NxpElement *p, NxpElement *var, GHashTable *named_args, NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *pol = NXP_ELEMENT_MPOLY(p);
    NxpElementVector *v;
    int i;
    mpz_t deg;
    GSList *l;

    l = pol->args->list;
    i = 0;
    do {
	if (nxp_element_eq(l->data, var, context, error)) {
	    l = pol->mon->list;
	    mpz_init_set_si(deg, -1);
	    do {
		v = NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(l->data));
		if (mpz_sgn(NXP_ELEMENT_INT(_element_vector_index(v, i))->value) < 0) {
		    g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("The first argument is not a polynomial in the given variable"));
		    mpz_clear(deg);
		    return NULL;
		}
		if (mpz_cmp(deg, NXP_ELEMENT_INT(_element_vector_index(v, i))->value) < 0)
		    mpz_set(deg, NXP_ELEMENT_INT(_element_vector_index(v, i))->value);
		l = l->next;
	    }
	    while(l);
	    return nxp_element_int_new_from_mpz(deg);
	}
	else {
	    i++;
	    l = l->next;
	}
    }
    while(l);
    return nxp_element_int_new(-1);
}

NxpElement *nxp_element_mpoly_euclidean(NxpElement *p1, NxpElement *p2, NxpEvalContext *context, GError **error)
{
    NxpIntPoly a1[1], a2[1];
    NxpIntPoly r[1];
    NxpElement *retval;
    GError *err = NULL;

    nxp_int_poly_init_set_mpoly(a1, p1, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_int_poly_init_set_mpoly(a2, p2, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_int_poly_init(r, a1->deg + 1);
    nxp_int_poly_pseudo_div_qr(a1, a2, r, NULL);
    retval = nxp_element_mpoly_new_from_int_poly(r, -1, NXP_ELEMENT_MPOLY(p1)->args->list->data);
    nxp_int_poly_clear(a1);
    nxp_int_poly_clear(a2);
    nxp_int_poly_clear(r);
    return retval;
}

static NxpElement *pol_mul(NxpOp op, NxpElement *a, NxpElement *b, NxpEvalContext *context, GError **error)
{
    NxpElement *mona, *monb;
    NxpElement *el, *mon, *nb;
    GSList *la, *lb;
    NxpElement *vector;

    NxpElementMpoly *aa, *bb;
    GSList *l;
    NxpElement *el_mon;

    l = _element_mpoly_same_args(a, b, context, error);
    aa = l->data;
    bb = l->next->data;
    g_slist_free(l);

    el_mon = nxp_element_list_new();
    for (la = aa->mon->list; la; la = la->next) {
	for (lb = bb->mon->list; lb; lb = lb->next) {
	    mona = la->data;
	    monb = lb->data;
	    nb = nxp_element_tree_simplify_op(NXP_OP_STAR, NXP_MONOMIAL_COEFF(mona), NXP_MONOMIAL_COEFF(monb), context, error);
	    if (nxp_element_is_true(nb, context, error)) {
		vector = nxp_element_tree_simplify_op(NXP_OP_PLUS, NXP_MONOMIAL_VECTOR(mona), NXP_MONOMIAL_VECTOR(monb), context, error);
		mon = nxp_monomial_new(nb, vector);
		nxp_element_list_prepend(el_mon, mon);
		NXP_ELEMENT_UNREF(mon);
		NXP_ELEMENT_UNREF(vector);
	    }
	    NXP_ELEMENT_UNREF(nb);
	}
    }
    el = nxp_element_mpoly_new(el_mon, aa->args);
    NXP_ELEMENT_UNREF(el_mon);
    _mpoly_sort(NXP_ELEMENT_MPOLY(el));
    _element_mpoly_reduce(el, context, error);
    NXP_ELEMENT_UNREF(aa);
    NXP_ELEMENT_UNREF(bb);
    return el;
}

static NxpElement* nxp_element_mpoly_gcd(NxpElement *a, NxpElement *b, GHashTable *named_args, NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    NxpIntPoly aa[1], bb[1], dd[1];
    NxpElementInt *modulus;
    NxpElement *retval;

    modulus = nxp_named_args_get_int(named_args, "modulus", context, error);

    nxp_int_poly_init_set_mpoly(aa, a, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_int_poly_init_set_mpoly(bb, b, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    nxp_int_poly_init(dd, MIN(aa->deg, bb->deg) + 1);

    if (modulus && mpz_sgn(modulus->value) != 0) {
	if (mpz_probab_prime_p(modulus->value, 10)) {
	    nxp_int_poly_init(dd, MIN(aa->deg, bb->deg) + 1);
	    nxp_int_poly_gcd_mod(dd, aa, bb, modulus->value);
	}
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("The optional modulus argument should be a prime number or zero"));
	    nxp_int_poly_clear(aa);
	    nxp_int_poly_clear(bb);
	    nxp_int_poly_clear(dd);
	    NXP_ELEMENT_UNREF(modulus);
	    return NULL;
	}
    }
    else
	nxp_int_poly_gcd(dd, aa, bb);

    nxp_int_poly_clear(aa);
    nxp_int_poly_clear(bb);
    retval = nxp_element_mpoly_new_from_int_poly(dd, -1, NXP_ELEMENT_MPOLY(a)->args);
    nxp_int_poly_clear(dd);
    NXP_ELEMENT_XUNREF(modulus);
    return retval;
}

NxpElement* nxp_element_mpoly_to_vector(NxpElement *p, NxpElement *var,
					GHashTable *named_args,
					NxpEvalContext *context,
					GError **error)
{
    NxpElementMpoly *pol = NXP_ELEMENT_MPOLY(p);
    NxpElementVector *v;
    NxpElement *deg;
    nxp_int_t d, k, j;
    int index;
    GSList *l, *la;
    NxpElement *term, *tmp, *tmp1;
    NxpElementVector *pow;
    GError *err = NULL;

    deg = nxp_element_mpoly_deg(p, var, NULL, context, error);
    d = nxp_element_int_get_si(deg, &err);
    NXP_ELEMENT_UNREF(deg);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    if (d < 0)
	d = 0;
    v = nxp_element_vector_new(d + 1, NULL);
    la = pol->args->list;
    index = 0;
    if (d == 0)
	_element_vector_index(v, 0) = NXP_ELEMENT_REF(p);
    else {
	do {
	    if (nxp_element_eq(la->data, var, context, error)) {
		for (l = pol->mon->list; l; l = l->next) {
		    term = NXP_ELEMENT_REF(NXP_MONOMIAL_COEFF(l->data));
		    pow = NXP_MONOMIAL_VECTOR(l->data);
		    for (j = 0, la = pol->args->list; j < pow->size; j++, la = la->next) {
			if (j != index) {
			    tmp = nxp_element_tree_simplify_op(NXP_OP_HAT, la->data, _element_vector_index(pow, j), context, error);
			    tmp1 = nxp_element_tree_simplify_op(NXP_OP_STAR, term, tmp, context, error);
			    NXP_ELEMENT_UNREF(tmp);
			    NXP_ELEMENT_UNREF(term);
			    term = tmp1;
			}
			else {
			    k = nxp_element_int_get_si(_element_vector_index(pow, j), &err);
			    if (err) {
				g_propagate_error(error, err);
				NXP_ELEMENT_UNREF(v);
				NXP_ELEMENT_UNREF(term);
				return NULL;
			    }
			}
		    }
		    if (!_element_vector_index(v, k))
			_element_vector_index(v, k) = term;
		    else {
			tmp = nxp_element_tree_simplify_op(NXP_OP_PLUS,
				    _element_vector_index(v, k),
				    term, context, &err);
			if (err) {
			    g_propagate_error(error, err);
			    NXP_ELEMENT_UNREF(v);
			    NXP_ELEMENT_UNREF(term);
			    return NULL;
			}
			NXP_ELEMENT_UNREF(_element_vector_index(v, k));
			NXP_ELEMENT_UNREF(term);
			_element_vector_index(v, k) = tmp;
		    }
		}
		for (k = 0; k < v->size; k++)
		    if (!_element_vector_index(v, k))
			_element_vector_index(v, k) = nxp_element_int_new(0);
		return v;
	    }
	    else {
		index++;
		la = la->next;
	    }
	}
	while(la);
	g_assert_not_reached();
    }
    return v;
}

/************************ Polynomials' factorization *************************/

NxpElement* nxp_element_mpoly_cont(NxpElement *pol, NxpEvalContext *context, GError **error)
{
    GSList *lst;
    NxpElement *c;
    mpz_t gcd;

    lst = NXP_ELEMENT_MPOLY(pol)->mon->list;
    if (NXP_ELEMENT_IS_INT(NXP_MONOMIAL_COEFF(lst->data))) {
	mpz_init_set(gcd, NXP_ELEMENT_INT(NXP_MONOMIAL_COEFF(lst->data))->value);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("Polynomial coefficients must be integers"));
	return NULL;
    }

    lst = lst->next;
    while(lst) {
	c = NXP_MONOMIAL_COEFF(lst->data);
	if (NXP_ELEMENT_IS_INT(c)) {
	    if (mpz_sgn(NXP_ELEMENT_INT(c)->value))
		mpz_gcd(gcd, gcd, NXP_ELEMENT_INT(c)->value);
	}
	else {
	    mpz_clear(gcd);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("Polynomial coefficients must be integers"));
	    return NULL;
	}
	lst = lst->next;
    }
    return nxp_element_int_new_from_mpz(gcd);
}

static NxpElement *
list_to_product(NxpElement *list)
{
    GSList *l = NXP_ELEMENT_LIST(list)->list;
    NxpElement *tree, *tmp;

    tree = NULL;
    for (; l; l = l->next) {
	if (tree == NULL)
	    tree = NXP_ELEMENT_REF(l->data);
	else {
	    tmp = nxp_element_tree_new(NXP_OP_STAR, tree, l->data);
	    NXP_ELEMENT_UNREF(tree);
	    tree = tmp;
	}
    }
    return tree;
}

static NxpElement* nxp_element_mpoly_factor(NxpElement *pol, GHashTable *named_args,
				     NxpEvalContext *context, GError **error)
{
    int len, deg, n, i, pow;
    NxpElement *tmp2, *tmp1, *tmp, *gcd, *ppcm, *c, *power, *list = nxp_element_list_new();
    GSList *lst, *sf, *l, *l1, *factors;
    GError *err = NULL;
    mpz_t prime;
    NxpIntPoly ppol[1], ttmp[1];
    NxpIntPoly *item;

    len = nxp_element_list_length(NXP_ELEMENT(NXP_ELEMENT_MPOLY(pol)->args));
    if (len > 1)
	return NXP_ELEMENT_REF(pol);

    ppcm = nxp_element_int_new(1);
    deg = 0;
    for (lst = NXP_ELEMENT_MPOLY(pol)->mon->list; lst; lst = lst->next) {
	if (NXP_ELEMENT_IS_FRAC(NXP_MONOMIAL_COEFF(lst->data))) {
	    tmp = nxp_element_frac_get_den(NXP_MONOMIAL_COEFF(lst->data));
	    gcd = nxp_element_int_gcd(NXP_ELEMENT_INT(tmp), NXP_ELEMENT_INT(ppcm), context, error);
	    tmp1 = nxp_element_tree_simplify_op(NXP_OP_STAR, tmp, ppcm,
		    context, error);
	    NXP_ELEMENT_UNREF(tmp);
	    tmp = nxp_element_tree_simplify_op(NXP_OP_SLASH, tmp1, gcd,
		    context, error);
	    NXP_ELEMENT_UNREF(tmp1);
	    NXP_ELEMENT_UNREF(gcd);
	    NXP_ELEMENT_UNREF(ppcm);
	    ppcm = tmp;
	}
	else if (!nxp_element_number_is_integer(NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(lst->data)))) {
	    /* At the moment, we don't like polynomials with irrational
	     * coefficients */
	    NXP_ELEMENT_UNREF(ppcm);
	    return NXP_ELEMENT_REF(pol);
	}
	n = nxp_element_int_get_si(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(lst->data)), 0), &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	else if (n < 0) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("The argument is not a polynomial"));
	    return NULL;
	}
	else if (n > deg)
	    deg = n;
    }

    tmp = nxp_element_tree_simplify_op(NXP_OP_STAR, pol, ppcm, context, error);
    tmp2 = nxp_element_mpoly_cont(tmp, context, error);
    tmp1 = nxp_element_tree_simplify_op(NXP_OP_SLASH, tmp, tmp2, context, error);
    NXP_ELEMENT_UNREF(tmp);
    tmp = nxp_element_tree_simplify_op(NXP_OP_SLASH, tmp2, ppcm, context, error);
    NXP_ELEMENT_UNREF(ppcm);
    NXP_ELEMENT_UNREF(tmp2);

    /* tmp is the "generalized" content of the polynomial ; we keep it only if
     * it is different of 1 */
    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(tmp), 1) != 0)
	nxp_element_list_append(list, tmp);
    NXP_ELEMENT_UNREF(tmp);

    /* tmp1 is the primitive polynomial with integer coefficients */



    /***************** Here, begins the factorization ***********************/

    nxp_int_poly_init(ppol, deg + 1);
    for (i = 0; i <= deg; i++)
	mpz_set_ui(ppol->coeff[i], 0);
    ppol->deg = deg;

    for (lst = NXP_ELEMENT_MPOLY(tmp1)->mon->list; lst; lst = lst->next) {
	c = NXP_MONOMIAL_COEFF(lst->data);
	n = nxp_element_int_get_si(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(lst->data)), 0), &err);
	mpz_set(ppol->coeff[n], NXP_ELEMENT_INT(c)->value);
    }
    nxp_int_poly_view(ppol);
    sf = nxp_int_poly_sff(ppol);
    NXP_ELEMENT_UNREF(tmp1);

    /* sf is a list containing polynomials Pm such that :
     *	    tmp1 = p1^1*p2^2*p3^3*...*pk^k. */

    mpz_init(prime);
    nxp_int_poly_init(ttmp, ((NxpIntPoly *)sf->data)->deg + 1);
    for (l = sf, pow = 1; l; l = l->next, pow++) {
	item = l->data;
	if (item->deg > 1) {
	    mpz_set_ui(prime, 3);
	    do {
		mpz_nextprime(prime, prime);
		nxp_int_poly_normalize(ttmp, item, prime);
		if (ttmp->deg == item->deg) {
		    nxp_int_poly_derive(ttmp, item);
		    nxp_int_poly_normalize(ttmp, ttmp, prime);
		    if (ttmp->deg == item->deg - 1) {
			nxp_int_poly_gcd_mod(ttmp, item, ttmp, prime);
			if (ttmp->deg == 0 && mpz_cmp_ui(ttmp->coeff[0], 1) == 0)
			    break;
		    }
		}
	    }
	    while(TRUE);
	    l1 = nxp_int_poly_ddf(item, prime);

	    if (pow > 1)
		power = nxp_element_int_new(pow);
	    else
		power = NULL;

	    if (l1->next) { /* l1 size is greater than 1 */
		factors = nxp_int_poly_get_factors(item, l1, prime);

		for (l1 = factors; l1; l1 = l1->next) {
		    tmp1 = nxp_element_mpoly_new_from_int_poly(l1->data, -1, NXP_ELEMENT_MPOLY(pol)->args);
		    nxp_int_poly_clear(l1->data);
		    if (power) {
			tmp2 = nxp_element_tree_new(NXP_OP_HAT, tmp1, power);
			nxp_element_list_append(list, tmp2);
			NXP_ELEMENT_UNREF(tmp2);
		    }
		    else
			nxp_element_list_append(list, tmp1);
		    NXP_ELEMENT_UNREF(tmp1);
		}
		g_slist_free(factors);
	    }
	    else {
		if (item->deg != 0 || mpz_cmp_ui(item->coeff[0], 1) != 0) {
		    tmp1 = nxp_element_mpoly_new_from_int_poly(item, -1, NXP_ELEMENT_MPOLY(pol)->args);
		    if (power) {
			tmp2 = nxp_element_tree_new(NXP_OP_HAT, tmp1, power);
			nxp_element_list_append(list, tmp2);
			NXP_ELEMENT_UNREF(tmp2);
		    }
		    else
			nxp_element_list_append(list, tmp1);
		    NXP_ELEMENT_UNREF(tmp1);
		}
	    }
	    NXP_ELEMENT_XUNREF(power);
	}
	else {
	    if (item->deg != 0 || mpz_cmp_ui(item->coeff[0], 1) != 0) {
		tmp1 = nxp_element_mpoly_new_from_int_poly(item, -1, NXP_ELEMENT_MPOLY(pol)->args);
		if (pow > 1) {
		    power = nxp_element_int_new(pow);
		    tmp = nxp_element_tree_new(NXP_OP_HAT, tmp1, power);
		    NXP_ELEMENT_UNREF(power);
		}
		else
		    tmp = NXP_ELEMENT_REF(tmp1);
		NXP_ELEMENT_UNREF(tmp1);
		nxp_element_list_append(list, tmp);
		NXP_ELEMENT_UNREF(tmp);
	    }
	}
    }

    nxp_int_poly_clear(ttmp);
    nxp_int_poly_clear(ppol);
    mpz_clear(prime);
    g_slist_free(sf);
    tmp = list_to_product(list);
    NXP_ELEMENT_UNREF(list);
    return tmp;
}

#define POL_BIND(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_MPOLY, \
							 NXP_ELEMENT_TYPE_MPOLY, \
    			                                 eval, symp, 0, NULL);
#define POL_BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_MPOLY, \
   			                                 eval, symp, 0, NULL);

int nxp_element_mpoly_get_type(void)
{
    static int nxp_element_type_mpoly = 0;

    if (nxp_element_type_mpoly == 0) {
	nxp_element_type_mpoly = nxp_element_next_type();
	mpoly_typeinfo = nxp_element_type_register_info(
					    nxp_element_type_mpoly,
					    sizeof(NxpElementMpolyTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_mpoly;
}

void nxp_element_mpoly_type_init() 
{
    mpoly_typeinfo = nxp_element_type_get_info(nxp_element_mpoly_get_type());
    mpoly_typeinfo->name = g_strdup("mpoly");
    mpoly_typeinfo->destroy_func = nxp_mpoly_destroy;
    mpoly_typeinfo->eval_func = pol_eval;
    mpoly_typeinfo->simplify_func = pol_simplify;
    mpoly_typeinfo->derive_func = pol_derive;
    mpoly_typeinfo->str_func = pol_repr_func;
    mpoly_typeinfo->repr_func = pol_repr_func;
    mpoly_typeinfo->truth_func = NULL;
    mpoly_typeinfo->iterator_func = NULL;
    mpoly_typeinfo->dependson_func = pol_dependson;
    mpoly_typeinfo->substitute_func = substitute_func;

    POL_BIND(NULL, pol_add, NXP_OP_PLUS);
    POL_BIND(NULL, pol_sub, NXP_OP_MINUS);
    POL_BIND(NULL, pol_mul, NXP_OP_STAR);
    POL_BIND(NULL, pol_div, NXP_OP_SLASH);

    POL_BIND_LEFT(NULL, pol_minus, NXP_OP_MINUS_LEFT);
}

#define D(str, prot, eval, derive, simp) nxp_bifunc_define_global(str, \
                                         prot, eval, derive, simp, context, error);

void nxp_element_mpoly_function_init(NxpEvalContext *context, GError **error)
{
    D("factor", &nxp_prototype__MPOLY,
	    NULL, NULL, nxp_element_mpoly_factor);
    D("polyGcd", &nxp_prototype__MPOLY_MPOLY,
	    NULL, NULL, nxp_element_mpoly_gcd);
    D("deg", &nxp_prototype__MPOLY_VARIABLE,
	    NULL, NULL, nxp_element_mpoly_deg);
    D("polyToVector", &nxp_prototype__MPOLY_VARIABLE,
	    nxp_element_mpoly_to_vector, NULL, nxp_element_mpoly_to_vector);
}
