/* -*- Mode: C; c-file-style: "bsd"; -*- */

#include "int.h"
#include "tree.h"
#include "mpoly.h"
#include "op_bindings.h"
#include "errors.h"
#include "i18n.h"

static NxpElement *
power(NxpOp op, NxpElement *a, NxpElement *b,
      NxpEvalContext *context, GError **error)
{
    NxpElement *pol, *pol1, *res;
    int pow;

    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(b), NULL);

    if (nxp_element_number_is_integer(NXP_ELEMENT_NUMBER(b)) && nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(b), 0) >= 0) {

	pow = nxp_element_r_get_value_as_int(NXP_ELEMENT_R(b));
	pol = NXP_ELEMENT_REF(a);
	res = nxp_element_int_new(1);
	while(pow != 0) {
	    if (pow & 1) {
		pol1 = res;
		res = nxp_element_tree_simplify_op(NXP_OP_STAR, pol, pol1,
					      context, error);
		NXP_ELEMENT_UNREF(pol1);
	    }
	    pol1 = nxp_element_tree_simplify_op(NXP_OP_STAR, pol, pol,
						context, error);
	    NXP_ELEMENT_UNREF(pol);
	    pol = pol1;
	    pow >>= 1;    
	}
	NXP_ELEMENT_UNREF(pol);
    }
    else
	return nxp_element_tree_new(op, a, b);
    return res;
}

static NxpElement *
add(NxpOp op, NxpElement *a, NxpElement *b,
	NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *pol;
    GSList *l;
    NxpElement *vector, *coeff;
    NxpElement *mon;
    NxpElement *mons;
    NxpElement *retval;
    gboolean done = FALSE;
    int i;

    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(b), NULL);

    if (nxp_element_is_true(b, context, error)) {
	pol = NXP_ELEMENT_MPOLY(a);
	mons = nxp_element_list_new();
	for (l = pol->mon->list; l; l = l->next) {
	    vector = NXP_MONOMIAL_VECTOR(l->data);
	    if (nxp_element_is_true(vector, context, error))
		nxp_element_list_append(mons, NXP_ELEMENT(l->data));
	    else {
		coeff = nxp_element_tree_simplify_op(op,
						NXP_MONOMIAL_COEFF(l->data),
						b, context, error);
		mon = nxp_monomial_new(coeff, vector);
		NXP_ELEMENT_UNREF(coeff);
		nxp_element_list_append(mons, mon);
		NXP_ELEMENT_UNREF(mon);
		done = TRUE;
	    }
	}
	if (!done) {
	    int size = nxp_element_list_length(NXP_ELEMENT(pol->args));
	    NxpElement *zero = nxp_element_int_new(0);
	    vector = nxp_element_vector_new(size, NULL);
	    for (i = 0; i < size; i ++)
		_element_vector_set(NXP_ELEMENT_VECTOR(vector), i, zero);
	    NXP_ELEMENT_UNREF(zero);
	    if (op == NXP_OP_PLUS)
		mon = nxp_monomial_new(b, vector);
	    else {
		NxpElement *nb = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(b));
		mon = nxp_monomial_new(nb, vector);
		NXP_ELEMENT_UNREF(nb);
	    }
	    NXP_ELEMENT_UNREF(vector);
	    nxp_element_list_append(mons, mon);
	    NXP_ELEMENT_UNREF(mon);
	}

	retval = nxp_element_mpoly_new(mons, NXP_ELEMENT(pol->args));
	NXP_ELEMENT_UNREF(mons);
    }
    else
	retval = NXP_ELEMENT_REF(a);
    return retval;
}

static NxpElement *
sub_n_p(NxpOp op, NxpElement *a, NxpElement *b,
	NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *pol;
    GSList *l;
    NxpElement *vector, *coeff;
    NxpElement *mon;
    NxpElement *mons, *tmp;
    NxpElement *retval;
    gboolean done = FALSE;
    int i;

    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(b), NULL);

    pol = NXP_ELEMENT_MPOLY(b);
    mons = nxp_element_list_new();
    for (l = pol->mon->list; l; l = l->next) {
	vector = NXP_MONOMIAL_VECTOR(l->data);
	if (nxp_element_is_true(vector, context, error)) {
	    coeff = nxp_element_number_opposed(NXP_MONOMIAL_COEFF(l->data));
	    tmp = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(l->data));
	    nxp_element_list_append(mons, tmp);
	    NXP_ELEMENT_UNREF(coeff);
	    NXP_ELEMENT_UNREF(tmp);
	}
	else {
	    coeff = nxp_element_tree_simplify_op(op,
					    a, NXP_MONOMIAL_COEFF(l->data),
					    context, error);
	    mon = nxp_monomial_new(coeff, vector);
	    NXP_ELEMENT_UNREF(coeff);
	    nxp_element_list_append(mons, mon);
	    NXP_ELEMENT_UNREF(mon);
	    done = TRUE;
	}
    }
    if (!done) {
	int size = nxp_element_list_length(NXP_ELEMENT(pol->args));
	NxpElement *zero = nxp_element_int_new(0);
	vector = nxp_element_vector_new(size, NULL);
	for (i = 0; i < size; i ++)
	    _element_vector_set(NXP_ELEMENT_VECTOR(vector), i, zero);
	NXP_ELEMENT_UNREF(zero);
	mon = nxp_monomial_new(a, vector);
	NXP_ELEMENT_UNREF(vector);
	nxp_element_list_append(mons, mon);
	NXP_ELEMENT_UNREF(mon);
    }

    retval = nxp_element_mpoly_new(mons, NXP_ELEMENT(pol->args));
    NXP_ELEMENT_UNREF(mons);
    return retval;
}

static NxpElement *
mul(NxpOp op, NxpElement *a, NxpElement *b,
	NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *pol;
    GSList *l;
    NxpElement *coeff;
    NxpElement *mon;
    NxpElement *mons;
    NxpElement *retval;

    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(b), NULL);

    if (nxp_element_is_true(b, context, error)) {
	pol = NXP_ELEMENT_MPOLY(a);
	mons = nxp_element_list_new();
	for (l = pol->mon->list; l; l = l->next) {
	    coeff = NXP_MONOMIAL_COEFF(l->data);
	    coeff = nxp_element_tree_simplify_op(op, coeff, b, context, error);
	    mon = nxp_monomial_new(coeff, NXP_MONOMIAL_VECTOR(l->data));
	    NXP_ELEMENT_UNREF(coeff);
	    
	    nxp_element_list_append(mons, mon);
	    NXP_ELEMENT_UNREF(mon);
	}
	retval = nxp_element_mpoly_new(mons, NXP_ELEMENT(pol->args));
	NXP_ELEMENT_UNREF(mons);
    }
    else
	retval = NXP_ELEMENT_REF(b);
    return retval;
}

static NxpElement *div_n_p(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElementMpoly *bb = NXP_ELEMENT_MPOLY(b);
    NxpElement *el, *monb, *mon, *nb, *vector;

    if (nxp_element_number_compare_int(a, 0) == 0)
	return NXP_ELEMENT_REF(a);

    if (nxp_element_list_length(NXP_ELEMENT(bb->mon)) == 1) {
	monb = bb->mon->list->data;
	nb = nxp_element_tree_simplify_op(NXP_OP_SLASH, a,
				    NXP_MONOMIAL_COEFF(monb), context, error);
	vector = nxp_element_tree_simplify_op(NXP_OP_MINUS_LEFT, NULL,
				    NXP_MONOMIAL_VECTOR(monb), context, error);
	mon = nxp_monomial_new(nb, vector);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(nb);
	el = nxp_element_mpoly_new_simple(mon, NXP_ELEMENT(bb->args));
	NXP_ELEMENT_UNREF(mon);
	return el;
    }
    else
	return nxp_element_tree_new(op, a, b);
}

#define BIND_SYM(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_NUMBER, \
   			                              eval, symp,  NXP_OP_BIND_SYMMETRIC, NULL);

#define BIND_P_N(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_NUMBER, \
   			                              eval, symp,  0, NULL);

#define BIND_N_P(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_NUMBER, NXP_ELEMENT_TYPE_MPOLY, \
   			                              eval, symp,  0, NULL);

void nxp_element_mpoly_number_operator_init(void)
{
    BIND_SYM(NULL, add, NXP_OP_PLUS);
    BIND_P_N(NULL, add, NXP_OP_MINUS);
    BIND_SYM(NULL, add, NXP_OP_PLUS);
    BIND_P_N(NULL, add, NXP_OP_MINUS);
    BIND_SYM(NULL, mul, NXP_OP_STAR);
    BIND_P_N(NULL, mul, NXP_OP_SLASH);
    BIND_N_P(NULL, div_n_p, NXP_OP_SLASH);
    BIND_N_P(NULL, sub_n_p, NXP_OP_MINUS);
    BIND_P_N(NULL, power, NXP_OP_HAT);
}
