/* -*- Mode: C; c-file-style: "bsd"; -*- */

#include <string.h>
#include "nxp_string.h"
#include "element_type.h"
#include "types_misc.h"
#include "op_bindings.h"
#include "int.h"
#include "tree.h"

#include "parser/plugins.h"

#define BIND(func,op)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_STRING, NXP_ELEMENT_TYPE_STRING, 	\
  				func,  			\
  				func, 			\
  				0, 					\
				NULL); 


static NxpElementTypeInfo *typeinfo = NULL;


static NxpElement* nxp_parse_string(const gchar *string) 
{
    return nxp_element_string_new(string);
}



/**
 * nxp_element_string_new:
 * @string: value of element; will _not_ keep this string
 * 
 * Create a new NxpElementString
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_string_new(const char *string)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementString));
    NXP_ELEMENT_STRING(elem)->string = g_strdup(string);
    return elem;
}


/**
 * nxp_element_string_new:
 * @string: value of element; will take ownership of this string
 * 
 * Create a new NxpElementString
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_string_new_with_str(char *string)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementString));
    NXP_ELEMENT_STRING(elem)->string = string;
    return elem;
}


static GString* str_func(NxpElement *elem, int *prec)
{
    NxpElementString *string = NXP_ELEMENT_STRING(elem);
    GString *gstr;
    if (prec) *prec = 0;
    gstr = g_string_new("\"");
    gstr = g_string_append(gstr, string->string);
    gstr = g_string_append_c(gstr, '"');
    return gstr;
}


static NxpElement* substitute_func(NxpElement *self,
				   NxpElement *var, NxpElement *el,
				   NxpEvalContext *context, GError **error)
{
    return NXP_ELEMENT_REF(self);
}

static void destroy_func(NxpElement *elem)
{
    g_free(NXP_ELEMENT_STRING(elem)->string);
}

static NxpElement* sum(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error)
{
    GString *aa;
    NxpElement *el;
    aa = g_string_new(NXP_ELEMENT_STRING(a)->string);
    aa = g_string_append(aa, NXP_ELEMENT_STRING(b)->string);
    el = nxp_element_string_new_with_str(aa->str);
    g_string_free(aa, FALSE);
    return el;
}

static NxpElement* gt(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) > 0);
}

static NxpElement* ge(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) >= 0);
}

static NxpElement* lt(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) < 0);
}

static NxpElement* le(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) <= 0);
}

static NxpElement* eq(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) == 0);
}

static NxpElement* ne(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new(strcmp(NXP_ELEMENT_STRING(a)->string,
				      NXP_ELEMENT_STRING(b)->string) != 0);
}

static NxpElement* and(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error)
{
    int aa = strlen(NXP_ELEMENT_STRING(a)->string);
    int bb = strlen(NXP_ELEMENT_STRING(b)->string);
    return nxp_element_int_new(aa?bb:aa);
}

static NxpElement* or(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error)
{
    int aa = strlen(NXP_ELEMENT_STRING(a)->string);
    int bb = strlen(NXP_ELEMENT_STRING(b)->string);
    return nxp_element_int_new(aa?aa:bb);
}

int nxp_element_string_get_type(void)
{
    static int nxp_element_type_string = 0;

    if (nxp_element_type_string == 0) {
	nxp_element_type_string = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_string,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_string;
}

void nxp_element_string_type_init()
{
    typeinfo = nxp_element_type_get_info(nxp_element_string_get_type());
    typeinfo->name = g_strdup("string");
    typeinfo->eval_func = nxp_eval_return_self;
    typeinfo->simplify_func = nxp_eval_return_self;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->destroy_func = destroy_func;
    typeinfo->substitute_func = substitute_func;

    nxp_parser_register_func(NXP_ATOM_CLASS_STRING, nxp_parse_string, 50);

    BIND(sum,NXP_OP_PLUS);

    // BIND(...,NXP_OP_MINUS);
    // BIND(...,NXP_OP_MINUS_LEFT);
    // BIND(...,NXP_OP_STAR);
    // BIND(...,NXP_OP_SLASH);
    // BIND(...,NXP_OP_HAT);
    // BIND(...,NXP_OP_DOT3);
    // BIND(...,NXP_OP_PLUS2_LEFT);
    // BIND(...,NXP_OP_PLUS2_RIGHT);
    // BIND(...,NXP_OP_MINUS2_LEFT);
    // BIND(...,NXP_OP_MINUS2_RIGHT);

    BIND(gt,NXP_OP_GT);
    BIND(ge,NXP_OP_GT_EQ);
    BIND(lt,NXP_OP_LT);
    BIND(le,NXP_OP_LT_EQ);
    BIND(eq,NXP_OP_EQ2);
    BIND(ne,NXP_OP_EXCL_EQ);

    BIND(and, NXP_OP_SLASH_BSLASH);
    BIND(or,  NXP_OP_BSLASH_SLASH);

    // BIND(...,NXP_OP_TILDE_LEFT);
    // BIND(...,NXP_OP_EXCL_RIGHT);

    // BIND(...,NXP_OP_EQ);
    // BIND(...,NXP_OP_COLON_EQ);
}



