/* -*- Mode: C; c-file-style: "soustroup"; -*- */
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "number.h"
#include "int.h"
#include "derive.h"
#include "element_type.h"
#include "element_output.h"
#include "errors.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

/* This function tries to simplify the unary minus when it is applied
 * to a int, real or complex value.
 */
NxpElement* elem_UMINUS(NxpElement* el) {
    if (NXP_ELEMENT_IS_NUMBER(el))
	return nxp_element_number_opposed(NXP_ELEMENT_NUMBER(el));
    else
	return elem_UNARY(NXP_OP_MINUS_LEFT, NULL, el);
}

/**
 * nxp_element_tree_new:
 * @op: the operator (#NxpOp) of the tree
 * @left: left branch of the binary tree
 * @right: right branch of the binary tree
 * 
 * Create an NxpElementTree element (binary tree)
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_tree_new(NxpOp op, NxpElement *left, NxpElement *right)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementTree));
    NXP_ELEMENT_TREE(elem)->left = left? NXP_ELEMENT_REF(left): NULL;
    NXP_ELEMENT_TREE(elem)->right = right? NXP_ELEMENT_REF(right): NULL;
    NXP_ELEMENT_TREE(elem)->op = op;

    return elem;
}


gboolean
nxp_element_eq(NxpElement * a, NxpElement * b,
	       NxpEvalContext * context, GError ** error)
{
    NxpElement     *ret;

    ret = nxp_element_tree_simplify_op(NXP_OP_EQ2, a, b, context, error);
    if (ret && NXP_ELEMENT_IS_NUMBER(ret)
	&& nxp_element_is_true(ret, context, error)) {
	NXP_ELEMENT_UNREF(ret);
	return TRUE;
    } else {
	NXP_ELEMENT_XUNREF(ret);
	return FALSE;
    }
}


static NxpElement* eval_unary_right_op(NxpOp op, NxpElement *elem, 
				       NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type;
    NxpElement *ret, *evaled;
    NxpElement *target = NULL;
    GError *error1 = NULL;

    g_return_val_if_fail(elem, NULL);
    
    evaled = NULL;
    ret = NULL;
    
    type = NXP_ELEMENT_GET_TYPE(elem);
    
    binding = nxp_get_operator_binding(op, type, NXP_ELEMENT_TYPE_NONE, NULL);
    if (binding) {
	target = elem;
    } else {
	binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_ANY,
					   NXP_ELEMENT_TYPE_NONE, NULL);
	if (binding) {
	    target = elem;
	} else {
	    evaled = nxp_element_eval(elem, context, &error1);
	    if (error1) {
		g_propagate_error(error, error1);
		NXP_ELEMENT_XUNREF(evaled);
		return NULL;
	    }
	    g_assert(evaled);
	    binding = nxp_get_operator_binding(op, NXP_ELEMENT_GET_TYPE(evaled),
					       NXP_ELEMENT_TYPE_NONE,
					       NULL);
	    if (binding) {
		target = evaled;
	    } 
	}
    }
    if (!binding) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("operation undefined or not implemented for %s %s %s"),
		    NXP_ELEMENT_GET_TYPE_NAME(elem),
		    nxp_op_get_info(op)->token, "");
	NXP_ELEMENT_XUNREF(evaled);
	return NULL;
    }
    if (!binding->eval_func) {
	g_warning("bug: eval not implemented for %s %s.",
		  NXP_ELEMENT_GET_TYPE_NAME(elem), 
		  nxp_op_get_info(op)->token);
	ret = NULL;
    } else {
	    ret = binding->eval_func(op, target, NULL, context, error);
    }
    NXP_ELEMENT_XUNREF(evaled);
    return ret;
}

static NxpElement* eval_unary_left_op(NxpOp op, NxpElement *elem,
				      NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type;
    NxpElement *ret, *evaled;
    NxpElement *target = NULL;
    GError *error1 = NULL;

    g_return_val_if_fail(elem, NULL);
    
    evaled = NULL;
    ret = NULL;

    type = NXP_ELEMENT_GET_TYPE(elem);

    binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_NONE, type, NULL);
    if (binding) {
	target = elem;
    } else {
	binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_NONE, NXP_ELEMENT_TYPE_ANY, NULL);
	if (binding) {
	    target = elem;
	} else {
	    evaled = nxp_element_eval(elem, context, &error1);
	    if (error1) {
		g_propagate_error(error, error1);
		NXP_ELEMENT_XUNREF(evaled);
		return NULL;
	    }
	    g_assert(evaled);
	    binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_NONE,
					       NXP_ELEMENT_GET_TYPE(evaled),
					       NULL);
	    if (binding) {
		target = evaled;
	    } 
	}
    }
    if (!binding) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("operation undefined or not implemented for %s %s %s"),
		    nxp_op_get_info(op)->token,
		    NXP_ELEMENT_GET_TYPE_NAME(elem), "");
	NXP_ELEMENT_XUNREF(evaled);
	return NULL;
    }
    if (!binding->eval_func) {
	g_warning("bug: eval not implemented for %s %s.",
		  nxp_op_get_info(op)->token,
		  NXP_ELEMENT_GET_TYPE_NAME(elem));
	ret = NULL;
    } else {
	    ret = binding->eval_func(op, NULL, target, context, error);
    }
    NXP_ELEMENT_XUNREF(evaled);
    return ret;
}


static NxpElement* eval_binary_op(NxpOp op, NxpElement *left, NxpElement *right,
				  NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type1, type2;
    NxpElement *ret, *left_eval, *right_eval;
    NxpElement *target_left = NULL;
    NxpElement *target_right = NULL;
    GError *error1 = NULL;

    g_return_val_if_fail(left && right, NULL);
    
    left_eval = NULL;
    right_eval = NULL;
    ret = NULL;

    type1 = NXP_ELEMENT_GET_TYPE(left);
    type2 = NXP_ELEMENT_GET_TYPE(right);

    binding = nxp_get_operator_binding(op, type1, type2, error);

    if (binding && binding->eval_func) {
	target_left = left;
	target_right = right;
    } else {
	binding = nxp_get_operator_binding(op, type1, NXP_ELEMENT_TYPE_ANY, error);
	if (binding && binding->eval_func) {
//	    g_warning("The second type is any !!\nI don't want it anymore !!\n");
	    target_left = left;
	    target_right = right;
	} else {
	    binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_ANY, type2, error);
	    if (binding && binding->eval_func) {
//		g_warning("The first type is any !!\nI don't want it anymore !!\n");
		target_left = left;
		target_right = right;
	    } else {
		right_eval = nxp_element_eval(right, context, &error1);
		if (error1) {
		    g_propagate_error(error, error1);
		    NXP_ELEMENT_XUNREF(right_eval);
		    return NULL;
		}
		type2 = NXP_ELEMENT_GET_TYPE(right_eval);
//		g_warning("the second object is evaled\n");
		binding = nxp_get_operator_binding(op, type1, type2, error);
		if (binding && binding->eval_func) {
		    target_left = left;
		    target_right = right_eval;
		} else {
		    left_eval = nxp_element_eval(left, context, &error1);
//		    g_warning("the first object is evaled\n");
		    if (error1) {
			g_propagate_error(error, error1);
			NXP_ELEMENT_XUNREF(left_eval);
			NXP_ELEMENT_XUNREF(right_eval);
			return NULL;
		    }
		    type1 = NXP_ELEMENT_GET_TYPE(left_eval);
		    binding = nxp_get_operator_binding(op, type1, type2, error);
		    if (binding && binding->eval_func) {
			target_left = left_eval;
			target_right = right_eval;
		    } else {
			binding = nxp_get_operator_binding(op, type1,
							   NXP_ELEMENT_TYPE_ANY,
							   error);
			if (binding && binding->eval_func) {
//			    g_warning("The second type is any !!\nI don't want it anymore !!\n");
			    target_left = left_eval;
			    target_right = right_eval;
			} else {
			    binding = nxp_get_operator_binding(op,
					    		NXP_ELEMENT_TYPE_ANY,
							type2, error);
			    if (binding && binding->eval_func) {
//				g_warning("The first type is any !!\nI don't want it anymore !!\n");
				target_left = left_eval;
				target_right = right_eval;
			    }
			}
		    }
		}
	    }
	}
    }
    if (!binding) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("operation undefined or not implemented for %s %s %s"),
		    type1? NXP_ELEMENT_GET_TYPE_NAME(left) : "",
		    nxp_op_get_info(op)->token,
		    type2? NXP_ELEMENT_GET_TYPE_NAME(right) : "");
	NXP_ELEMENT_XUNREF(left_eval);
	NXP_ELEMENT_XUNREF(right_eval);
	return NULL;
    }
    if (!binding->eval_func) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("operation undefined or not implemented for %s %s %s"),
		    type1? NXP_ELEMENT_GET_TYPE_NAME(left) : "",
		    nxp_op_get_info(op)->token,
		    type2? NXP_ELEMENT_GET_TYPE_NAME(right) : "");
	ret = NULL;
    } else {
	if (binding->flags & NXP_OP_BINDING_REVERSE)
	    ret = binding->eval_func(op, target_right, target_left, context, error);
	else
	    ret = binding->eval_func(op, target_left, target_right, context, error);
    }
    NXP_ELEMENT_XUNREF(left_eval);
    NXP_ELEMENT_XUNREF(right_eval);
    return ret;
}


static NxpElement* derive_binary_op(NxpOp op, NxpIdentifier *id,
				    NxpElement *left, NxpElement *right, 
				    NxpEvalContext *context, GError **error)
{
    NxpOpDeriveFunc derive_func;
    NxpElementType type1, type2;
    NxpElement *ret, *left_eval, *right_eval;

    g_return_val_if_fail(left && right, NULL);
    
    left_eval = NULL;
    right_eval = NULL;
    ret = NULL;

    type1 = NXP_ELEMENT_GET_TYPE(left);
    type2 = NXP_ELEMENT_GET_TYPE(right);

    derive_func = get_derive_function(op);

    if (derive_func) {
	return derive_func(op,id,left,right, context, NULL);
    } else {
	g_set_error(error, nxp_error_domain, 
		    NXP_ERROR_NO_DERIVATION, 
		    _("Operator '%s' can't be derived!"),
		    (nxp_op_get_info(op))->token);
	return NULL;
    }
}

static NxpElement* derive_unary_left_op(NxpOp op, 
					NxpIdentifier *id,
					NxpElement *right, 
					NxpEvalContext *context,
					GError **error)
{
    NxpOpDeriveFunc derive_func;
    NxpElementType typer;
    NxpElement *ret, *right_eval;

    g_return_val_if_fail(right, NULL);
    
    right_eval = NULL;
    ret = NULL;

    typer = NXP_ELEMENT_GET_TYPE(right);

    derive_func = get_derive_function(op);

    if (derive_func) {
	return derive_func(op, id, NULL, right, context, NULL);
    } else {
	printf("Don't know\n");
    }

    return NULL;

}

/**
 * nxp_element_tree_eval_op: -eval an imaginary tree
 * @op: operator
 * @left: left branch
 * @right: right branch
 * @error: NXP_ERROR_NOT_IMPLEMENTED if no operator binding is found
 * 
 * Invoke the appropriate function to evaluate an imaginary tree
 * composed by the operator @op with @left and @right as branches.
 * 
 * Return value: the result
 **/
NxpElement* nxp_element_tree_eval_op(NxpOp op,
				     NxpElement *left,
				     NxpElement *right,
				     NxpEvalContext *context,
				     GError **error)
{
    switch (nxp_op_get_info(op)->kind)
    {
    case NXP_OP_BINARY:
	return eval_binary_op(op, left, right, context, error);
    case NXP_OP_UNARY_LEFT:
	return eval_unary_left_op(op, right, context, error);
    case NXP_OP_UNARY_RIGHT:
	return eval_unary_right_op(op, left, context, error);
    default:
	g_assert_not_reached();
	return NULL;
    }
}

NxpElement* nxp_element_tree_derive_op(NxpOp op, 
				       NxpIdentifier* id,
				       NxpElement *left,
				       NxpElement *right, 
				       NxpEvalContext *context, 
				       GError **error)
{
    switch (nxp_op_get_info(op)->kind) 
    {
    case NXP_OP_BINARY:
	return derive_binary_op(op, id, left, right, context, error);
    case NXP_OP_UNARY_LEFT:
	return derive_unary_left_op(op, id, right, context, error);
    case NXP_OP_UNARY_RIGHT:
	return NULL;
	//return eval_unary_right_op(op, left, error);
    default:
	g_assert_not_reached();
	return NULL;
    }
}


static NxpElement* eval(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(elem, NULL);

    return nxp_element_tree_eval_op(NXP_ELEMENT_TREE(elem)->op,
				  NXP_ELEMENT_TREE(elem)->left,
				  NXP_ELEMENT_TREE(elem)->right,
				  context, error);
}

static NxpElement* derive(NxpElement *elem, NxpIdentifier *id, NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    g_return_val_if_fail(elem, NULL);
    retval = nxp_element_tree_derive_op(NXP_ELEMENT_TREE(elem)->op,
					id,
					NXP_ELEMENT_TREE(elem)->left,
					NXP_ELEMENT_TREE(elem)->right,
					context, error);
    return retval;
}

static void destroy(NxpElement *elem)
{
    if (NXP_ELEMENT_TREE(elem)->left)
	NXP_ELEMENT_UNREF(NXP_ELEMENT_TREE(elem)->left);
    if (NXP_ELEMENT_TREE(elem)->right)
	NXP_ELEMENT_UNREF(NXP_ELEMENT_TREE(elem)->right);
}


/**
 * str_or_repr_func:
 * @element: 
 * @precedence: 
 * @str_or_repr: 1->str 2->repr 
 * 
 * 
 * 
 * Return value: 
 **/
static inline GString* str_or_repr_func(NxpElement *element, int *precedence, gint str_or_repr)
{
    GString *str_tree, *str_left, *str_right;
    gint prec_tree, prec_left = -1000, prec_right = -1000;
    NxpElementTree *tree = NXP_ELEMENT_TREE(element);
    const NxpOpInfo *opinfo;
    NxpElementTypeInfo *typeinfo;

    opinfo = nxp_op_get_info(tree->op);
    prec_tree = opinfo->precedence;
    str_tree = g_string_new(NULL);

    if (tree->left)
    {
	typeinfo = NXP_ELEMENT_GET_TYPE_INFO(tree->left);
	if (str_or_repr == 1) {
	    if (typeinfo->str_func)
		str_left = typeinfo->str_func(tree->left, &prec_left);
	    else 
		str_left = g_string_new("??");
	} else {
	    if (typeinfo->repr_func)
		str_left = typeinfo->repr_func(tree->left, &prec_left);
	    else
		str_left = g_string_new("??");
	}
	if (prec_tree < prec_left)	// <=
	    str_tree = g_string_append(str_tree, "(");
	str_tree = g_string_append(str_tree, str_left->str);
	if (prec_tree < prec_left)	// <=
	    str_tree = g_string_append(str_tree, ")");
	g_string_free(str_left, TRUE);
    }

    if (opinfo->kind == NXP_OP_BINARY || opinfo->kind == NXP_OP_UNARY_LEFT)
	str_tree = g_string_append_c(str_tree, ' ');
    str_tree = g_string_append(str_tree, opinfo->token);
//    if (opinfo->kind == NXP_OP_BINARY || opinfo->kind == NXP_OP_UNARY_RIGHT)
	str_tree = g_string_append_c(str_tree, ' ');

    if (tree->right)
    {
	typeinfo = NXP_ELEMENT_GET_TYPE_INFO(tree->right);
	if (str_or_repr == 1) {
	    if (typeinfo->str_func)
		str_right = typeinfo->str_func(tree->right, &prec_right);
	    else
		str_right = g_string_new("??");
	} else {
	    if (typeinfo->repr_func)
		str_right = typeinfo->repr_func(tree->right, &prec_right);
	    else
		str_right = g_string_new("??");
	}
	if (prec_tree <= prec_right)
	    str_tree = g_string_append(str_tree, "(");
	str_tree = g_string_append(str_tree, str_right->str);
	if (prec_tree <= prec_right)
	    str_tree = g_string_append(str_tree, ")");
	g_string_free(str_right, TRUE);
    }
    if (precedence) *precedence = prec_tree;
    return str_tree;
}


static GString* str_func(NxpElement *element, int *precedence)
{
    return str_or_repr_func(element, precedence, 1);   
}


static GString* repr_func(NxpElement *element, int *precedence)
{
    return str_or_repr_func(element, precedence, 2);
}

static NxpElement* nxp_element_tree_reduce(NxpOp op, NxpElement *left, NxpElement *right, NxpEvalContext *context, GError **error)
{
    switch (op) {
    case NXP_OP_STAR:
	if (NXP_ELEMENT_IS_NUMBER(left) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(left), 0))
	    return nxp_element_int_new(0);
	else if (NXP_ELEMENT_IS_NUMBER(right) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(right), 0))
	    return nxp_element_int_new(0);
	else if (NXP_ELEMENT_IS_NUMBER(right) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(right), 1))
	    return NXP_ELEMENT_REF(left);
	else if (NXP_ELEMENT_IS_NUMBER(left) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(left), 1))
	    return NXP_ELEMENT_REF(right);
	else
	    return nxp_element_tree_new(op, left, right);

    case NXP_OP_PLUS:
	if (NXP_ELEMENT_IS_NUMBER(left) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(left), 0))
	    return NXP_ELEMENT_REF(right);
	else if (NXP_ELEMENT_IS_NUMBER(right) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(right), 0))
	    return NXP_ELEMENT_REF(left);
	else
	    return nxp_element_tree_new(op, left, right);

    case NXP_OP_MINUS:
	if (NXP_ELEMENT_IS_NUMBER(left) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(left), 0))
	    return nxp_element_tree_new(NXP_OP_MINUS_LEFT, NULL, right);
	else if (NXP_ELEMENT_IS_NUMBER(right) && !nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(right), 0))
	    return NXP_ELEMENT_REF(left);
	else
	    return nxp_element_tree_new(op, left, right);
    default:
	return nxp_element_tree_new(op, left, right);
    }
}

	
static NxpElement* simplify_binary_op(NxpOp op, NxpElement *left, NxpElement *right,
				      NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type1, type2;
    NxpElement *ret;

    g_return_val_if_fail(left && right, NULL);

    type1 = NXP_ELEMENT_GET_TYPE(left);
    type2 = NXP_ELEMENT_GET_TYPE(right);

    binding = nxp_get_operator_binding(op, type1, type2, error);

    if (!binding) { 
	binding = nxp_get_operator_binding(op, type1, NXP_ELEMENT_TYPE_ANY, error); 
	if (!binding) { 
	    binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_ANY, type2, error);
	    if (!binding) {
		ret = nxp_element_tree_reduce(op, left, right, context, error);
		return ret;
	    }
	}
    }

    if (!binding->simp_func)
	ret = nxp_element_tree_reduce(op, left, right, context, error);
    else {
		if (binding->flags & NXP_OP_BINDING_REVERSE) 
			ret = binding->simp_func(op, right, left, context, error);
		else 
	    	ret = binding->simp_func(op, left, right, context, error);
    }
    return ret; 
}


static NxpElement* simplify_unary_left_op(NxpOp op, NxpElement *left, NxpElement *right,
					  NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type;
    NxpElement *ret; //, *right_simp;

    g_return_val_if_fail(right, NULL);

/*    // This is to go off, someday
    right_simp = nxp_element_simplify(right, context, NULL);
    if (!right_simp) return nxp_element_tree_new(op, left, right);*/

    type = NXP_ELEMENT_GET_TYPE(right); //right_simp

    binding = nxp_get_operator_binding(op, 0, type, error);

    if (!binding) {
  	binding = nxp_get_operator_binding(op, 0, NXP_ELEMENT_TYPE_ANY, error); 
  	if (!binding) { 
	    ret = nxp_element_tree_new(op, NULL, right); //right_simp
	    //NXP_ELEMENT_UNREF(right_simp);
	    return ret;
	}
    }

    if (!binding->simp_func)
	ret = nxp_element_tree_new(op, NULL, right); //right_simp
    else
	ret = binding->simp_func(op, NULL, right, context, error); //right_simp

    //NXP_ELEMENT_UNREF(right_simp);
    return ret; 
}


static NxpElement* simplify_unary_right_op(NxpOp op, NxpElement *left, NxpElement *right,
					   NxpEvalContext *context, GError **error)
{
    NxpOpBinding *binding;
    NxpElementType type;
    NxpElement *ret; //, *left_simp;

    g_return_val_if_fail(left, NULL);

    /* // This is to go off, someday
    left_simp = nxp_element_simplify(left,context, NULL);
    if (!left_simp) return nxp_element_tree_new(op, left, right);*/

    type = NXP_ELEMENT_GET_TYPE(left); //left_simp

    binding = nxp_get_operator_binding(op, type, 0, error);

    if (!binding) { 
  	binding = nxp_get_operator_binding(op, NXP_ELEMENT_TYPE_ANY, 0, error); 
  	if (!binding) { 
	    ret = nxp_element_tree_new(op, left, NULL); //left_simp
	    //NXP_ELEMENT_UNREF(left_simp);
	    return ret;
	}
    }

    if (!binding->simp_func)
	ret = nxp_element_tree_new(op, left, NULL); //left_simp
    else
	ret = binding->simp_func(op, left, NULL, context, error);  //left_simp

    //NXP_ELEMENT_UNREF(left_simp);
    return ret; 
}

NxpElement* nxp_element_tree_simplify_op(NxpOp op, 
					 NxpElement *left,
					 NxpElement *right, 
					 NxpEvalContext *context, 
					 GError **error)
{
    switch (nxp_op_get_info(op)->kind) 
    {
    case NXP_OP_BINARY:
	return simplify_binary_op(op, left, right, context, error);
    case NXP_OP_UNARY_LEFT:
	return simplify_unary_left_op(op, left, right, context, error);
    case NXP_OP_UNARY_RIGHT:
	return simplify_unary_right_op(op, left, right, context, error);
    default:
	g_assert_not_reached();
	return NULL;
    }
}


static NxpElement* substitute(NxpElement *elem, NxpElement *var, NxpElement *el,
			      NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElement *left = NULL, *right = NULL;
    GError *err = NULL;

    g_return_val_if_fail(elem, NULL);

    if (NXP_ELEMENT_TREE(elem)->left) {
	left = nxp_element_substitute(NXP_ELEMENT_TREE(elem)->left, var,
				      el, context, &err);
	if (err) {
	    g_warning("Erreur in tree\n");
	    g_propagate_error(error, err);
	    return NULL;
	}
    }
    if (NXP_ELEMENT_TREE(elem)->right) {
	right = nxp_element_substitute(NXP_ELEMENT_TREE(elem)->right, var,
				       el, context, &err);
	if (err) {
	    g_warning("Erreur in tree\n");
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(left);
	    return NULL;
	}
    }
    retval = nxp_element_tree_new(NXP_ELEMENT_TREE(elem)->op, left, right);
    NXP_ELEMENT_XUNREF(left);
    NXP_ELEMENT_XUNREF(right);
    return retval;
}

static NxpElement* simplify(NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElement *e, *e1, *rt;
    NxpElement *left_simp, *right_simp;
    GError *err = NULL;
    g_return_val_if_fail(el, NULL);

    rt = NULL;

    if (NXP_ELEMENT_TREE(el)->op == NXP_OP_COLON_EQ || NXP_ELEMENT_TREE(el)->op == NXP_OP_EQ) {
	    left_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->left);
	    right_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->right);
    }
    else if (NXP_ELEMENT_TREE(el)->op == NXP_OP_PLUS2_LEFT) {
	left_simp = NULL;
	right_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->right);
    }
    else if (NXP_ELEMENT_TREE(el)->op == NXP_OP_MINUS2_LEFT) {
	left_simp = NULL;
	right_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->right);
    }
    else {
	if (NXP_ELEMENT_TREE(el)->left) {
	    left_simp = nxp_element_simplify(NXP_ELEMENT_TREE(el)->left, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_XUNREF(left_simp);
		return NULL;
		//g_warning("cannot simplify left element\n");
		//left_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->left);
		//g_error_free(err);
		//err = NULL;
	    }
	}
	else
	    left_simp = NULL;

	if (NXP_ELEMENT_TREE(el)->right) {
	    right_simp = nxp_element_simplify(NXP_ELEMENT_TREE(el)->right, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_XUNREF(left_simp);
		NXP_ELEMENT_XUNREF(right_simp);
		return NULL;
		//g_warning("cannot simplify right element\n");
		//right_simp = NXP_ELEMENT_REF(NXP_ELEMENT_TREE(el)->right);
		//g_error_free(err);
		//err = NULL;
	    }
	}
	else
	    right_simp = NULL;
    }

    e = nxp_element_tree_simplify_op(NXP_ELEMENT_TREE(el)->op,
				     left_simp, right_simp,
				     context,
				     &err);
    if (err) {
	g_propagate_error(error, err);
	NXP_ELEMENT_XUNREF(left_simp);
	NXP_ELEMENT_XUNREF(right_simp);
	NXP_ELEMENT_XUNREF(e);
	return NULL;
    }

    NXP_ELEMENT_XUNREF(left_simp);
    NXP_ELEMENT_XUNREF(right_simp);

    if (e) {
	if (NXP_ELEMENT_GET_TYPE(e) == NXP_ELEMENT_TYPE_TREE) {
	    e1 = NXP_ELEMENT_REF(e);

//	    e1 = nxp_element_tree_simplify_tree(NXP_ELEMENT_TREE(e),
//						context, error);
	    g_assert(*error == NULL);
	    if (e1) {
		NXP_ELEMENT_XUNREF(e);
		rt = e1;
	    }
	}
	else
	    rt = e;
    }
    else
	rt = NXP_ELEMENT_REF(el);
    return rt;
}

static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement* el;
    gboolean result;
    GString *str;
    el = nxp_element_simplify(elem, context, error);
    if (el != elem && !NXP_ELEMENT_IS_TREE(el)) {
	/* Could I remove this condition ?????? */
	g_print ("truth_func : positive value : %s\n", __FILE__);
	result = nxp_element_is_true(el, context, error);
	NXP_ELEMENT_UNREF(el);
	return result;
    }
    NXP_ELEMENT_UNREF(el);
    str = nxp_element_repr(elem);
    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDECIDABLE, _("Cannot decide if %s is true."), str->str);
    g_string_free(str, TRUE);
    return FALSE;
}

static NxpElement* eq(NxpOp op,
		      NxpElement *a, NxpElement *b,
		      NxpEvalContext *context,
		      GError **error) {
    NxpElementTree* aa = NXP_ELEMENT_TREE(a);
    NxpElementTree* bb = NXP_ELEMENT_TREE(b);
    NxpElement *ret1 = NULL, *ret2 = NULL;

    if (aa->op == bb->op) {
	switch (nxp_op_get_info(aa->op)->kind)
	{
	case NXP_OP_BINARY:
	    ret1 = nxp_element_tree_simplify_op(NXP_OP_EQ2, aa->left, bb->left, context, error);
	    g_assert(*error == NULL);
	    ret2 = nxp_element_tree_simplify_op(NXP_OP_EQ2, aa->right, bb->right, context, error);
	    g_assert(*error == NULL);
	    if (ret1 && ret2 
		    && NXP_ELEMENT_GET_TYPE(ret1) == NXP_ELEMENT_TYPE_INT
		    && NXP_ELEMENT_INT(ret1)->value
		    && NXP_ELEMENT_GET_TYPE(ret2) == NXP_ELEMENT_TYPE_INT
		    && NXP_ELEMENT_INT(ret2)->value) {
		NXP_ELEMENT_UNREF(ret1);
		return ret2;
	    }
	    break;
	case NXP_OP_UNARY_LEFT:
	    ret2 = nxp_element_tree_simplify_op(NXP_OP_EQ2, aa->right, bb->right, context, error);
	    g_assert(*error == NULL);
	    if (ret2 && NXP_ELEMENT_GET_TYPE(ret2) == NXP_ELEMENT_TYPE_INT
		    && NXP_ELEMENT_INT(ret2)->value)
		return ret2;
	    break;
	case NXP_OP_UNARY_RIGHT:
	    ret1 = nxp_element_tree_simplify_op(NXP_OP_EQ2, aa->left, bb->left, context, error);
	    g_assert(*error == NULL);
	    if (ret1 && NXP_ELEMENT_GET_TYPE(ret1) == NXP_ELEMENT_TYPE_INT
		    && NXP_ELEMENT_INT(ret1)->value)
		return ret2;
	    break;
	default:
	    g_assert_not_reached();
	}
	if (ret1)
	    NXP_ELEMENT_UNREF(ret1);
	if (ret2)
	    NXP_ELEMENT_UNREF(ret2);
    }
    return nxp_element_int_new(0);
}

#define BIND_FUNC(func,op)    nxp_bind_operator_function (op, 	  \
  				NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_TREE, \
  				func,  		               	  \
  				func,   	  	   	  \
  				0, 				  \
				NULL); 

#define BIND(eval,symp,op)    nxp_bind_operator_function (op,	    \
				NXP_ELEMENT_TYPE_TREE, NXP_ELEMENT_TYPE_TREE, \
				eval,				    \
				symp,				    \
				0,				    \
				NULL);

static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc func,
              gpointer user_data,
              NxpEvalContext *context)
{
    NxpElementTree *tree = NXP_ELEMENT_TREE(elem);

    if (tree->left) {
        func(tree->left, user_data);
        if (tree->left->type_info->iterator_func)
            tree->left->type_info->iterator_func(tree->left, func, user_data, context);
    }

    if (tree->right) {
        func(tree->right, user_data);
        if (tree->right->type_info->iterator_func)
            tree->right->type_info->iterator_func(tree->right, func, user_data, context);
    }
}

static gboolean
dependson_func(NxpElement *self, NxpElement *variable,
	       NxpEvalContext *context, GError **error)
{
    NxpElementTree *tree = NXP_ELEMENT_TREE(self);
    gboolean retval = FALSE;
    GError *err = NULL;
    if (tree->left)
	retval = nxp_element_depends_on(tree->left, variable, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return FALSE;
    }
    if (tree->right)
	retval |= nxp_element_depends_on(tree->right, variable, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return FALSE;
    }
    return retval;
}

int nxp_element_tree_get_type(void)
{
    static int nxp_element_type_tree = 0;
    if (nxp_element_type_tree == 0) {
	nxp_element_type_tree = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_tree,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_tree;
}

void nxp_element_tree_type_init()
{
    typeinfo = nxp_element_type_get_info(nxp_element_tree_get_type());
    typeinfo->name = g_strdup("tree");
    typeinfo->eval_func = eval;
    typeinfo->simplify_func = simplify;
    typeinfo->derive_func = derive;
    typeinfo->substitute_func = substitute;
    typeinfo->destroy_func = destroy;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = truth_func;
    typeinfo->iterator_func = iterator_func;
    typeinfo->dependson_func = dependson_func;

    BIND( NULL, eq, NXP_OP_EQ2);

    /* [TODO] to complete ... */
}




