/*-*- c -*- */
#include <gsl/gsl_fft_complex.h>
#include <glib.h>

#include "nxp_kernel.h"
#include "gsl.h"
#include "i18n.h"

static NxpElementVector* gsl_cpacked_array_to_nxp_vector(gsl_complex_packed_array a,
							 int size) {
    int i;
    NxpElementVector *v = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));

    for (i = 0; i < size; i++) {
	NXP_ELEMENT_VECTOR_ELEM(v, i) =
	    nxp_element_complex_new_from_dbl(a[i] + a[i+1] * 1i);
    }
    return v;
}

static gsl_complex_packed_array
nxp_vector_to_gsl_cpacked_array(NxpElementVector * v, GError ** error)
{
    int             i,
                    l = NXP_ELEMENT_VECTOR_SIZE(v);
    gsl_complex_packed_array a = g_new(double, l * 2);
    for (i = 0; i < l; i++) {
	NxpElement     *x = NXP_ELEMENT_VECTOR_ELEM(v, i);
	if (NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_INT) {
	    a[i] = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(x));
	    a[i + 1] = 0.00f;
	}
	else if (NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_REAL) {
	    a[i] = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(x));
	    a[i + 1] = 0.00f;
	}
	else if (NXP_ELEMENT_GET_TYPE(x) == NXP_ELEMENT_TYPE_COMPLEX) {
	    a[i] = (double) CXREAL(nxp_element_complex_get_value(x));
	    a[i + 1] = (double) CXIMAG(nxp_element_complex_get_value(x));
	}
	else {
	    g_free(a);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
			_("Vector elements must not be compound!"));
	    return NULL;
	}
    }
    return a;
}


#define RADIX2(x) static NxpElement* nxp_gsl_fft_complex_radix2_##x(NxpElementVector *v, ETC) {\
    GError *err = NULL;\
    NxpElement *r = NULL;\
    int ec;\
    gsl_complex_packed_array a;\
    a = nxp_vector_to_gsl_cpacked_array(v, &err);\
    if (err) { g_propagate_error(error, err); return NULL; }\
    ec = gsl_fft_complex_radix2_##x(a, NXP_ELEMENT_VECTOR_SIZE(v), 1);\
    if (ec) { g_free(a); GSL_RETURN(NULL); }\
    r = (NxpElement*) gsl_cpacked_array_to_nxp_vector(a, NXP_ELEMENT_VECTOR_SIZE(v));\
    GSL_RETURN(r);\
}

RADIX2(forward);
RADIX2(dif_forward);
RADIX2(backward);
RADIX2(dif_backward);
RADIX2(inverse);
RADIX2(dif_inverse);

#undef RADIX2

static NxpElement* nxp_gsl_fft_complex_radix2_transform(NxpElementVector *v, 
							NxpElementInt *sign, ETC) {
    GError *err = NULL;
    NxpElement *r = NULL;
    int ec;
    gsl_complex_packed_array a;
    a = nxp_vector_to_gsl_cpacked_array(v, &err);
    if (err) { g_propagate_error(error, err); return NULL; }
    ec = gsl_fft_complex_radix2_transform(a, NXP_ELEMENT_VECTOR_SIZE(v), 1, 
					  (nxp_element_r_get_value_as_int(sign)>0)?forward:backward);
    if (ec) { g_free(a); GSL_RETURN(NULL); }
    r = (NxpElement*) gsl_cpacked_array_to_nxp_vector(a, NXP_ELEMENT_VECTOR_SIZE(v));
    GSL_RETURN(r);
}

static NxpElement* nxp_gsl_fft_complex_radix2_dif_transform(NxpElementVector *v, 
							NxpElementInt *sign, ETC) {
    GError *err = NULL;
    NxpElement *r = NULL;
    int ec;
    gsl_complex_packed_array a;
    a = nxp_vector_to_gsl_cpacked_array(v, &err);
    if (err) { g_propagate_error(error, err); return NULL; }
    ec = gsl_fft_complex_radix2_dif_transform(a, NXP_ELEMENT_VECTOR_SIZE(v), 1, 
					  (nxp_element_r_get_value_as_int(sign)>0)?forward:backward);
    if (ec) { g_free(a); GSL_RETURN(NULL); }
    r = (NxpElement*) gsl_cpacked_array_to_nxp_vector(a, NXP_ELEMENT_VECTOR_SIZE(v));
    GSL_RETURN(r);
}


void
nxp_gsl_fft_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;
nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"fft");
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2forward");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_forward, NULL, nxp_gsl_fft_complex_radix2_forward, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2difForward");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_dif_forward, NULL, nxp_gsl_fft_complex_radix2_dif_forward, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2backward");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_backward, NULL, nxp_gsl_fft_complex_radix2_backward, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2difBackward");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_dif_backward, NULL, nxp_gsl_fft_complex_radix2_dif_backward, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2inverse");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_inverse, NULL, nxp_gsl_fft_complex_radix2_inverse, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2difInverse");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_fft_complex_radix2_dif_inverse, NULL, nxp_gsl_fft_complex_radix2_dif_inverse, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2transform");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR_INT, nxp_gsl_fft_complex_radix2_transform, NULL, nxp_gsl_fft_complex_radix2_transform, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","fft", "radix2difTransform");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR_INT, nxp_gsl_fft_complex_radix2_dif_transform, NULL, nxp_gsl_fft_complex_radix2_dif_transform, context, NULL); nxp_identifier_unref(id);
}
