/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <string.h>
#include "nxp_kernel.h"
#include "gsl.h"
#include "i18n.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_complex_math.h>
#include "types/types.h"

int gsl_is_on_error;
char* gsl_error_message;

void nxp_gsl_error_handler(const char* reason, const char* file, int line, int gsl_errno) {
    if (!gsl_is_on_error) {
	/* This way, we get the first of the error messages */
	gsl_is_on_error = 1;
	gsl_error_message = strdup(reason);
    }
}

void nxp_gsl_poly_roots_init(NxpEvalContext*);
void nxp_gsl_physical_init(NxpEvalContext*);
void nxp_gsl_basics_init(NxpEvalContext*);
void nxp_gsl_complexes_init(NxpEvalContext*);
void nxp_gsl_permutation_init(NxpEvalContext*);
void nxp_gsl_linalg_init(NxpEvalContext*);
void nxp_gsl_stats_init(NxpEvalContext*);
void nxp_gsl_rng_init(NxpEvalContext*);
void nxp_gsl_qrng_init(NxpEvalContext*);
void nxp_gsl_randist_init(NxpEvalContext*);
void nxp_gsl_histogram_init(NxpEvalContext*);
void nxp_gsl_interp_init(NxpEvalContext*);
void nxp_gsl_roots_init(NxpEvalContext*);
void nxp_gsl_fft_init(NxpEvalContext*);
void nxp_gsl_integration_init(NxpEvalContext*);

void nxp_module_init(NxpEvalContext *context, GError **error)
{
    gsl_is_on_error = 0;
    gsl_error_message = NULL;

    gsl_set_error_handler(nxp_gsl_error_handler);

    nxp_gsl_basics_init(context);
    nxp_gsl_complexes_init(context);
    nxp_gsl_poly_roots_init(context);
    nxp_gsl_sf_init(context);
    nxp_gsl_permutation_init(context);
    nxp_gsl_linalg_init(context);
    nxp_gsl_stats_init(context);
    nxp_gsl_rng_init(context);
    nxp_gsl_qrng_init(context);
    nxp_gsl_randist_init(context);
    nxp_gsl_histogram_init(context);
    nxp_gsl_interp_init(context);
    nxp_gsl_roots_init(context);
    nxp_gsl_integration_init(context);

    nxp_gsl_fft_init(context);

    nxp_gsl_physical_init(context);
}


/* MISC */

NxpElement *nxp_pair_new(NxpElement *x,
			 NxpElement *y) {
    NxpElement *vector = nxp_element_vector_new(2, NULL);
    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(vector),0) = x;
    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(vector),1) = y;
    return vector;
}

double* nxp_vector_to_double_buffer(NxpElementVector* v, NxpEvalContext *context, GError **error) {
    double *buffer;
    NxpElement *el,*ev;
    int i;
    buffer = g_new(double, v->size);
 
    for (i=0; i<v->size; ++i) {
	el = NXP_ELEMENT_VECTOR_ELEM(v, i);

	ev = nxp_element_eval(el, context, NULL); /* FIXME */

	if (NXP_ELEMENT_IS_R(ev))
	    *(buffer+i) = (double)nxp_element_r_get_value_as_double(NXP_ELEMENT_R(ev));
	else {
	    g_free(buffer);
	    g_set_error(error, nxp_error_domain,
			NXP_ERROR_TYPE_CHECK, _("Vector can contain only integers and reals"));
	    NXP_ELEMENT_UNREF(ev);
	    return NULL;
	}
	NXP_ELEMENT_UNREF(ev);
    }
    return buffer;
}

gsl_complex nxp_gsl_complex_from_real(double real) {
    return gsl_complex_rect(real,0);
}
gsl_complex nxp_gsl_complex_from_nxp_complex(NxpElementComplex* c) 
{
    nxp_complex_t cpx = nxp_element_complex_get_value(c);
    return gsl_complex_rect(CXREAL(cpx), CXIMAG(cpx));
}

NxpElement* nxp_complex_from_gsl_complex(gsl_complex c) 
{
    return nxp_element_complex_new_from_dbl(GSL_REAL(c) + GSL_IMAG(c) * I);
}

NxpElement* nxp_double_buffer_to_vector(double* buffer, int size) {
    int i;
    NxpElement* v = nxp_element_vector_new(size, NULL);
    for (i=0;i<size;++i) 
	NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(v),i) = nxp_element_real_new(buffer[i]);
    return v;
    
}

gsl_vector* nxp_vector_to_gsl_vector(NxpElementVector *v, NxpEvalContext *context, GError **error) {
    gsl_vector* new;
    GError *err = NULL;
    int i;
    double *buffer = nxp_vector_to_double_buffer(v, context, &err); 
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    new = gsl_vector_alloc(v->size);
    for (i=0; i<v->size; i++)
	gsl_vector_set(new, i, buffer[i]);
    return new;
}

NxpElement* gsl_vector_to_nxp_vector(gsl_vector* v) {
    int i,size = v->size;
    NxpElement *new;
    new = nxp_element_vector_new(size, NULL);
    if (!new) return NULL;
    for (i=0; i<size; ++i)
	NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(new), i) = 
	    nxp_element_real_new(gsl_vector_get(v, i));
    gsl_vector_free(v);
    return new;
}

gsl_matrix* nxp_matrix_to_gsl_matrix(NxpElementMatrix *m, GError **error) {
    int i,j;
    int width = m->width;
    int height= m->height;
    gsl_matrix *new = gsl_matrix_alloc(width, height);
    for(i=0;i<width;++i) {
	for (j=0;j<height;++j) {
	    NxpElement *e = NXP_ELEMENT_MATRIX_ELEM_XY(m, i, j);
	    if (NXP_ELEMENT_IS_R(e)) {
		gsl_matrix_set(new, j, i,
			nxp_element_r_get_value_as_double(NXP_ELEMENT_R(e)));
	    } else {
		gsl_matrix_free(new);
		g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
			    _("Matrix must contain only integers and reals"));
		return NULL;
	    }
	}
    }
    return new;
}

NxpElement* gsl_matrix_to_nxp_matrix(gsl_matrix* m) {
    NxpElement *new;
    int width = m->size1;
    int height = m->size2;
    int i,j;
    new = nxp_element_matrix_new(width, height, NULL);
    for (i=0;i<width;++i) {
	for (j=0;j<height;++j) {
	    NXP_ELEMENT_MATRIX_ELEM_YX(NXP_ELEMENT_MATRIX(new), i, j) = 
		nxp_element_real_new(gsl_matrix_get(m, i, j));
	}
    }
    gsl_matrix_free(m);
    return new;
}

/* returns 0 if it is not constant
   returns 1 if it contains only integers and reals
   returns 2 if it contains complexes */
gint nxp_matrix_is_constant(NxpElementMatrix *m) {
    gint result = 1;
    int width = m->width;
    int height = m->height;
    int i,j;
    for (i=0; i<width; ++i) {
	for (j=0; j<height; ++j) {
	    NxpElement *el = NXP_ELEMENT_MATRIX_ELEM_XY(m, i, j);
	    if (NXP_ELEMENT_IS_COMPLEX(el)) {
		result = 2;
	    } else if (!NXP_ELEMENT_IS_NUMBER(el)) return 0;
	}
    }
    return result;
}


gsl_matrix_complex* nxp_matrix_to_gsl_cmatrix(NxpElementMatrix* m, GError **error) {
    int i,j;
    int width = m->width;
    int height= m->height;
    gsl_matrix_complex *new = gsl_matrix_complex_alloc(width, height);
    for(i=0;i<width;++i) {
	for (j=0;j<height;++j) {
	    NxpElement *e = NXP_ELEMENT_MATRIX_ELEM_YX(m, i, j);
	    if (NXP_ELEMENT_IS_COMPLEX(e)) {
		gsl_matrix_complex_set(new, j, i, 
				       nxp_gsl_complex_from_nxp_complex(NXP_ELEMENT_COMPLEX(e)));
	    } else if (NXP_ELEMENT_IS_R(e)) {
		gsl_matrix_complex_set(new, j, i,
				       nxp_gsl_complex_from_real(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(e))));
	    } else {
		g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, 
			    _("Matrix should contain only integers, reals and complex values"));
		gsl_matrix_complex_free(new);
		return NULL;
	    }
	}
    }
    return new;
}

NxpElement* gsl_cmatrix_to_nxp_matrix(gsl_matrix_complex* m) {
    NxpElement *new;
    int width = m->size1;
    int height = m->size2;
    int i,j;
    new = nxp_element_matrix_new(width, height, NULL);
    for (i=0;i<width;++i)
	for (j=0;j<height;++j) {
	    NXP_ELEMENT_MATRIX_ELEM_YX(NXP_ELEMENT_MATRIX(new), i, j) = 
		nxp_complex_from_gsl_complex(gsl_matrix_complex_get(m, i, j));
	}
    gsl_matrix_complex_free(m);
    return new;
}

gsl_vector_complex* nxp_vector_to_gsl_cvector(NxpElementVector* m, GError **error) {
    int i;
    int size = m->size;
    gsl_vector_complex *new = gsl_vector_complex_alloc(size);

    for(i=0;i<size;++i) {
	NxpElement *e = NXP_ELEMENT_VECTOR_ELEM(m, i);
	if (NXP_ELEMENT_IS_COMPLEX(e)) {
	    gsl_vector_complex_set(new, i,
				   nxp_gsl_complex_from_nxp_complex(NXP_ELEMENT_COMPLEX(e)));
	} else if (NXP_ELEMENT_IS_R(e)) {
	    gsl_vector_complex_set(new, i,
				   nxp_gsl_complex_from_real(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(e))));
	} else if (NXP_ELEMENT_IS_COMPLEX(e)) {
	    gsl_vector_complex_set(new, i,
				   nxp_gsl_complex_from_nxp_complex(NXP_ELEMENT_COMPLEX(e)));
	} else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, 
			_("Vector should contain only integers, reals and complex values"));
	    gsl_vector_complex_free(new);
	    return NULL;
	}
    }
    return new;
}

NxpElement* gsl_cvector_to_nxp_vector(gsl_vector_complex* m) {
    NxpElement *new;
    int size = m->size;
    int i;
    new = nxp_element_vector_new(size, NULL);
    for (i=0;i<size;++i) {
	NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(new), i) = 
	    nxp_complex_from_gsl_complex(gsl_vector_complex_get(m, i));
    }
    gsl_vector_complex_free(m);
    return new;
}
