#include "mathml.h"
#include "types/frac.h"
#include "nxp_kernel.h"

static GHashTable *symbols;

/* Some MathML entities in UTF-8 */
#define ENT_af                   "\342\201\241"
#define ENT_InvisibleTimes       "\342\201\242"
#define ENT_DifferentialD        "\340\241\242"
#define ENT_CapitalDifferentialD "\340\241\241"
#define ENT_ExponentialE         "\340\241\243"
#define ENT_ImaginaryI           "\342\205\210"
#define ENT_Times                "\303\227"
#define ENT_TripleDot            "\342\203\233"
#define ENT_geq                  "\342\211\245"
#define ENT_leq                  "\342\211\244"
#define ENT_NotEqual             "\342\211\240"
#define ENT_wedge                "\342\210\247"
#define ENT_vee                  "\342\210\250"
#define ENT_not                  "\302\254"
#define ENT_Equals		 "\075"
#define ENT_Assign               "\342\211\224"
#define ENT_alpha                "\316\261"
#define ENT_beta                 "\316\262"
#define ENT_gamma                "\316\263"
#define ENT_delta                "\316\264"
#define ENT_epsi                 "\316\265"
#define ENT_zeta                 "\316\266"
#define ENT_eta                  "\316\267"
#define ENT_theta                "\316\270"
#define ENT_iota                 "\316\271"
#define ENT_kappa                "\316\272"
#define ENT_lambda               "\316\273"
#define ENT_mu                   "\316\274"
#define ENT_nu                   "\316\275"
#define ENT_xi                   "\316\276"
#define ENT_pi                   "\317\200"
#define ENT_rho                  "\317\201"
#define ENT_sigma                "\317\203"
#define ENT_tau                  "\317\204"
#define ENT_upsilon              "\317\205"
#define ENT_phi                  "\317\225"
#define ENT_chi                  "\317\207"
#define ENT_psi                  "\317\210"
#define ENT_omega                "\317\211"
#define ENT_Gamma                "\316\223"
#define ENT_Delta                "\316\224"
#define ENT_Theta                "\316\230"
#define ENT_Lambda               "\316\233"
#define ENT_Xi                   "\316\236"
#define ENT_Pi                   "\316\240"
#define ENT_Sigma                "\316\243"
#define ENT_Upsilon              "\316\245"
#define ENT_Psi                  "\316\250"
#define ENT_Phi                  "\316\246"
#define ENT_Omega                "\316\251"
#define ENT_vert                 "\174"



static GString* htmlEscape(const char* string) {
    guint i, size;
    GString *str;
    i = 0;
    size = strlen(string);
    str = g_string_new("");
    while(i < size) {
	switch(string[i]) {
	case '<':
	    str = g_string_append(str, "&lt;");
	    break;
	case '>':
	    str = g_string_append(str, "&gt;");
	    break;
	case '&':
	    str = g_string_append(str, "&amp;");
	    break;
	default:
	    str = g_string_append_c(str, string[i]);
	}
	i++;
    }
    return str;
}

static GString* string2mathml(NxpElement *elem, int *prec, GError **error)
{
    NxpElementString *str = NXP_ELEMENT_STRING(elem);
    GString *gstr, *escaped = htmlEscape(str->string);
    if (prec) *prec = 0;
    gstr = g_string_new("<mtext>");
    gstr = g_string_append(gstr, escaped->str);
    g_string_free(escaped, TRUE);
    gstr = g_string_append(gstr, "</mtext>");
    return gstr;
}

static GString* matrix2mathml(NxpElement *elem, int *precedence, GError **error)
{
    GString *str = g_string_new("<mrow><mo>[</mo><mtable>");
    GError *err = NULL;
    int i, j;
    NxpElement *el;
    NxpElementMatrix* m;
    GString* old;
    
    if (precedence) *precedence = 0;

    m = NXP_ELEMENT_MATRIX(elem);
    for (i = 0; i < m->height; i++) {
      str = g_string_append(str, "<mtr>");
      for (j = 0; j < m->width; j++) {
	el = _element_matrix_get(m, i, j);

	old = nxp_mathml(el, NULL, &err); 
	if (err) {
	    if (old) g_string_free(old, TRUE);
	    g_string_free(str, TRUE);
	    g_propagate_error(error, err);
	    return NULL;
	}

	NXP_ELEMENT_UNREF(el);

	str = g_string_append(str, "<mtd><mrow>");
	str = g_string_append(str, old->str);
	str = g_string_append(str, "</mrow></mtd>");
	g_string_free(old, TRUE);
      }
      str = g_string_append(str, "</mtr>");
    }
    str = g_string_append(str, "</mtable><mo>]</mo></mrow>");

    return str;
}

static GString* vector2mathml(NxpElement *elem, int *precedence, GError **error)
{
    GString *string = g_string_new("<mrow><mo>{</mo>");
    GError *err = NULL;
    int i;
    NxpElement *el;
    NxpElementVector* v;
    GString* old;

    if (precedence) *precedence = 0;

    v = NXP_ELEMENT_VECTOR(elem);
    for (i = 0; i < v->size; i++) {
	el = _element_vector_get(v, i);

	old = nxp_mathml(el, NULL, &err); 
	if (err) {
	    if (old) g_string_free(old, TRUE);
	    g_string_free(string, TRUE);
	    g_propagate_error(error, err);
	    return NULL;
	}

	NXP_ELEMENT_UNREF(el);
	old = g_string_prepend(old, "<mrow>");
	old = g_string_append(old, "</mrow>");
	string = g_string_append(string, old->str);
	string = g_string_append(string, "<mo>,</mo>");
	g_string_free(old, TRUE);
    }
    if (v->size) string = g_string_truncate(string, (string)->len - 10);
    string = g_string_append(string, "<mo>}</mo></mrow>");
    
    return string;
}

static GString* func2mathml(NxpElement *elem, int *prec, GError **error)
{
    GString *str;
    GSList *arg;
    NxpElementFunction *func = NXP_ELEMENT_FUNCTION(elem);

    /* TODO: abs can be a special case, too...? */


    if (strcmp(nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(func)->ident),"abs") == 0
	    || strcmp(nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(func)->ident),"::abs") == 0) {
        /* abs is a special case! */
        str = nxp_mathml(NXP_ELEMENT(NXP_ELEMENT_LIST(func->simple_args)->list->data), NULL, NULL);
        str = g_string_prepend(str, "<mrow><mo form='prefix' fence='true' "
                               "stretchy='true' lspace='0em' rspace='0em'>" ENT_vert "</mo>");
        str = g_string_append(str, "<mo form='postfix' fence='true' "
                              "stretchy='true' lspace='0em' rspace='0em'>" ENT_vert "</mo></mrow>");
    } else {
    /* sqrt is a special case! */
        if (strcmp(nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(func)->ident),"sqrt") == 0
		|| strcmp(nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(func)->ident),"::sqrt") == 0) {
            str = nxp_mathml(NXP_ELEMENT(NXP_ELEMENT_LIST(func->simple_args)->list->data), NULL, NULL);
            str = g_string_prepend(str, "<msqrt>");
            str = g_string_append(str, "</msqrt>");
        } else {
            str = nxp_identifier_to_gstring(NXP_ELEMENT_VARIABLE(func)->ident);
            
            str = g_string_prepend(str, "<mrow><mi>");
            str = g_string_append(str, "</mi><mo>" ENT_af "</mo><mrow><mo>(</mo>");
            
            if (func->simple_args) {
                for (arg = NXP_ELEMENT_LIST(func->simple_args)->list; arg; arg = arg->next)
                {
                    GString *str1;
                    NxpElement *el;
                    el = NXP_ELEMENT(arg->data);
                    str1 = nxp_mathml(el, NULL, NULL);
                    str = g_string_append(str, str1->str);
                    g_string_free(str1, TRUE);
                    if (arg->next)
                        str = g_string_append(str, "<mo>,</mo>");
                }
            }
            str = g_string_append(str, "<mo>)</mo></mrow></mrow>");
        }
    }
    if (prec) *prec = 0;
    return str;
}

static GString* mn(NxpElement *el, int *precedence, GError **error) {
    GString *repr;
    if (precedence) *precedence = 0;
    repr = nxp_element_repr(el);
    repr = g_string_prepend(repr,"<mn>");
    repr = g_string_append(repr,"</mn>");
    return repr;
}

static GString* mpoly2mathml(NxpElement *elem, int *precedence, GError **error)
{
    GSList *args, *mon;
    NxpElementMpoly *pol = NXP_ELEMENT_MPOLY(elem);
    NxpElement *nb;
    GString *string, *var, *var1, *coeff, *coeff1;
    int prec;
    int i, pow, n_mon;
    int s, c, c1;
    gboolean star = FALSE, powers;

    if (precedence) *precedence = 0;
    string = g_string_new("");
    for (n_mon = 0, mon = pol->mon->list; mon; mon = mon->next, n_mon++) {
	NxpElementNumber *numb = NXP_ELEMENT_NUMBER(NXP_MONOMIAL_COEFF(mon->data));
	s = nxp_element_number_compare_int(numb, 0);
	c = nxp_element_number_compare_int(numb, 1);
	c1 = nxp_element_number_compare_int(numb, -1);

	coeff = g_string_new("");

	if (s != 0) {
	    if (string->str[0] && s > 0)
		g_string_append_printf(coeff, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_PLUS)->token);
	    else if (string->str[0] && s < 0)
		g_string_append_printf(coeff, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_MINUS)->token);
	    else if (string->str[0] == 0 && s < 0) {
		g_string_append_printf(coeff, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_MINUS_LEFT)->token);
	    }

	    powers = nxp_element_is_true(NXP_MONOMIAL_VECTOR(mon->data),
					 NULL, NULL);
	    if ((c && c1) || !powers) {
		if (s > 0) {
		    coeff1 = nxp_mathml(NXP_ELEMENT(numb), &prec, NULL);
		}
		else {
		    nb = nxp_element_number_opposed(numb);
		    coeff1 = nxp_mathml(nb, &prec, NULL);
		    NXP_ELEMENT_UNREF(nb);
		}
		if (prec > 40) {
		    coeff1 = g_string_prepend(coeff1, "<mrow><mo>(</mo>");
		    coeff1 = g_string_append(coeff1, "<mo>)</mo></mrow>");
		}
		coeff = g_string_append(coeff, coeff1->str);
		g_string_free(coeff1, TRUE);
	    }

	    string = g_string_append(string, coeff->str);

	    if (powers) {
		if (c && c1)
		    g_string_append_printf(string, "<mo>"ENT_InvisibleTimes"</mo>");

		var = g_string_new("");
		for (args = pol->args->list, i = 0; args; args = args->next, i++) {
		    var1 = nxp_mathml(NXP_ELEMENT(args->data), &prec, error);
		    if (prec >= 40) {
			var1 = g_string_prepend(var1, "<mrow><mo>(</mo>");
			var1 = g_string_append(var1, "<mo>)</mo></mrow>");
		    }
		    pow = nxp_element_r_get_value_as_int(NXP_ELEMENT_R(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mon->data)), i)));
		    if (pow) {
			if (var->str[0] != 0) {
			    var = g_string_append(var, "<mo>"ENT_InvisibleTimes"</mo>");
			    star = TRUE;
			}
			if (pow == 1)
			    g_string_append_printf(var, "%s", var1->str);
			else
			    g_string_append_printf(var, "<mrow><msup>%s<mn>%d</mn></msup></mrow>", var1->str, pow);
		    }
		    g_string_free(var1, TRUE);
		}
		string = g_string_append(string, var->str);
		g_string_free(var, TRUE);
	    }
	}
    }
    if (string->str[0] == 0)
	string = g_string_append(string, "<mn>0</mn>");

    if (precedence) {
	if (n_mon > 1)
	    * precedence = 50;
	else if (star)
	    * precedence = 40;
	else
	    * precedence = 30;
    }

    return string;
}

static GString* tree2mathml(NxpElement *elem, int *prec, GError **error) 
{
    NxpElementTree *tree = NXP_ELEMENT_TREE(elem);
    const NxpOpInfo *opinfo;
    GString *result, *tmp = NULL;
    gint left = 0, right = 0, myprecedence;

    int type_left = 0, type_right = 0;

    // Get operator information and precedence
    opinfo = nxp_op_get_info(tree->op);
    myprecedence = opinfo->precedence;
    if (prec) *prec = myprecedence;
    if (opinfo->op == NXP_OP_SLASH) myprecedence = 10000;

    result = g_string_new("");

    if (tree->left) {
	type_left = NXP_ELEMENT_GET_TYPE(tree->left);

	tmp = nxp_mathml(tree->left, &left, NULL);
	if (myprecedence < left) {
	    tmp = g_string_prepend(tmp, "<mrow><mo>(</mo>");
	    tmp = g_string_append(tmp, "<mo>)</mo></mrow>");
	}
	result = g_string_append(result, tmp->str);
	g_string_free(tmp, TRUE);
    }


    if (tree->right)
	type_right = NXP_ELEMENT_GET_TYPE(tree->right);

    switch(tree->op) {
    case NXP_OP_PLUS:
    case NXP_OP_MINUS:
    case NXP_OP_MINUS_LEFT:
	g_string_append_printf(result, "<mo>%s</mo>", opinfo->token);
	break;
    case NXP_OP_STAR:
/*	if ((type_left == NXP_ELEMENT_TYPE_INT ||
	     type_left == NXP_ELEMENT_TYPE_REAL ||
	     type_left == NXP_ELEMENT_TYPE_COMPLEX) &&
	    (type_right == NXP_ELEMENT_TYPE_INT ||
	     type_right == NXP_ELEMENT_TYPE_REAL ||
	     type_right == NXP_ELEMENT_TYPE_COMPLEX)) {
	    result = g_string_append(result, "<mo>"ENT_Times"</mo>");
	} else
	    result = g_string_append(result, "<mo>"ENT_InvisibleTimes"</mo>");*/
	if (NXP_ELEMENT_IS_NUMBER(tree->right))
	    result = g_string_append(result, "<mo>"ENT_Times"</mo>");
	else
	    result = g_string_append(result, "<mo>"ENT_InvisibleTimes"</mo>");
	break;
    case NXP_OP_SLASH:
	result = g_string_prepend(result, "<mfrac>");
	break;
    case NXP_OP_HAT:
	result = g_string_prepend(result, "<msup>");
	break;
    case NXP_OP_DOT3:
	result = g_string_append(result, "<mo>"ENT_TripleDot"</mo>");
	break;
    case NXP_OP_PLUS2_LEFT:
	result = g_string_prepend(result, "<mo>+</mo><mo>+</mo>");
	break;
    case NXP_OP_MINUS2_LEFT:
	result = g_string_prepend(result, "<mo>-</mo><mo>-</mo>");
	break;
    case NXP_OP_PLUS2_RIGHT:
	result = g_string_append(result, "<mo>+</mo><mo>+</mo>");
	break;
    case NXP_OP_MINUS2_RIGHT:
	result = g_string_append(result, "<mo>-</mo><mo>-</mo>");
	break;
    case NXP_OP_GT:
	result = g_string_append(result, "<mo>&gt;</mo>");
	break;
    case NXP_OP_LT:
	result = g_string_append(result, "<mo>&lt;</mo>");
	break;
    case NXP_OP_EQ2:
	result = g_string_append(result, "<mo>=</mo>");
	break;
    case NXP_OP_GT_EQ:
	result = g_string_append(result, "<mo>"ENT_geq"</mo>");
	break;
    case NXP_OP_LT_EQ:
	result = g_string_append(result, "<mo>"ENT_leq"</mo>");
	break;
    case NXP_OP_EXCL_EQ:
	result = g_string_append(result, "<mo>"ENT_NotEqual"</mo>");
	break;
    case NXP_OP_SLASH_BSLASH:
	result = g_string_append(result, "<mo>"ENT_wedge"</mo>");
	break;
    case NXP_OP_BSLASH_SLASH:
	result = g_string_append(result, "<mo>"ENT_vee"</mo>");
	break;
    case NXP_OP_TILDE_LEFT:
	result = g_string_append(result, "<mo>"ENT_not"</mo>");
	break;
    case NXP_OP_EXCL_RIGHT:
	result = g_string_append(result, "<mo>!</mo>");
	break;
    case NXP_OP_EQ:
	result = g_string_append(result, "<mo>"ENT_Equals"</mo>");
	break;
    case NXP_OP_COLON_EQ:
	result = g_string_append(result, "<mo>"ENT_Assign"</mo>");
	break;
    case NXP_OP_PERCENT:
	// ---------- CHECK WHAT TODO
	break;
    default:
	break;
    }

    if (tree->right) {
	tmp = nxp_mathml(tree->right, &right, NULL);
	if (myprecedence < right) {
	    tmp = g_string_prepend(tmp, "<mrow><mo>(</mo>");
	    tmp = g_string_append(tmp, "<mo>)</mo></mrow>");
	}
	result = g_string_append(result, tmp->str);
	g_string_free(tmp, TRUE);
    }

    if (tree->op == NXP_OP_SLASH)
	result = g_string_append(result, "</mfrac>");
    if (tree->op == NXP_OP_HAT)
	result = g_string_append(result, "</msup>");

    result = g_string_prepend(result, "<mrow>");
    result = g_string_append(result, "</mrow>");

    return result;
}

static GString* mcomplex(NxpElement *elem, int *precedence, GError **error)
{
    GString *string, *mon;
    GString *old;
    int i, length = 0, diff;
    NxpElementComplex* p;
    NxpElementVector *vector;
    NxpElement *coeff, *coeff1;
    gboolean prod;

    if (precedence) *precedence = 0;
    p = NXP_ELEMENT_COMPLEX(elem);
    vector = p->vector;
    string = g_string_new("");
    for (i = 0; i <= 1; i++) {
	coeff = _element_vector_index(vector, i);
	diff = nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(coeff), 0);
	if (diff) {
	    mon = g_string_new("");
	    prod = FALSE;
	    length ++;
	    if (diff < 0) {
		coeff1 = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(coeff));
		old = nxp_mathml(coeff1, NULL, NULL);
		NXP_ELEMENT_UNREF(coeff1);
		if (*string->str)
		    g_string_append_printf(string, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_MINUS)->token);
		else
		    g_string_append_printf(mon, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_MINUS_LEFT)->token);
		if (i == 0 || nxp_element_number_compare_int(
					NXP_ELEMENT_NUMBER(coeff), -1)) {
		    g_string_append_printf(mon, "%s%s", old->str,
			    i == 0 ? "" : "<mo>"ENT_InvisibleTimes"</mo>");
		    prod = (i != 0);
		}
	    }
	    else {
		old = nxp_mathml(coeff, NULL, NULL);
		if (*string->str)
		    g_string_append_printf(string, "<mo>%s</mo>", nxp_op_get_info(NXP_OP_PLUS)->token);
		if (i == 0 || nxp_element_number_compare_int(
					NXP_ELEMENT_NUMBER(coeff), 1)) {
		    g_string_append_printf(mon, "%s%s", old->str,
			    i == 0 ? "" : "<mo>"ENT_InvisibleTimes"</mo>");
		    prod = (i != 0);
		}
	    }
	    if (i == 1)
		g_string_append_printf(mon, "<mi>"ENT_ImaginaryI"</mi>");
	    g_string_free(old, TRUE);
	    if (prod) {
		mon = g_string_prepend(mon, "<mrow>");
		mon = g_string_append(mon, "</mrow>");
	    }
	    string = g_string_append(string, mon->str);
	    g_string_free(mon, TRUE);
	}
    }

    if (precedence) {
	if (length > 1)
	    * precedence = 50;
	else
	    * precedence = 40;
    }
    return string;
}

static GString* frac(NxpElement *el, int *precedence, GError **error) {
    GString *repr, *num_repr, *den_repr;
    NxpElement *num, *den;
    if (precedence) *precedence = 0;

    repr = g_string_new("");
    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(el), 0) >= 0) {
	num = nxp_element_frac_get_num(el);
	den = nxp_element_frac_get_den(el);
	num_repr = nxp_mathml(num, NULL, NULL);
	den_repr = nxp_mathml(den, NULL, NULL);
	NXP_ELEMENT_UNREF(num);
	NXP_ELEMENT_UNREF(den);
	g_string_printf(repr,
			"<mfrac>%s%s</mfrac>",
			num_repr->str, den_repr->str);
    }
    else {
	den = nxp_element_frac_get_num(el);
	num = nxp_element_number_opposed(NXP_ELEMENT_NUMBER(den));
	NXP_ELEMENT_UNREF(den);
	den = nxp_element_frac_get_den(el);
	num_repr = nxp_mathml(num, NULL, NULL);
	den_repr = nxp_mathml(den, NULL, NULL);
	NXP_ELEMENT_UNREF(num);
	NXP_ELEMENT_UNREF(den);
	g_string_printf(repr,
			"<mrow><mo>-</mo><mfrac>%s%s</mfrac></mrow>",
			num_repr->str, den_repr->str);
    }
    g_string_free(num_repr, TRUE);
    g_string_free(den_repr, TRUE);
    return repr;
}


static GString* mi(NxpElement *el, int *precedence, GError **error)
{
    GString *repr;
    const char *mathml = NULL;
    if (precedence) *precedence = 0;
    repr = nxp_element_repr(el);
    if (strcmp(repr->str, "void")) {
	if ((mathml = g_hash_table_lookup(symbols, repr->str)))
	    g_string_assign(repr, mathml);
	repr = g_string_prepend(repr,"<mi>");
	repr = g_string_append(repr,"</mi>");
    }
    return repr;
}

GString* nxp_mathml(NxpElement* element, gint *precedence, GError **error) {
    NxpElementTypeInfo *info;
    NxpMathMLrepr func;

    info = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(element));

    if (info->type == NXP_ELEMENT_TYPE_VOID) return g_string_new("void");

    func = (NxpMathMLrepr)nxp_element_type_get_data(info, "mathml");

    if (func)
	return (*func)(element, precedence, NULL);
    return g_string_new("??");
}

static NxpElement* mathml(NxpElement* element, ETC)
{
    gint precedence;
    GString *mathml = nxp_mathml(element, &precedence, NULL);
    NxpElement *rr = nxp_element_string_new(mathml->str);
    g_string_free(mathml, TRUE);
    return rr;
}

static NxpElement* version(ETC)
{
    GString *version = g_string_new("Num\316\243\317\207\317\201");
    g_string_append_printf(version," %s",nxp_kernel_version());
    NxpElement *rr = nxp_element_string_new(version->str);
    g_string_free(version, TRUE);
    return rr;
}


static void nxp_mathml_init_symbols(GHashTable *hash)
{
#define ADD(name)  g_hash_table_insert(hash, #name, ENT_##name)
#define ADD2(name, replacement)  g_hash_table_insert(hash, #name, replacement)
    ADD(alpha);
    ADD(beta);
    ADD(gamma);
    ADD(delta);
    ADD2(epsilon, ENT_epsi);
    ADD(zeta);
    ADD(eta);
    ADD(theta);
    ADD(iota);
    ADD(kappa);
    ADD(lambda);
    ADD(mu);
    ADD(nu);
    ADD(xi);
    ADD(pi);
    ADD(rho);
    ADD(sigma);
    ADD(tau);
    ADD(upsilon);
    ADD(phi);
    ADD(chi);
    ADD(psi);
    ADD(omega);

    ADD(Gamma);
    ADD(Delta);
    ADD(Theta);
    ADD(Lambda);
    ADD(Xi);
    ADD(Pi);
    ADD(Sigma);
    ADD(Upsilon);
    ADD(Psi);
    ADD(Phi);
    ADD(Omega);
#undef ADD
}

static void nxp_mathml_init(NxpEvalContext *context) {
    NxpIdentifier *id, *vid;
    GError *err = NULL;

    id  = nxp_identifier_new();
    vid = nxp_identifier_new();

    nxp_identifier_append_separator(id);
    nxp_identifier_append_separator(vid);

    nxp_identifier_append_component(id, "mathml");
    nxp_identifier_append_component(vid, "version");

    nxp_bifunc_define(id, &nxp_prototype__ANY, mathml, NULL, NULL, context, NULL);
    if (err) {
	g_warning("%s",err->message);
	err = NULL;
    }

    nxp_bifunc_define(vid,&nxp_prototype__VOID, version, NULL, NULL, context, &err);
    if (err) g_warning("%s",err->message);

    nxp_identifier_unref(id);
    nxp_identifier_unref(vid);

    symbols = g_hash_table_new(g_str_hash, g_str_equal);
    nxp_mathml_init_symbols(symbols);

#define REG_FUNC(datatype, reprfunc) nxp_element_type_set_data(nxp_element_type_get_info(datatype),\
                                                               "mathml", reprfunc)
    REG_FUNC(NXP_ELEMENT_TYPE_REAL, mn);

    REG_FUNC(NXP_ELEMENT_TYPE_FRAC, frac);
    
    REG_FUNC(NXP_ELEMENT_TYPE_INT, mn);

    REG_FUNC(NXP_ELEMENT_TYPE_VARIABLE, mi);

    REG_FUNC(NXP_ELEMENT_TYPE_COMPLEX, mcomplex);
    
    REG_FUNC(NXP_ELEMENT_TYPE_TREE, tree2mathml);

    REG_FUNC(NXP_ELEMENT_TYPE_MPOLY, mpoly2mathml);

    REG_FUNC(NXP_ELEMENT_TYPE_FUNCTION, func2mathml);

    REG_FUNC(NXP_ELEMENT_TYPE_STRING, string2mathml);
    REG_FUNC(NXP_ELEMENT_TYPE_MATRIX, matrix2mathml);
    REG_FUNC(NXP_ELEMENT_TYPE_VECTOR, vector2mathml);

#undef REG_FUNC
}

void nxp_module_init(NxpEvalContext *context, GError **error)
{
    nxp_mathml_init(context);
}
